/*
 * COPYRIGHT NOTICE
 * 
 * The information contained herein and which follows is proprietary information
 * and is the property of PrismTech Corporation. This information is not to be
 * divulged, released, copied, reproduced, or conveyed to unauthorized
 * personnel, companies, or other institutions without the direct and expressed
 * approval in writing of PrismTech Corporation.
 * 
 * DISCLAIMER
 * 
 * Neither PrismTech Limited nor PrismTech Corporation ("PrismTech") are under
 * any obligation to support the use of this code which is provided without
 * warranty of any kind. PrismTech shall (except as otherwise prohibited by law)
 * have no liability whatsoever in relation to its use.
 */
package com.prismtech.idt.idlmm.constraints;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.core.runtime.IStatus;
import org.eclipse.emf.validation.IValidationContext;

import com.prismtech.idt.idlmm.Contained;
import com.prismtech.idt.idlmm.Container;
import com.prismtech.idt.idlmm.EnumDef;
import com.prismtech.idt.idlmm.EnumMember;
import com.prismtech.idt.idlmm.InterfaceDecl;
import com.prismtech.idt.idlmm.InterfaceDef;
import com.prismtech.idt.idlmm.ModuleDef;
import com.prismtech.idt.idlmm.Named;
import com.prismtech.idt.idlmm.Repository;
import com.prismtech.idt.idlmm.ValueDecl;
import com.prismtech.idt.idlmm.ValueDef;


/**
 * @see <a
 *      href="https://scarab.prismtech.com/twiki/bin/view/MDD/IdtModelChecker">IDT-ModelConstraints</a>
 * @constrain 65, 68
 */
public class CheckContainerUniqueContainedNames extends IdlmmConstraint
{
   @Override
   public IStatus validate (IValidationContext ctx)
   {
      Container target = (Container) ctx.getTarget ();

      List<Contained> contents = null;

      if (target instanceof ModuleDef)
      {
         contents = getContents ((ModuleDef) target);
      }
      else if (target instanceof InterfaceDef)
      {
         contents = getContents ((InterfaceDef) target);
      }
      else
      {
         contents = getContents (target);
      }

      Map<String, Contained> containedSet = new HashMap<String, Contained> ();
      List<Contained> duplicates = new ArrayList<Contained> ();

      /*
       * Find duplicate names
       */
      for (Contained contained : contents)
      {
         String name = contained.getName ();

         if (name == null || name.length () == 0)
         {
            continue;
         }

         if (containedSet.containsKey (name))
         {
            Contained resolution = resolveConflict (contained, containedSet.get (name));
            if (resolution != null)
            {
               containedSet.put (name, resolution);
            }
            else
            {
               duplicates.add (contained);
            }
         }
         else
         {
            containedSet.put (name, contained);
         }

         if (contained instanceof EnumDef)
         {
            for (EnumMember member : EnumDef.class.cast (contained).getMembers ())
            {
               String memberName = member.getName ();
               if (memberName == null || memberName.length () == 0)
               {
                  continue;
               }
               if (containedSet.containsKey (memberName))
               {
                  duplicates.add (member);
               }
               else
               {
                  containedSet.put (memberName, member);
               }
            }
         }
      }

      if (duplicates.size () == 0)
      {
         return ctx.createSuccessStatus ();
      }
      Object[] args = new Object[]
      {
            duplicates.get (0).getAbsoluteName (), describe (target)
      };

      ctx.addResults (duplicates);

      return ctx.createFailureStatus (args);
   }

   /**
    * Checks is it valid to have two Contained entities with same name within
    * same container.
    * 
    * @param c1
    * @param c2
    * @return <code>{@link Contained}</code> which "dominates" over the second
    *         one. It means that for pair (Forward declaration, Definition) the
    *         Definition will be returned. null it <code>c1</code> and
    *         <code>c2</code> can't be resided within one container.
    */
   protected Contained resolveConflict (Contained c1, Contained c2)
   {
      assert c1.getAbsoluteName ().equals (c2.getAbsoluteName ());

      if (c1 instanceof InterfaceDecl && c2 instanceof InterfaceDecl)
      {
         return c1;
      }
      else if (c1 instanceof InterfaceDecl && c2 instanceof InterfaceDef)
      {
         return c2;
      }
      else if (c1 instanceof InterfaceDef && c2 instanceof InterfaceDecl)
      {
         return c1;
      }
      else if (c1 instanceof ValueDecl && c2 instanceof ValueDecl)
      {
         return c1;
      }
      else if (c1 instanceof ValueDecl && c2 instanceof ValueDef)
      {
         return c2;
      }
      else if (c1 instanceof ValueDef && c2 instanceof ValueDecl)
      {
         return c1;
      }

      return null;
   }

   /**
    * Returns all the {@link Contained} ({@link Named}) objects that should be
    * considered for uniqueness, in the context of a {@link ModuleDef}.
    * <p>
    * This returns the union of all {@link Contained} objects from the specified
    * <code>module</code>, <em>and</em> any other {@link ModuleDef}
    * instances that represent the same module. Such {@link ModuleDef} objects
    * may be found in the same {@link Repository} and/or in other
    * {@link Repository#getIncludes() included} repositories.
    */
   protected List<Contained> getContents (ModuleDef target)
   {
      return target.getContents (); // TODO(ejp) Implement as described
   }

   /**
    * Returns all the {@link Contained} ({@link Named}) objects that should be
    * considered for uniqueness, in the context of a {@link InterfaceDef}.
    * <p>
    * This returns the union of all the {@link Contained} objects from the
    * specified {@link InterfaceDef}, <em>and</em> any
    * {@link InterfaceDef#getBaseInterfaces() inherited} {@link InterfaceDef}s.
    */
   protected List<Contained> getContents (InterfaceDef target)
   {
      return target.getAllContents ();
   }

   protected List<Contained> getContents (Container target)
   {
      return target.getContents ();
   }
}
