/*
 * COPYRIGHT NOTICE The information contained herein and which follows is
 * proprietary information and is the property of PrismTech Corporation. This
 * information is not to be divulged, released, copied, reproduced, or conveyed
 * to unauthorized personnel, companies, or other institutions without the
 * direct and expressed approval in writing of PrismTech Corporation. DISCLAIMER
 * Neither PrismTech Limited nor PrismTech Corporation ("PrismTech") are under
 * any obligation to support the use of this code which is provided without
 * warranty of any kind. PrismTech shall (except as otherwise prohibited by law)
 * have no liability whatsoever in relation to its use.
 */
package com.prismtech.idt.idlmm.constraints;

import org.eclipse.core.runtime.IStatus;
import org.eclipse.emf.validation.IValidationContext;

import com.prismtech.idt.idlmm.ConstantDef;
import com.prismtech.idt.idlmm.EnumDef;
import com.prismtech.idt.idlmm.EnumMember;
import com.prismtech.idt.idlmm.Expression;
import com.prismtech.idt.idlmm.IDLType;


/**
 * An enum constant can only be defined using a scoped name for the enumerator.
 * The scoped name is resolved using the normal scope resolution rules Section
 * 3.20, "Names and Scoping" on page 3-67. For example: enum Color { red, green,
 * blue }; const Color FAVORITE_COLOR = red; module M { enum Size { small,
 * medium, large }; }; const M::Size MYSIZE = M::medium; The constant name for
 * the RHS of an enumerated constant definition must denote one of the
 * enumerators defined for the enumerated type of the constant. For example:
 * const Color col = red; // is OK but const Color another = M::medium; // is an
 * error
 * 
 * @see <a
 *      href="https://scarab.prismtech.com/twiki/bin/view/MDD/IdtModelChecker">IDT-ModelConstraints</a>
 * @constrain 50
 */
public class CheckEnumConstant extends IdlmmConstraint
{

   /*
    * (non-Javadoc)
    * 
    * @see org.eclipse.emf.validation.AbstractModelConstraint#validate(org.eclipse.emf.validation.IValidationContext)
    */
   @Override
   public IStatus validate (IValidationContext ctx)
   {
      ConstantDef target = ConstantDef.class.cast (ctx.getTarget ());
      IDLType type = target.getType ();

      if (type instanceof EnumDef)
      {
         EnumDef enumDef = EnumDef.class.cast (type);
         boolean found = false;
         Expression value = target.getValue ();
         if (value == null) 
         {
            return ctx.createSuccessStatus ();
         }
         for (EnumMember member : enumDef.getMembers ())
         {
            if(member.equals (value.getExpressionValue ()))
            {
               found = true;
               break;
            }
         }
         if (!found)
         {
            return ctx.createFailureStatus (new Object[]
            {
                  describe (target), enumDef.getName ()
            });
         }

      }
      return ctx.createSuccessStatus ();

   }
}
