package com.prismtech.idt.idlmm.constraints;

import org.eclipse.core.runtime.IStatus;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.validation.IValidationContext;

import com.prismtech.idt.idlmm.Contained;
import com.prismtech.idt.idlmm.HomeDef;
import com.prismtech.idt.idlmm.IDLType;
import com.prismtech.idt.idlmm.InterfaceDef;
import com.prismtech.idt.idlmm.SequenceDef;
import com.prismtech.idt.idlmm.StructDef;
import com.prismtech.idt.idlmm.StructMember;
import com.prismtech.idt.idlmm.ValueDef;
import com.prismtech.idt.idlmm.ValueMemberDef;
import com.prismtech.idt.idlmm.Visibility;

/**
 * A primary key type must be a concrete type with at least one public state member.
 * A primary key type may not contain private state members.
 * A primary key type may not contain any members whose type is a CORBA interface reference type, including
 *   references for interfaces, abstract interfaces, and local interfaces.
 * These constraints apply recursively to the types of all of the members; that is, members that are structs, unions, value
 *   types, sequences or arrays may not contain interface reference types. If the type of a member is a value type or contains
 *   a value type, it must meet all of the above constraints.
 *   
 *   CCM Spec 6.7.2.1
 * 
 * @see <a href="https://scarab.prismtech.com/twiki/bin/view/MDD/IdtModelChecker">IDT-ModelConstraints</a>
 * 
 * @constraint 70, 71, 72
 */

public class CheckHomePrimaryKeyMembers extends IdlmmConstraint
{

   @Override
   public IStatus validate (IValidationContext ctx)
   {
      HomeDef target = (HomeDef) ctx.getTarget ();
      ValueDef primaryKey = target.getPrimarykey ();
      return validateValue (ctx, primaryKey);
   }
   
   public IStatus validateValue(IValidationContext ctx, ValueDef val)
   {
      if(val.isAbstract () == true)
      {
         Object[] args = new Object[]
         {
               val.getAbsoluteName (), "is abstract"
         };
   
         return ctx.createFailureStatus (args);
      }
      EList<ValueMemberDef> members = val.getMembers ();
      if(members.isEmpty ())
      {
         Object[] args = new Object[]
         {
               val.getAbsoluteName (), "has no public member"
         };
   
         return ctx.createFailureStatus (args);
      }
      for(ValueMemberDef member : members)
      {
         if(member.getAccess () == Visibility.PRIVATE_MEMBER)
         {
            Object[] args = new Object[]
            {
                  val.getAbsoluteName (), "has a private member : " + member.getName ()
            };
      
            return ctx.createFailureStatus (args);
         }
         IStatus status = validateMemberType (ctx, val, member.getType ());
         if(!status.isOK ())
         {
            return status;
         }
      }

      return ctx.createSuccessStatus ();
   }
   
   public IStatus validateMemberType(IValidationContext ctx, ValueDef val, IDLType type)
   {
      if(type instanceof InterfaceDef)
      {
         Object[] args = new Object[]
         {
               val.getAbsoluteName (), "has a member whose type is an interface"
         };
   
         return ctx.createFailureStatus (args);            
      }

      if(type instanceof ValueDef)
      {
         IStatus status = validateValue (ctx, (ValueDef)type);
         if(!status.isOK ())
         {
            return status;
         }
      }
      
      if(type instanceof StructDef)
      {
         for(StructMember structMember : ((StructDef)type).getMembers ())
         {
            IStatus status = validateMemberType (ctx, val, structMember.getType ());
            if(!status.isOK ())
            {
               return status;
            }
         }
      }
      
      if(type instanceof SequenceDef)
      {     
         IStatus status = validateMemberType (ctx, val, ((SequenceDef)type).getOwnedType ());
         if(!status.isOK ())
         {
            return status;
         }
      }
      
      return ctx.createSuccessStatus ();
   }
   

}
