package com.prismtech.idt.idlmm.impl;

import java.util.List;

import org.eclipse.emf.ecore.EAnnotation;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EOperation;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;

import com.prismtech.idt.idlmm.Contained;
import com.prismtech.idt.idlmm.Container;
import com.prismtech.idt.idlmm.IdlmmPackage;


/**
 * This is an 'extension class' for {@link Contained}.
 * <p>
 * It provides implementations for the operation signatures defined in the UML
 * model. In the process of transforming from the UML2 model to the ECORE model,
 * the resulting {@link EOperation} objects, in the ECORE, have an
 * {@link EAnnotation} associated with them. The value of these annotations gets
 * used as the body of the generated methods that implement the interface
 * signatures. This annotation value is constructed such that it makes a simple
 * call to the corresponding method on the corresponding 'extension class'.
 * 
 * @author ejp
 */
public class ContainedExt
{
   /**
    * @see Contained#getAbsoluteName()
    */
   public static String getAbsoluteName (Contained self)
   {
      String name = self.getName ();

      if (name == null)
      {
         name = "<noname>";
      }

      if (! (self.getContainer () instanceof Contained))
      {
         return "::" + name;
      }

      return getAbsoluteName ((Contained) self.getContainer ()) + "::" + name;
   }

   @SuppressWarnings("unchecked")
   public static void setContainer (Contained self, Container container)
   {
      if (container == null)
      {
         EObject eContainer = self.eContainer ();

         if (eContainer != null)
         {
            EStructuralFeature eContainingFeature = self.eContainingFeature ();
            Object value = eContainer.eGet (eContainingFeature);
            List<Contained> contents = (List<Contained>) value;

            contents.remove (self);
         }
      }
      else
      {
         EReference eReference = getContainmentReference (container, self);
         Object value = container.eGet (eReference);
         List<Contained> contents = (List<Contained>) value;

         contents.add (self);
      }
   }

   /**
    * Based on the actual {@link EObject#eClass() types} of the specified
    * {@link Container} and {@link Contained} objects, this method will return
    * the appropriate {@link EReference} from the <code>container</code>
    * object that is able to contain the <code>contained</code> object.
    * 
    * @exception IllegalArgumentException
    *               If there is either zero, or more than one, suitable
    *               containment references found.
    */
   private static EReference getContainmentReference (Container container, Contained contained)
   {
      EReference result = null;

      EClass containerType = container.eClass ();
      EClass containedType = contained.eClass ();

      for (EReference candidate : containerType.getEAllContainments ())
      {
         EClass eReferenceType = (EClass) candidate.getEType ();

         if (eReferenceType.isSuperTypeOf (containedType) == true)
         {
            if (result == null)
            {
               result = candidate;
            }
            else
            {
               throw new IllegalArgumentException (containedType.getName () + ".setContainer (" + containerType.getName ()
                     + ") : Multiple candidate containment references");
            }
         }
      }

      if (result == null)
      {
         throw new IllegalArgumentException (containedType.getName () + ".setContainer (" + containerType.getName ()
               + ") : No candidate containment references");
      }

      return result;
   }

   // public static String getID (Contained self, String id)
   // {
   // if (self.eIsSet (IdlmmPackage.Literals.CONTAINED__ID))
   // {
   // return id;
   // }
   // else
   // {
   // StringBuilder result = new StringBuilder ("IDL:");
   //
   // String prefix = self.getPrefix ();
   //
   // if ( (prefix != null) && (prefix.length () > 0))
   // {
   // result.append (prefix);
   // result.append ('/');
   // result.append (self.getName ());
   // }
   // else
   // {
   // String absname = self.getAbsoluteName ();
   //            
   // absname = absname.replace ("::", "/");
   // absname = absname.substring (1);
   //            
   // result.append (absname);
   // }
   //
   // String version = self.getVersion ();
   //
   // if ( (version == null) || (version.length () == 0))
   // {
   // result.append (":1.0");
   // }
   // else
   // {
   // result.append (':');
   // result.append (version);
   // }
   //
   // return result.toString ();
   // }
   // }

   public static String getID (Contained self, String id)
   {
      if (self.eIsSet (IdlmmPackage.Literals.CONTAINED__ID))
      {
         return id;
      }
      else
      {
         StringBuilder result = new StringBuilder ("IDL:");

         result.append (getIDNamePart (self));
         result.append (getIDVersionPart (self));

         return result.toString ();
      }
   }

   private static String getIDNamePart (Contained self)
   {
      StringBuilder result = new StringBuilder ();
      String prefix = self.getPrefix ();

      if (prefix == null) // got no prefix
      {
         Container container = self.getContainer ();

         if (container instanceof Contained)
         {
            result.append (getIDNamePart ( ((Contained) container)));
            result.append ('/');
            result.append (self.getName ());
         }
         else
         {
            result.append (self.getName ());
         }
      }
      else if (prefix.length () == 0) // got a blank prefix
      {
         result.append (self.getName ());
      }
      else
      // got a non-blank prefix
      {
         result.append (prefix);
         result.append ('/');
         result.append (self.getName ());
      }

      return result.toString ();
   }

   private static Object getIDVersionPart (Contained self)
   {
      String version = self.getVersion ();

      if ( (version == null) || (version.length () == 0))
      {
         return ":1.0";
      }
      else
      {
         StringBuilder result = new StringBuilder (10);
         return result.append (':').append (version).toString ();
      }
   }
}
