package com.prismtech.idt.idlmm.impl;

import java.util.Set;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.ECollections;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EAnnotation;
import org.eclipse.emf.ecore.EOperation;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.impl.EObjectImpl;
import org.eclipse.emf.ecore.util.DelegatingEcoreEList.UnmodifiableEList;

import com.prismtech.idt.idlmm.Contained;
import com.prismtech.idt.idlmm.Container;
import com.prismtech.idt.idlmm.EnumDef;
import com.prismtech.idt.idlmm.IRObject;
import com.prismtech.idt.idlmm.InterfaceDecl;
import com.prismtech.idt.idlmm.InterfaceDef;
import com.prismtech.idt.idlmm.ValueDecl;
import com.prismtech.idt.idlmm.ValueDef;
import com.prismtech.idt.idlmm.util.ScopedName;


/**
 * This is an 'extension class' for {@link Container}.
 * <p>
 * It provides implementations for the operation signatures defined in the UML
 * model. In the process of transforming from the UML2 model to the ECORE model,
 * the resulting {@link EOperation} objects, in the ECORE, have an
 * {@link EAnnotation} associated with them. The value of these annotations gets
 * used as the body of the generated methods that implement the interface
 * signatures. This annotation value is constructed such that it makes a simple
 * call to the corresponding method on the corresponding 'extension class'.
 */
public class ContainerExt
{
   public static Contained lookup (Container self, String search_name, boolean global, boolean inherited)
   {
      ScopedName sname = new ScopedName (search_name);

      if (sname.isAbsolute ())
      {
         sname = sname.ltrim (1);
         return lookDown (self.getRepository (), sname, inherited);
      }
      else
      {
         return lookUp (self, sname, inherited);
      }
   }

   private static Contained lookUp (Container self, ScopedName sname, boolean inherited)
   {
      Contained contained = lookDown (self, sname, inherited);
      if (contained != null)
      {
         return contained;
      }
      else if (self instanceof Contained)
      {
         Container parent = ((Contained) self).getContainer ();
         if (parent == null)
         {
            return null;
         }
         return lookUp (parent, sname, inherited);
      }
      else
      {
         return null;
      }
   }

   private static Contained lookDown (Container self, ScopedName sname, boolean inherited)
   {
      Contained found = null;
      EList<Contained> contents;
      if (self instanceof InterfaceDef && inherited)
      {
         contents = ((InterfaceDef) self).getAllContents ();
      }
      else if (self instanceof ValueDef && inherited)
      {
         contents = ((ValueDef) self).getAllContents ();
      }
      else
      {
         contents = self.getContents ();
      }
      for (Contained c : contents)
      {
         if (c.getName ().equals (sname.get (0)))
         {
            if (sname.size () == 1)
            {
               if (c instanceof InterfaceDecl || c instanceof ValueDecl)
               {
                  found = c;
               }
               else
               {
                  return c;
               }
            }
            else
            {
               if (c instanceof InterfaceDecl || c instanceof ValueDecl)
               {
                  continue;
               }
               else
               {
                  return lookDown ((Container) c, sname.ltrim (1), inherited);
               }
            }
         }
         else if (c instanceof EnumDef)
         {
            Contained enumMember = lookDown ((Container) c, sname, inherited);
            if (enumMember != null)
            {
               return enumMember;
            }
         }

      }
      return found;
   }

   /**
    * Implementation of {@link Container#getContents()}.
    */
   public static EList<Contained> getContents (Container self)
   {
      return getContents (self, Contained.class);
   }

   public static <T> EList<T> getContents (Container self, Class<T> type)
   {
      return new UnmodifiableEList<T> (
            (EObjectImpl)self, 
            (EStructuralFeature)IdlmmPackageImpl.init ().getContainer_Contents (), 
            select (self.eContents (), type));
   }

   public static <T> EList<T> select (EList<?> source, Class<T> type)
   {
      EList<T> result = new BasicEList<T> ();

      for (Object o : source)
      {
         if (type.isInstance (o))
         {
            result.add (type.cast (o));
         }
      }

      return ECollections.unmodifiableEList (result);
   }

   /**
    * @see Container#lookup(String)
    */
   public static Contained lookup (Container self, String name)
   {
      return self.lookup (name, true, true);
   }

   public static EList<IRObject> getPrerequisites (Container self)
   {
      Set<IRObject> result = new ArraySet<IRObject> ();
      for (Contained contained : self.getContents ())
      {
         result.addAll (contained.getPrerequisites ());
      }
      return new BasicEList<IRObject> (result);
   }
}
