package com.prismtech.idt.idlmm.impl;

import java.math.BigInteger;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EAnnotation;
import org.eclipse.emf.ecore.EOperation;

import com.prismtech.idt.idlmm.AliasDef;
import com.prismtech.idt.idlmm.ArrayDef;
import com.prismtech.idt.idlmm.ConstantDef;
import com.prismtech.idt.idlmm.EnumDef;
import com.prismtech.idt.idlmm.ExceptionDef;
import com.prismtech.idt.idlmm.FixedDef;
import com.prismtech.idt.idlmm.IDLType;
import com.prismtech.idt.idlmm.IdlmmFactory;
import com.prismtech.idt.idlmm.IntegerLiteral;
import com.prismtech.idt.idlmm.InterfaceDef;
import com.prismtech.idt.idlmm.ModuleDef;
import com.prismtech.idt.idlmm.NativeDef;
import com.prismtech.idt.idlmm.PrimitiveDef;
import com.prismtech.idt.idlmm.PrimitiveKind;
import com.prismtech.idt.idlmm.Repository;
import com.prismtech.idt.idlmm.SequenceDef;
import com.prismtech.idt.idlmm.StringDef;
import com.prismtech.idt.idlmm.StructDef;
import com.prismtech.idt.idlmm.UnionDef;
import com.prismtech.idt.idlmm.ValueBoxDef;
import com.prismtech.idt.idlmm.ValueDef;
import com.prismtech.idt.idlmm.WStringDef;


/**
 * This is an 'extension class' for {@link Repository}.
 * <p>
 * It provides implementations for the operation signatures defined in the UML
 * model. In the process of transforming from the UML2 model to the ECORE model,
 * the resulting {@link EOperation} objects, in the ECORE, have an
 * {@link EAnnotation} associated with them. The value of these annotations gets
 * used as the body of the generated methods that implement the interface
 * signatures. This annotation value is constructed such that it makes a simple
 * call to the corresponding method on the corresponding 'extension class'.
 * 
 * @author ejp
 */
public class RepositoryExt
{
   private static IdlmmFactory getFactory (Repository self)
   {
      return (IdlmmFactory) self.eClass ().getEPackage ().getEFactoryInstance ();
   }

   private static IntegerLiteral createIntegerLiteral (Repository self, int value)
   {
      IntegerLiteral literal = getFactory (self).createIntegerLiteral ();
      literal.setExpressionValue (BigInteger.valueOf (value));
      return literal;
   }

   /**
    * Implementation of {@link Repository#createString(int)}
    */
   public static StringDef createString (Repository self, int bound)
   {
      StringDef result = getFactory (self).createStringDef ();

      result.setBound (createIntegerLiteral (self, bound));

      return result;
   }

   /**
    * Implementation of {@link Repository#createWString(int)}
    */
   public static WStringDef createWString (Repository self, int bound)
   {
      WStringDef result = getFactory (self).createWStringDef ();

      result.setBound (createIntegerLiteral (self, bound));

      return result;
   }

   /**
    * Implementation of {@link Repository#createSequence(int, IDLType)}
    */
   @SuppressWarnings("unchecked")
   public static SequenceDef createSequence (Repository self, int bound, IDLType elementType)
   {
      SequenceDef result = getFactory (self).createSequenceDef ();

      result.setBound (createIntegerLiteral (self, bound));
      result.setType (elementType);

      return result;
   }

   /**
    * Implementation of {@link Repository#createArray(int, IDLType)}
    */
   @SuppressWarnings("unchecked")
   public static ArrayDef createArray (Repository self, int length, IDLType elementType)
   {
      ArrayDef result = getFactory (self).createArrayDef ();

      result.setLength (createIntegerLiteral (self, length));
      result.setType (elementType);

      return result;
   }

   /**
    * Implementation of {@link Repository#createFixed(int, int)}
    */
   @SuppressWarnings("unchecked")
   public static FixedDef createFixed (Repository self, int digits, int scale)
   {
      FixedDef result = getFactory (self).createFixedDef ();

      result.setDigits (createIntegerLiteral (self, digits));
      result.setScale (createIntegerLiteral (self, scale));
      
      return result;
   }

   /**
    * Implementation of {@link Repository#createPrimitive(PrimitiveKind)}
    */
   public static PrimitiveDef createPrimitive (Repository self, PrimitiveKind kind)
   {
      PrimitiveDef result = getFactory (self).createPrimitiveDef ();

      result.setKind (kind);

      return result;
   }

   public static EList<AliasDef> getAliases (Repository self)
   {
      return ContainerExt.getContents (self, AliasDef.class);
   }

   public static EList<ConstantDef> getConstants (Repository self)
   {
      return ContainerExt.getContents (self, ConstantDef.class);
   }

   public static EList<EnumDef> getEnums (Repository self)
   {
      return ContainerExt.getContents (self, EnumDef.class);
   }

   public static EList<ExceptionDef> getExceptions (Repository self)
   {
      return ContainerExt.getContents (self, ExceptionDef.class);
   }

   public static EList<InterfaceDef> getInterfaces (Repository self)
   {
      return ContainerExt.getContents (self, InterfaceDef.class);
   }

   public static EList<ModuleDef> getModules (Repository self)
   {
      return ContainerExt.getContents (self, ModuleDef.class);
   }

   public static EList<NativeDef> getNatives (Repository self)
   {
      return ContainerExt.getContents (self, NativeDef.class);
   }

   public static EList<StructDef> getStructs (Repository self)
   {
      return ContainerExt.getContents (self, StructDef.class);
   }

   public static EList<UnionDef> getUnions (Repository self)
   {
      return ContainerExt.getContents (self, UnionDef.class);
   }

   public static EList<ValueBoxDef> getValueBoxes (Repository self)
   {
      return ContainerExt.getContents (self, ValueBoxDef.class);
   }

   public static EList<ValueDef> getValueTypes (Repository self)
   {
      return ContainerExt.getContents (self, ValueDef.class);
   }

   /**
    * Helper for {@link #getAllIncludes(Repository, boolean)}
    */
   private static EList<Repository> getAllIncludes (Repository self, EList<Repository> includes, boolean inclusive)
   {
      for (Repository include : self.getIncludes ())
      {
         if (includes.contains (include) == false)
         {
            getAllIncludes (include, includes, true);
         }
      }

      if (inclusive)
      {
         includes.add (self);
      }

      return includes;
   }

   /**
    * @see Repository#getAllIncludes(boolean)
    */
   public static EList<Repository> getAllIncludes (Repository self, boolean inclusive)
   {
      return getAllIncludes (self, new BasicEList<Repository> (), inclusive);
   }
}
