/***************************************************************************************************
 * - COPYRIGHT NOTICE - * * The information contained herein and which follows is proprietary *
 * information and is the property of PrismTech Corporation. This * information is not to be
 * divulged, released, copied, reproduced, or * conveyed to unauthorized personnel, companies, or
 * other institutions * without the direct and expressed approval in writing of PrismTech *
 * Corporation. * * * - DISCLAIMER - * Neither PrismTech Limited nor PrismTech Corporation
 * ("PrismTech") * are under any obligation to support the use of this code which is * provided
 * without warranty of any kind. PrismTech shall (except as * otherwise prohibited by law) have no
 * liability whatsoever in relation * to its use. * *
 **************************************************************************************************/

package com.prismtech.idt.imports.idl.driver;

import java.io.File;

import org.eclipse.core.runtime.IStatus;

import com.prismtech.idt.idlmm.Repository;
import com.prismtech.idt.imports.idl.importer.IIdlImporter;
import com.prismtech.idt.imports.idl.parser.IIdlParser;
import com.prismtech.idt.imports.idl.preprocessor.IIdlPreprocessor;


/**
 * Driver is a facade hiding and managing IDL preprocessing, parsing, importing and validation
 * functionality. Driver provides interfaces which can be used both by command-line and GUI clients.
 * 
 * @see IdlImportDriverFactory
 * @see IIdlPreprocessor
 * @see IIdlParser
 * @see IIdlImporter
 */
public interface IIdlImportDriver
{
   public static final String PLUGIN_ID = IIdlImportDriver.class.getPackage ().getName ();

   /**
    * Factory method that creates a new {@link IIdlImportDriverParams} instance. The returned object
    * encapsulates all of the input and output involved in an invocation of
    * {@link #build(IIdlImportDriverParams)}.
    * <p>
    * Note that the returned object can only be used as input to the {@link IIdlImportDriver} that
    * was used to create it.
    */
   public IIdlImportDriverParams createDriverParams (File... files);

   /**
    * Returns the {@link IIdlPreprocessor} instance used by this {@link IIdlImportDriver}
    */
   public IIdlPreprocessor getPreprocessor ();

   /**
    * Returns the {@link IIdlParser} instance used by this {@link IIdlImportDriver}
    */
   public IIdlParser getParser ();

   /**
    * Returns the {@link IIdlImporter} instance used by this {@link IIdlImportDriver}
    */
   public IIdlImporter getImporter ();

   /**
    * Set the {@link IValidator} that will be used by the {@link #build(IIdlImportDriverParams)} to
    * perform the validation of {@link Repository} objects produced by the {@link IIdlImporter}
    */
   public void setValidator (IValidator validator);

   /**
    * Performs the build process :
    * <ul>
    * <li>Preprocess the source IDL</li>
    * <li>Parse the preprocessed IDL</li>
    * <li>Import the information into a {@link Repository}</li>
    * <li>Validate the resulting {@link Repository}</li>
    * </ul>
    * The result of the build process is not represented by a single {@link IStatus} because it is
    * possible for some IDL files to be processed correctly, while processing others may encounter
    * errors or warnings. Therefore, status information is 'returned' separately for each IDL file
    * via the {@link IIdlImportDriverParam#getStatus()} method.
    * 
    * @param params
    *           Describes each of the IDL source files to be processed, along with any options for
    *           preprocess, parse, import, validate stages.
    * 
    * @throws IllegalArgumentException
    *            If the specified {@link IIdlImportDriverParams} instance was not created by this
    *            {@link IIdlImportDriver} instance.
    */
   public void build (IIdlImportDriverParams params);
}
