/***************************************************************************************************
 * - COPYRIGHT NOTICE - * * The information contained herein and which follows is proprietary *
 * information and is the property of PrismTech Corporation. This * information is not to be
 * divulged, released, copied, reproduced, or * conveyed to unauthorized personnel, companies, or
 * other institutions * without the direct and expressed approval in writing of PrismTech *
 * Corporation. * * * - DISCLAIMER - * Neither PrismTech Limited nor PrismTech Corporation
 * ("PrismTech") * are under any obligation to support the use of this code which is * provided
 * without warranty of any kind. PrismTech shall (except as * otherwise prohibited by law) have no
 * liability whatsoever in relation * to its use. * *
 **************************************************************************************************/
package com.prismtech.idt.imports.idl.internal.driver;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.StringWriter;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.emf.common.util.BasicDiagnostic;
import org.eclipse.emf.ecore.util.Diagnostician;

import com.prismtech.idt.idlmm.Repository;
import com.prismtech.idt.imports.idl.driver.IIdlImportDriver;
import com.prismtech.idt.imports.idl.driver.IIdlImportDriverParam;
import com.prismtech.idt.imports.idl.driver.IIdlImportDriverParams;
import com.prismtech.idt.imports.idl.driver.IValidator;
import com.prismtech.idt.imports.idl.importer.IIdlImporter;
import com.prismtech.idt.imports.idl.importer.IIdlImporterParams;
import com.prismtech.idt.imports.idl.importer.IdlImporterFactory;
import com.prismtech.idt.imports.idl.parser.IIdlParser;
import com.prismtech.idt.imports.idl.parser.IIdlParserParams;
import com.prismtech.idt.imports.idl.parser.IIdlParserStatus;
import com.prismtech.idt.imports.idl.parser.IdlParserAST;
import com.prismtech.idt.imports.idl.parser.IdlParserFactory;
import com.prismtech.idt.imports.idl.preprocessor.IIdlPreprocessor;
import com.prismtech.idt.imports.idl.preprocessor.IIdlPreprocessorParams;
import com.prismtech.idt.imports.idl.preprocessor.IIdlPreprocessorStatus;
import com.prismtech.idt.imports.idl.preprocessor.IdlPreprocessorFactory;
import com.prismtech.trace.Trace;


public class IdlImportDriver implements IIdlImportDriver
{
   private static final boolean DEBUG = false;

   private IIdlPreprocessor m_preprocessor;

   private IIdlParser m_parser;

   private IIdlImporter m_importer;

   private IValidator m_validator;

   public IdlImportDriver ()
   {
      m_preprocessor = IdlPreprocessorFactory.create ();
      m_parser = IdlParserFactory.create ();
      m_importer = IdlImporterFactory.create ();
   }

   public IIdlImportDriverParams createDriverParams (File... files)
   {
      IdlImportDriverParams result = new IdlImportDriverParams (this);

      for (File file : files)
      {
         result.add (file);
      }

      return result;
   }

   public IIdlPreprocessor getPreprocessor ()
   {
      return m_preprocessor;
   }

   public IIdlParser getParser ()
   {
      return m_parser;
   }

   public IIdlImporter getImporter ()
   {
      return m_importer;
   }

   public void setValidator (IValidator validator)
   {
      m_validator = validator;
   }

   public void build (IIdlImportDriverParams params)
   {
      if (DEBUG)
      {
         Trace.begin (this, ".build (", params, ")");
      }

      try
      {
         if ( (params instanceof IdlImportDriverParams) == false)
         {
            throw new IllegalArgumentException ();
         }
         if ( ((IdlImportDriverParams) params).getDriver () != this)
         {
            throw new IllegalArgumentException ("Wrong driver");
         }

         for (IIdlImportDriverParam param : params)
         {
            build ((IdlImportDriverParam) param);
         }
      }
      finally
      {
         if (DEBUG)
         {
            Trace.end ();
         }
      }
   }

   private void build (IdlImportDriverParam param)
   {
      if (DEBUG)
      {
         Trace.begin (this, ".build (", param, ")");
      }

      try
      {
         preprocessIdl (param);
         
         if (param.getPreprocessorOutput () == null)
         {
            return;
         }

         parseIdl (param);

         if (param.getAST () == null)
         {
            return;
         }

         importIdl (param);

         if (param.getRepository () == null)
         {
            return;
         }

         validateIdl (param);
      }
      finally
      {
         if (DEBUG)
         {
            Trace.end ();
         }
      }
   }

   private IStatus preprocessIdl (IdlImportDriverParam file)
   {
      if (DEBUG)
      {
         Trace.begin (this, ".preprocessIdl (", file, ")");
      }

      try
      {
         IIdlPreprocessorParams preprocessorParams = file.getPreprocessorParams ();
         File input = file.getIdlFile ();
         StringWriter output = new StringWriter ();
         Map<File, List<File>> dependencies = new HashMap<File, List<File>> ();
         IIdlPreprocessorStatus pstatus = m_preprocessor.preprocess (preprocessorParams, input, output, dependencies);
         IStatus status = new IdlPreprocessorStatusAdapter(pstatus);
         file.addStatus (status);
         file.setPreprocessorOutput (output.toString ());
         file.setDependencies (dependencies);
         return status;
      }
      finally
      {
         if (DEBUG)
         {
            Trace.end ();
         }
      }
   }

   private IStatus parseIdl (IdlImportDriverParam file)
   {
      if (DEBUG)
      {
         Trace.begin (this, ".parseIdl (", file, ")");
      }

      try
      {
         IIdlParserParams parseParams = file.getParserParams ();
         String preprocessorOutput = file.getPreprocessorOutput ();
         ByteArrayInputStream input = new ByteArrayInputStream (preprocessorOutput.getBytes ());
         IIdlParserStatus pstatus = m_parser.parse (parseParams, input);
         IStatus status = new IdlParserStatusAdapter(pstatus);
         file.addStatus (status);
         file.setAST (m_parser.getAST ());
         return status;
      }
      finally
      {
         if (DEBUG)
         {
            Trace.end ();
         }
      }
   }

   private IStatus importIdl (IdlImportDriverParam file)
   {
      if (DEBUG)
      {
         Trace.begin (this, ".importIdl (", file, ")");
      }

      try
      {
         IIdlImporterParams importParams = file.getImporterParams ();
         IdlParserAST ast = file.getAST ();
         IStatus status = m_importer.build (importParams, ast);
         file.addStatus (status);
         file.setRepository (m_importer.getRepository ());
         return status;
      }
      finally
      {
         if (DEBUG)
         {
            Trace.end ();
         }
      }
   }

   private IStatus validateIdl (IdlImportDriverParam file)
   {
      if (DEBUG)
      {
         Trace.begin (this, ".validateIdl (", file, ")");
      }

      try
      {
         Repository repository = file.getRepository ();

         /*
          * Run basic EMF validation
          */
         BasicDiagnostic diagnostics = new BasicDiagnostic ();

         if (Diagnostician.INSTANCE.validate (repository, diagnostics) == false)
         {
            file.addStatus (BasicDiagnostic.toIStatus (diagnostics));
         }

         /*
          * Run any IValidator that has been plugged in
          */
         if (m_validator == null)
         {
            file.addStatus (new Status (IStatus.WARNING, IIdlImportDriver.PLUGIN_ID,
                  "Repository has not been validated because the driver has no validator"));
         }
         else
         {
            file.addStatus (m_validator.validate (file.getRepository ()));
         }

         return file.getStatus ();
      }
      finally
      {
         if (DEBUG)
         {
            Trace.end ();
         }
      }
   }

   @Override
   public String toString ()
   {
      return getClass ().getSimpleName ();
   }
}
