package com.prismtech.idt.imports.idl.internal.importer;

import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;

import com.prismtech.idt.idlmm.Contained;
import com.prismtech.idt.idlmm.IdlmmFactory;
import com.prismtech.idt.idlmm.IdlmmPackage;
import com.prismtech.idt.imports.idl.importer.IIdlPragmaHandler;
import com.prismtech.idt.imports.idl.importer.IdlImporterPlugin;
import com.prismtech.idt.imports.idl.internal.importer.builderutil.RepositoryIDHelper;


public class IDPragmaHandler implements IIdlPragmaHandler
{
   public boolean matches (String pragmaText)
   {
      return pragmaText.matches ("ID[ \t]+.*");
   }

   public IStatus create (CreateContext ctx)
   {
      String text = ctx.getPragmaText ();

      assert (matches (ctx.getPragmaText ()));

      /*
       * Remove "ID" and whitespace from the start
       */
      text = text.substring (2).trim ();

      /*
       * Find the first '"'
       */
      int firstQuote = text.indexOf ('"');

      /*
       * If we didn't find a quote something's wrong
       */
      if (firstQuote == -1)
      {
         return error (ctx, "Malformed #pragma ID <name> \"<id>\" : '#pragma " + ctx.getPragmaText () + "'");
      }
      
      /*
       * Get the name bit preceding the first quote, and trim off whitespace
       */
      String name = text.substring (0, firstQuote).trim ();

      /*
       * Find the last '"'
       */
      int lastQuote = text.lastIndexOf ('"');

      /*
       * If we didn't find a quote something's wrong
       */
      if (lastQuote == -1)
      {
         return error (ctx, "Malformed #pragma ID <name> \"<id>\" : '#pragma " + ctx.getPragmaText () + "'");
      }
      
      /*
       * Get the ID bit between the first and last quotes
       */
      String id = text.substring (firstQuote + 1, lastQuote);

      /*
       * Check we got the name bit
       */
      if ( (name == null) || (name.length () == 0))
      {
         return error (ctx, "Failed to extract the name part of : '#pragma " + ctx.getPragmaText () + "'");
      }

      /*
       * Check we got the id bit
       */
      if ( (id == null) || (id.length () == 0))
      {
         return error (ctx, "Failed to extract the id part of : '#pragma " + ctx.getPragmaText () + "'");
      }
      
      /*
       * Check the id is valid
       */
      if(!RepositoryIDHelper.validRepositoryID (id))
      {
         return error(ctx, "Invalid repository id " + id);
      }

      /*
       * Resolve the name to find the object to set the id on
       */
      Contained target = ctx.getParent ().lookup (name);

      if (target == null)
      {
         return error (ctx, "Failed to resolve '" + name + "' from : '#pragma " + ctx.getPragmaText () + "'");
      }

      /*
       * Check if taget already has id explicitly set
       */
      if (target.eIsSet (IdlmmPackage.Literals.CONTAINED__ID))
      {
         if (id.equals (target.getID ()) == false)
         {
            return error (ctx, "Cannot set repository id on '" + name + "' to '" + id + "'. It's already set to '"
                  + target.getID () + "'");
         }
      }

      /*
       * Everything is ok so set the id on the target and create the ScopePragma
       */
      target.setID (id);

      ctx.setPragma (IdlmmFactory.eINSTANCE.createScopePragma ());

      return new Status (IStatus.OK, IdlImporterPlugin.PLUGIN_ID, "");
   }

   private IStatus error (CreateContext ctx, String message)
   {
      return new IdlImporterStatus (ctx.getAst (), IStatus.ERROR, message);
   }

   public IStatus bind (CreateContext ctx)
   {
      return new Status (IStatus.OK, IdlImporterPlugin.PLUGIN_ID, "");
   }
   
   public static void main (String[] args)
   {
      IDPragmaHandler h = new IDPragmaHandler ();
      
      System.out.println (h.matches ("ID wefwfww"));
      System.out.println (h.matches ("ID  wefwfww"));
      System.out.println (h.matches ("ID\twefwfww"));
   }
}
