/*******************************************************************************
 * - COPYRIGHT NOTICE - * * The information contained herein and which follows
 * is proprietary * information and is the property of PrismTech Corporation.
 * This * information is not to be divulged, released, copied, reproduced, or *
 * conveyed to unauthorized personnel, companies, or other institutions *
 * without the direct and expressed approval in writing of PrismTech *
 * Corporation. * * * - DISCLAIMER - * Neither PrismTech Limited nor PrismTech
 * Corporation ("PrismTech") * are under any obligation to support the use of
 * this code which is * provided without warranty of any kind. PrismTech shall
 * (except as * otherwise prohibited by law) have no liability whatsoever in
 * relation * to its use. * *
 ******************************************************************************/
package com.prismtech.idt.imports.idl.internal.importer;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.MultiStatus;
import org.eclipse.core.runtime.Status;

import com.prismtech.idt.idlmm.Repository;
import com.prismtech.idt.imports.idl.importer.IIdlImporter;
import com.prismtech.idt.imports.idl.importer.IIdlImporterParams;
import com.prismtech.idt.imports.idl.importer.IIdlPragmaHandler;
import com.prismtech.idt.imports.idl.importer.IIdlSourceComposer;
import com.prismtech.idt.imports.idl.importer.IdlImporterPlugin;
import com.prismtech.idt.imports.idl.parser.IdlParserAST;
import com.prismtech.idt.imports.idl.parser.util.ASTDumper;


/**
 * This is a wrapper for {@link IdlTreeParser} parser.
 */
public class IdlImporter implements IIdlImporter
{
   private boolean DEBUG_LEVEL = false;

   private List<IStatus> m_status = new ArrayList<IStatus> ();

   private Repository m_repository;

   private List<IIdlPragmaHandler> m_pragmaHandlers = new ArrayList<IIdlPragmaHandler> ();

   private IIdlSourceComposer m_composer = new DefaultIdlSourceComposer ();

   public IdlImporter ()
   {
      addPragmaHandler (new PrefixPragmaHandler ());
      addPragmaHandler (new VersionPragmaHandler ());
      addPragmaHandler (new IDPragmaHandler ());
   }

   public IIdlImporterParams createImporterParams ()
   {
      return new IdlImporterParams ();
   }

   public void setSourceComposer (IIdlSourceComposer composer)
   {
      m_composer = (composer == null) ? new DefaultIdlSourceComposer () : composer;
   }

   private static class PragmaHandlerWrapper implements IIdlPragmaHandler
   {
      private final IIdlPragmaHandler m_handler;

      PragmaHandlerWrapper (IIdlPragmaHandler handler)
      {
         m_handler = handler;
      }

      public IStatus bind (CreateContext ctx)
      {
         try
         {
            return m_handler.bind (ctx);
         }
         catch (Throwable e)
         {
            return error ("Caught exception from " + describe (m_handler) + ".bind ()", e);
         }
      }

      public IStatus create (CreateContext ctx)
      {
         IStatus result;
         
         try
         {
            result = m_handler.create (ctx);
            
            if (result.isOK () && (ctx.getPragma () == null))
            {
               result = error (describe (m_handler) + " didn't create a Pragma", null);
            }
         }
         catch (Throwable e)
         {
            result = error ("Caught exception from " + describe (m_handler) + ".create ()", e);
         }
         
         return result;
      }

      public boolean matches (String pragmaText)
      {
         try
         {
            return m_handler.matches (pragmaText);
         }
         catch (Throwable e)
         {
            error ("Caught exception from " + describe (m_handler) + ".matches (\""+ pragmaText + "\")", e);
            return false;
         }
      }
      
      private static String describe (IIdlPragmaHandler handler)
      {
         return handler.getClass ().getName ();
      }

      private IStatus error (String message, Throwable e)
      {
         IStatus status = new Status (IStatus.ERROR, IdlImporterPlugin.PLUGIN_ID, message, e);

         return status;
      }
   }
   /*
    * (non-Javadoc)
    * 
    * @see com.prismtech.idt.imports.idl.importer.IIdlImporter#addPragmaType(java.lang.String,
    *      java.lang.Class)
    */
   public void addPragmaHandler (IIdlPragmaHandler handler)
   {
      m_pragmaHandlers.add (new PragmaHandlerWrapper (handler));
   }

   /*
    * (non-Javadoc)
    * 
    * @see com.prismtech.idt.imports.idl.driver.IImporter#build(com.prismtech.idt.IdlParserAST
    *      com.prismtech.idt.imports.idl.driver.IDependencyResolver)
    */
   public IStatus build (IIdlImporterParams params, IdlParserAST ast)
   {
      /**
       * this code is used while debugging and should be commented in delivery
       * version
       */
      // DEBUG_LEVEL = true;
      /** end of debug code */

      if (DEBUG_LEVEL)
      {
         ASTDumper.dump (System.out, ast);
      }

      m_status.clear ();
      IdlTreeParser treeParser = new IdlTreeParser (m_composer, m_pragmaHandlers);
      this.addStatus (treeParser.parse (ast));
      m_repository = treeParser.getRootObject (Repository.class);
      if (DEBUG_LEVEL)
      {
         ASTDumper.fullDump (System.out, ast);
      }

      return this.getStatus ();
   }

   public Repository getRepository ()
   {
      return m_repository;
   }

   public IStatus getStatus ()
   {
      switch (m_status.size ())
      {
         case 0:
         {
            return new Status (IStatus.OK, IdlImporterPlugin.PLUGIN_ID, 0, "", null); //$NON-NLS-1$
         }
         case 1:
         {
            return m_status.get (0);
         }
         default:
         {
            MultiStatus status = new MultiStatus (IdlImporterPlugin.PLUGIN_ID, 0, "", null);//$NON-NLS-1$

            for (IStatus s : m_status)
            {
               status.add (s);
            }
            return status;
         }
      }
   }

   /**
    * Adds the specified {@link IStatus} to our {@link #m_status} list. If the
    * specified {@link IStatus} is a
    * {@link IStatus#isMultiStatus() multi status}, then its children are
    * added.
    */
   private void addStatus (IStatus status)
   {
      if (status.isMultiStatus ())
      {
         for (IStatus s : status.getChildren ())
         {
            this.addStatus (s);
         }
      }
      else
      {
         m_status.add (status);
      }
   }
}
