/*
 * COPYRIGHT NOTICE The information contained herein and which follows is
 * proprietary information and is the property of PrismTech Corporation. This
 * information is not to be divulged, released, copied, reproduced, or conveyed
 * to unauthorized personnel, companies, or other institutions without the
 * direct and expressed approval in writing of PrismTech Corporation. DISCLAIMER
 * Neither PrismTech Limited nor PrismTech Corporation ("PrismTech") are under
 * any obligation to support the use of this code which is provided without
 * warranty of any kind. PrismTech shall (except as otherwise prohibited by law)
 * have no liability whatsoever in relation to its use.
 */

package com.prismtech.idt.imports.idl.internal.importer.builderutil;

import org.eclipse.core.runtime.IStatus;

import com.prismtech.idt.idlmm.ComponentDef;
import com.prismtech.idt.idlmm.Contained;
import com.prismtech.idt.idlmm.Container;
import com.prismtech.idt.idlmm.ExceptionDef;
import com.prismtech.idt.idlmm.Expression;
import com.prismtech.idt.idlmm.HomeDef;
import com.prismtech.idt.idlmm.IDLType;
import com.prismtech.idt.idlmm.InterfaceDef;
import com.prismtech.idt.idlmm.Value;
import com.prismtech.idt.idlmm.ValueDecl;
import com.prismtech.idt.idlmm.ValueDef;
import com.prismtech.idt.imports.idl.internal.importer.IdlImporterStatus;
import com.prismtech.idt.imports.idl.internal.importer.StatusHolder;
import com.prismtech.idt.imports.idl.parser.IdlParserAST;


/**
 * Utility class that performs names resolution
 * 
 * @since 12.07.2007 - 16:46:50
 * @author Aleksey Yashchenko
 * @version $Revision: 1.13 $
 */
public class ScopedNameResolver
{
   private final ResolvingContext m_resolvingContext;

   private StatusHolder m_statusHolder;

   public ScopedNameResolver (ResolvingContext cx, StatusHolder statusHolder)
   {
      m_resolvingContext = cx;
      m_statusHolder = statusHolder;
   }

   public Expression resolveExpression (ScopedNameNodePropertiesHolder snHolder)
   {
      String scopedName = snHolder.getScopedName ();
      Expression expression = ScopedNameResolver.resolveEntity (m_resolvingContext.getCurrentContext (), scopedName,
            Expression.class);

      if (expression == null)
      {
         m_statusHolder.addStatusItem (new IdlImporterStatus (snHolder.getASTs ()[0], IStatus.ERROR, "Undefined expression: "
               + scopedName));
      }

      return expression;
   }

   public IDLType resolveType (ScopedNameNodePropertiesHolder snHolder)
   {
      String scopedName = snHolder.getScopedName ();
      IDLType type = ScopedNameResolver.resolveEntity (m_resolvingContext.getCurrentContext (), scopedName, IDLType.class);

      if (type == null)
      {
         m_statusHolder.addStatusItem (new IdlImporterStatus (snHolder.getASTs ()[0], IStatus.ERROR, "Undefined type: "
               + scopedName));
      }

      return type;
   }

   public ExceptionDef resolveException (ScopedNameNodePropertiesHolder snHolder)
   {
      String scopedName = snHolder.getScopedName ();
      ExceptionDef ex = ScopedNameResolver.resolveEntity (m_resolvingContext.getCurrentContext (), scopedName, ExceptionDef.class);

      if (ex == null)
      {
         m_statusHolder.addStatusItem (new IdlImporterStatus (snHolder.getASTs ()[0], IStatus.ERROR, "Undefined exception: "
               + scopedName));
      }
      return ex;
   }

   public InterfaceDef resolveInterfaceDef (ScopedNameNodePropertiesHolder snHolder)
   {
      String name = snHolder.getScopedName ();
      Container context = m_resolvingContext.getCurrentContext ();

      Contained resolved = context.lookup (name);

      if (resolved instanceof InterfaceDef)
      {
         return (InterfaceDef) resolved;
      }

      String message = null;
      IdlParserAST ast = snHolder.getASTs ()[0];

      if (resolved == null)
      {
         message = "Undefined interface: " + name;
      }
      else
      {
         message = "Not an interface : " + name;
      }

      m_statusHolder.addStatusItem (new IdlImporterStatus (ast, IStatus.ERROR, message));

      return null;
   }

   public ValueDef resolveValueDef (ScopedNameNodePropertiesHolder snHolder)
   {
      String scopedName = snHolder.getScopedName ();
      Value value = ScopedNameResolver.resolveEntity (m_resolvingContext.getCurrentContext (), scopedName, Value.class);

      if (value == null)
      {
         m_statusHolder.addStatusItem (new IdlImporterStatus (snHolder.getASTs ()[0], IStatus.ERROR, "Undefined value: "
               + scopedName));
      }
      else if (value instanceof ValueDecl)
      {
         m_statusHolder.addStatusItem (new IdlImporterStatus (snHolder.getASTs ()[0], IStatus.ERROR, scopedName
               + " must be fully specified before inheritting from it"));
         return null;
      }
      return (ValueDef) value;
   }

   /**
    * Resolves an IDL entity by its scoped name within multiple Repositories
    * 
    * @param <T>
    * @param referrer
    * @param typeScopedName
    * @param entityType
    * @param resolver
    * @param currentFile
    * @return
    */
   private static <T> T resolveEntity (Container referrer, String scopedName, Class<T> entityType)
   {
      Object entity = referrer.lookup (scopedName);

      if (entityType.isInstance (entity))
      {
         return entityType.cast (entity);
      }
      return null;
   }
   
   public Contained resolveContained(ScopedNameNodePropertiesHolder snHolder)
   {
      String name = snHolder.getScopedName ();
      Container context = m_resolvingContext.getCurrentContext ();
      return context.lookup (name);      
   }

   public ComponentDef resolveComponentDef (ScopedNameNodePropertiesHolder snHolder)
   {
      String name = snHolder.getScopedName ();
      Container context = m_resolvingContext.getCurrentContext ();

      Contained resolved = context.lookup (name);

      if (resolved instanceof ComponentDef)
      {
         return (ComponentDef) resolved;
      }

      String message = null;
      IdlParserAST ast = snHolder.getASTs ()[0];

      if (resolved == null)
      {
         message = "Undefined component: " + name;
      }
      else
      {
         message = "Not a component : " + name;
      }

      m_statusHolder.addStatusItem (new IdlImporterStatus (ast, IStatus.ERROR, message));

      return null;
   }

   public HomeDef resolveHomeDef (ScopedNameNodePropertiesHolder snHolder)
   {
      String name = snHolder.getScopedName ();
      Container context = m_resolvingContext.getCurrentContext ();

      Contained resolved = context.lookup (name);

      if (resolved instanceof HomeDef)
      {
         return (HomeDef) resolved;
      }

      String message = null;
      IdlParserAST ast = snHolder.getASTs ()[0];

      if (resolved == null)
      {
         message = "Undefined home: " + name;
      }
      else
      {
         message = "Not a home : " + name;
      }

      m_statusHolder.addStatusItem (new IdlImporterStatus (ast, IStatus.ERROR, message));

      return null;
   }
}
