package com.prismtech.idt;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.Vector;
import java.util.Map.Entry;

import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.emf.ecore.xmi.impl.XMIResourceFactoryImpl;
import org.eclipse.emf.validation.model.IConstraintStatus;

import com.prismtech.idt.idlmm.IRObject;
import com.prismtech.idt.idlmm.IdlmmPackage;
import com.prismtech.idt.idlmm.Repository;
import com.prismtech.idt.imports.idl.driver.EMFValidator;
import com.prismtech.idt.imports.idl.driver.IIdlImportDriver;
import com.prismtech.idt.imports.idl.driver.IIdlImportDriverParam;
import com.prismtech.idt.imports.idl.driver.IIdlImportDriverParams;
import com.prismtech.idt.imports.idl.driver.IdlImportDriverFactory;
import com.prismtech.idt.imports.idl.importer.IIdlImporterStatus;
import com.prismtech.idt.imports.idl.internal.driver.IdlPreprocessorStatusAdapter;
import com.prismtech.idt.imports.idl.parser.IIdlParserStatus;

public class IDLImporter {

 // idlFile to be imported
 private File idlFile;
 
 // includes directories
 private List<File> includeDirs;
 
 // macros definitions
 private Map<String, String> macroDefs;

 // output directory
 private File outputDir;
 
 // Import Driver
 private IIdlImportDriver driver;
 
 // Imported model
 private Repository model;
 
 // debug flag
 private boolean debug;


 /**
  * Default constructor
  */
 public IDLImporter() {
  // Register meta-model package
  IdlmmPackage.eINSTANCE.eClass();

  // Register factory to load/save idlmm files
  Resource.Factory.Registry.INSTANCE.getExtensionToFactoryMap().put("idlmm", new XMIResourceFactoryImpl());

  // create Import Driver
  driver = IdlImportDriverFactory.create ();
  
  // set IDT's EMFValidator
  driver.setValidator(new EMFValidator());

  idlFile = null;
  includeDirs = new Vector<File>();
  macroDefs = new Hashtable<String, String>();
  outputDir = new File(System.getProperty("user.dir"));
  model = null;
  debug = false;
 }

 /**
  * Set IDL File to be imported
  */
 public void setIdlFile(File file) {
  if (!file.isFile()) {
   file = null;
   throw new IllegalArgumentException(
     file + " is not a valid file.");
  }
  this.idlFile = file;
 }
 
 /**
  * Add an include directory to be considered for pre-processing
  * @param includeDir The include directory
  */
 public void addIncludeDir(File includeDir) {
  includeDirs.add(includeDir);
 }
 
 /**
  * Add a macro definition (without value) to be considered for pre-processing
  * @param name The macro name
  */
 public void addMacro(String name) {
  addMacro(name, "");
 }
 
 /**
  * Add a macro definition (with a value) to be considered for pre-processing
  * @param name The macro name
  * @param value The macro value
  */
 public void addMacro(String name, String value) {
  macroDefs.put(name, value);
 }
 
 
 /**
  * Remove a macro definition. If this macro wasn't defined, nothing is done.
  * @param name The macro name
  */
 public void removeMacro(String name) {
  macroDefs.remove(name);
 }
 
 /**
  * Return the model imported from IDL File. 
  * This operation has to be called after doImport() call. 
  * Otherwise null is returned.
  * @return the imported model, or null if doImport() wasn't called 
  * or if an error occured during import.
  */
 public Repository getModel() {
  return model;
 }
 
 
 /**
  * Return output directory, specified with -o option.
  * If not specified, it's "user.dir" property.
  * @return the output directory.
  */
 public File getOutputDir() {
  return outputDir;
 }
 
 /**
  * Parse command line options.
  * Returns the list of non-recognized options (not considered)
  * Accepted options are:<ul>
  *     <li>--help  : print help
  *     <li>-I(dir) : add include dir
  *     <li>-D(macro=value) : define a macro
  *     <li>-U(macro) : undefine a macro
  *     <li>-O(dir) : output directory for model file
  *     <li>--debug : print debug information
  *     <li>(file) : IDL file to be imported
  * @param args The command line options
  * @return the ignored options
  */
 public List<String> parseOptions(List<String> args) 
  throws Exception
 {
  List<String> ignored = new Vector<String>();
  
  for (String arg : args) {

   // if arg doesn't start with "-", it's an idl filename
   if (!arg.startsWith("-")) {
    if (idlFile == null) {
     setIdlFile(new File(arg));
    } else {
     // only 1 idl file authorized. Ignore this one
     ignored.add(arg);
    }
   }

   else if (arg.equals("--help")) {
    printHelp();
    // exit program
    System.exit(0);
   }
   
   else if (arg.equals("--debug")) {
    debug = true;
   }
   
   // -I<include-dir>
   else if (arg.startsWith("-I")) {
    String include = arg.substring(2);
    if (include.length() == 0) {
     throw new Exception(
     "'-I' without specified directory is not a valid option.");
    }

    addIncludeDir(new File(include));
   }

   // -D<name>=<value>
   else if (arg.startsWith("-D")) {
    String macro = arg.substring(2);
    if (macro.length() == 0) {
     throw new Exception(
     "'-D' without specified name is not a valid option.");
    }

    int i = macro.indexOf('=');
    if (i == -1) {
     addMacro(macro);
    } else {
     String name = macro.substring(0, i);
     String value = macro.substring(i+1);
     addMacro(name, value);
    }
   }
   
   // -U<name>
   else if (arg.startsWith("-U")) {
    String name = arg.substring(2);
    if (name.length() == 0) {
     throw new Exception(
     "'-U' without specified name is not a valid option.");
    }
    removeMacro(name);
   }
   
   // -O<dir>
   else if (arg.startsWith("-O")) {
    String dirname = arg.substring(2);
    if (dirname.length() == 0) {
     throw new Exception(
     "'-O' without specified directory is not a valid option.");
    }
    
    File dir = new File(dirname);
    if (!dir.exists()) {
     if (!dir.mkdirs()) {
      throw new Exception(
       "Failed to create output directory " + dir);
     }
    }
    if (!dir.isDirectory()) {
     throw new Exception(
       arg + " is not a valid directory.");
    }
    outputDir = dir;
   }
   
   else {
    ignored.add(arg);
   }
  }
  
  return ignored;
 }

 /**
  * Actually do import of IDL file as model file.
  * @return a list of error status. 
  * If list is empty, import was successful.
  */
 public List<IStatus> doImport() 
 {
  // Create Driver parameters
  IIdlImportDriverParams params = driver.createDriverParams ();
  
  // add IDL file to parameters
  IIdlImportDriverParam parameter = params.add (idlFile);

  // add includes to parameters
  configureIncludePaths (parameter);

  // add macros to parameters
  configureMacroDefinitions (parameter);

  IStatus internalError = null;
  try {
   // import IDL and build model 
   driver.build (params);
   
   // get resulting model
   model = parameter.getRepository ();
   
  } catch (Throwable e) {
   internalError = new Status(IStatus.ERROR, 
            this.getClass().getPackage().toString(), 
            "Internal Error", 
            e);
  }

  if (debug) {
   System.out.println("\n------- AST -------");
   com.prismtech.idt.imports.idl.parser.util.ASTDumper.fullDump(System.out, 
     parameter.getAST());
   System.out.println("-------------------\n");
  }
  
  // extract list of non-OK status from Driver parameters
  List<IStatus> result = extractErrorStatusList(params);
  
  // if internal error occurs, add its status
  if (internalError != null)
   result.add(internalError);
  
  return result;
 }

 /**
  * Add includes directories in Driver parameters
  * @param param driver parameters
  */
 private void configureIncludePaths(IIdlImportDriverParam param) {
  try {
   // set working directory
   String currentDir = System.getProperty("user.dir");
   param.getPreprocessorParams ().setWorkingDirectory(new File(currentDir));
 
   // add include directories
   for (File dir : includeDirs) {
    param.getPreprocessorParams ().addIncludeDirectory(dir);
   }
 
  } catch (FileNotFoundException e) {
   // shouldn't happen
   e.printStackTrace();
  }
 }

 /**
  * Add macros definitions in Driver parameters
  * @param param driver parameters
  */
 private void configureMacroDefinitions(IIdlImportDriverParam param) {
  for (Entry<String, String> e : macroDefs.entrySet()) {
   param.getPreprocessorParams ().addMacroDefinition(e.getKey(), e.getValue());
  }
 }


 /**
  * Save the importing resulting model file.
  * This operation should be called after doImport() and only if 
  * all resulting status are OK. Otherwise, an Exception is thrown.
  * @throws Exception if model wasn't created.
  */
 public void saveIdlModelFile() 
  throws Exception
 {
  if (model == null)
  {
   throw new Exception("Internal error: built model is null !");
  }

  String modelFilename = idlFile.getName().replaceAll("idl$", "idlmm");
  File modelFile = new File(outputDir, modelFilename);
  
  URI uri = URI.createFileURI(modelFile.getAbsolutePath());

  ResourceSet resourceSet = new ResourceSetImpl();
  Resource resource = resourceSet.createResource(uri);
  resource.getContents().add(model);

  resource.save (null);
  
  System.out.println("Model created: " + modelFile);
  
  if (debug) {
   System.out.print("\n---- generated model ----");
   FileReader fr = new FileReader(modelFile);
   int c;
   while ((c = fr.read()) != -1) {
    System.out.print((char)c);
   }
   System.out.print("-------------------------\n");
  }
 }

 
 /**
  * Extract the list of IStatus from driver parameters
  * @param params Driver parameters
  * @return the list of IStatus
  */
 private static List<IStatus> extractErrorStatusList(IIdlImportDriverParams params) {
  List<IStatus> result = new Vector<IStatus>();
  for (IIdlImportDriverParam param : params) {
   addErrorStatus(param.getStatus(), result);
  }
  return result;
 }

 
 
 /**
  * Add a IStatus to a list of IStatus only if the status is not OK.
  * @param status the status to be added
  * @param result the list for status to be added
  */
 private static void addErrorStatus(IStatus status, List<IStatus> result) {
  // if multi-status, add all non-OK children status
  if (status.isMultiStatus ())
  {
   for (IStatus child : status.getChildren ())
   {
    addErrorStatus(child, result);
   }
  }
  else
  {
   // if status is not OK, add it
   if (!status.isOK()) {
    result.add(status);
   }
  }
 }

 /**
  * Print error messages from a list of IStatus.
  * @param status the list of IStatus
  */
 public static void printErrorStatus(List<IStatus> status) {
  for (IStatus s : status) {
   printErrorStatus(s);
  }
 }
 
 /**
  * Print error messages from a non-OK IStatus.
  * @param status the IStatus
  */
 @SuppressWarnings("restriction")
 public static void printErrorStatus(IStatus status) {
  if (status.isOK())
   return;
  
  if (status instanceof IdlPreprocessorStatusAdapter) {
   IdlPreprocessorStatusAdapter parserStatus = (IdlPreprocessorStatusAdapter) status;
   System.err.println("Error in " + parserStatus.getFilename() +
     " at line " + parserStatus.getLine() +
     " (char "   + parserStatus.getColumn() + 
     "): "       + parserStatus.getMessage());
  }
  
  else if (status instanceof IIdlParserStatus) {
   IIdlParserStatus parserStatus = (IIdlParserStatus) status;
   System.err.println("Error in " + parserStatus.getFilename() +
     " at line " + parserStatus.getLine() +
     " (char "   + parserStatus.getColumn() + 
     "): "       + parserStatus.getMessage());
  }

  else if (status instanceof IIdlImporterStatus) {
   IIdlImporterStatus importStatus = (IIdlImporterStatus) status;
   System.err.println("Error in " + importStatus.getFilename() + 
     " at line " + importStatus.getLine() + 
     " (char "   + importStatus.getColumn() + 
     "): "       + importStatus.getMessage());
  }

  else if (status instanceof IConstraintStatus) {
   IConstraintStatus constraintStatus = (IConstraintStatus) status;

   String filename = "(unknown file)";
   String line = "?";
   String column = "?";

   EObject target = constraintStatus.getTarget();
   if (target instanceof IRObject) {
    IRObject irobject = (IRObject) target;
    String source = irobject.getSource();
    filename = source.substring(0, source.indexOf('#'));
    line = source.substring(source.indexOf('#')+1, source.indexOf(','));
    column = source.substring(source.indexOf(',')+1);

   }
   System.err.println("Error in " + filename + 
     " at line " + line + 
     " (char "   + column + 
     "): "       + constraintStatus.getMessage());
  }
  else if (status.getClass().toString().startsWith("class org.eclipse.emf.common")) {
   // Produced model is not a valid EMF model
   // This could be because of previous errors during import
   System.err.println("Cannot produce a valid model (probably because of previous errors):");
   System.err.println("   " + status.getMessage());
  }
  else {
   System.err.println(status.getClass()+":");
   if (status.getMessage() != null && status.getMessage().length() > 0)
    System.err.println("  " + status.getMessage());
   if (status.getException() != null)
    status.getException().printStackTrace();
  }
 }

 /**
  * Print help message.
  */
 public static void printHelp() {
  String pkg = IDLImporter.class.getPackage().getName();
  System.out.println("Usage: java -jar " + pkg + ".jar [OPTION]... [IDL file]");
  System.out.println("Import the IDL file into a model file.");
  System.out.println("");
  System.out.println("Options:");
  printOptions();
 }
 
 /**
  * Print available options
  */
 public static void printOptions() {
  System.out.println("  -Idir         \t includes dir in search path for preprocessor");
  System.out.println("  -Dname[=value]\t defines name for preprocessor");
  System.out.println("  -Uname        \t undefines name for preprocessor");
  System.out.println("  -Odir         \t output directory for the generated model. Default is current directory");
  System.out.println("  --help        \t print this help");
  System.out.println("");  
 }
 
 /**
  * Main operation.
  * @param args
  */
 public static void main(String[] args) {
  System.out.println("IDLImporter...");
  
  if (args.length == 0) {
   printHelp();;
   return;
  }

  // create IDL importer
  IDLImporter importer = new IDLImporter();

  try {
   List<String> argList = new Vector<String>();
   for (String arg : args) {
    argList.add(arg);
   }
   
   // parse command line
   argList = importer.parseOptions(argList);
   
   if (!argList.isEmpty()) {
    throw new Exception("Unrecognized options: " + argList);
   }

   // do import
   List<IStatus> status = importer.doImport();
   
   // check is there are error status
   if (status.isEmpty()) {
    // no error status: save model file
    importer.saveIdlModelFile();
   } else {
    // print error status
    IDLImporter.printErrorStatus(status);
   }

  } catch (Exception e) {
   System.err.println(e.getMessage());
   System.err.println("");
   printHelp();
   System.exit(1);
  }
 }

}
