package com.thalesgroup.ccm.codegenerator;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import org.eclipse.core.runtime.IStatus;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.ecore.EObject;

import com.prismtech.idt.IDLImporter;
import com.prismtech.idt.idlmm.ConsumesDef;
import com.prismtech.idt.idlmm.HomeDef;
import com.prismtech.idt.idlmm.ProvidesDef;
import com.prismtech.idt.idlmm.Repository;
import com.prismtech.idt.idlmm.UsesDef;
import com.thalesgroup.ccm.codegenerator.container.CommonServices;


public class Generator {

	private IDLImporter importer;
	private AcceleoDriver acceleoDriver;
	private boolean saveModel;
	private String lang;
	
	public Generator() 
		throws Exception
	{
		importer = new IDLImporter();
		acceleoDriver = null;
		saveModel = false;
		lang = null;
	}
	

	/**
	 * Parse command line options.
	 * Returns the list of non-recognized options (not considered)
	 * Accepted options are:<ul>
	 *     <li>--home=(absName)[,session|,extension] : home (absolute name) to be considered for generation and container type (by default session)
	 *     <li>--lang=(java|cpp) : language of generate code (required option)
	 *     <li>--lfc_interceptors=[true|false] : generated lifecycle interception code (default = true)
	 *     <li>--help : print help
	 *     <li>Plus options accepted by com.prismtech.idt.IDLImporter.parseOptions()
	 * @param args The command line options
	 * @return the ignored options
	 */
	public List<String> parseOptions(List<String> args) 
		throws Exception
	{
		List<String> ignored = new Vector<String>();
		
		for (String arg : args) {

			if (arg.startsWith("--home=")) {
				String homeName = arg.substring(7);
				String contType = "session";
				int idx = homeName.indexOf(',');
				if (idx != -1) {
					contType = homeName.substring(idx+1);
					homeName = homeName.substring(0, idx);
				}
				
				if (homeName.length() == 0) {
					throw new Exception(
					    "'--home=' option specified without any home name");
				}
				
				// select home with container type for code generation 
				CommonServices.selectHome(homeName, contType);
			}

			else if (arg.startsWith("--lang=")) {
            	if (lang != null) {
                    throw new Exception(
                    "'--lang=' option must be used only once");
                }
            	lang = arg.substring(7);
                if (lang.length() == 0) {
                    throw new Exception(
                        "'--lang=' option specified without language");
                }
                if (!lang.equals("cpp") && !lang.equals("java")) {
                    throw new Exception(
                    "Unknown language for '--lang=' option: " + lang);
                }
			}
			
			else if (arg.startsWith("--lfc_interceptors=")) {
				String opt = arg.substring(19);
				if (opt.length() == 0) {
					throw new Exception(
					    "'--lfc_interceptors=' option specified without value");
				}
				
				if (!opt.toUpperCase().equals("TRUE") && 
				    !opt.toUpperCase().equals("FALSE"))
				{
					throw new Exception(
				    	"'--lfc_interceptors=' option specified with a wrong value: " + opt);
				}

				// set LIFECYCLE_INTERCEPTOR_PROPERTY
				System.setProperty(CommonServices.CONTAINER_INTERCEPTOR_PROPERTY, opt);
			}

			else if (arg.startsWith("--save-model")) {
				saveModel = true;
			}
			
			else if (arg.equals("--help")) {
				printHelp();
				// exit program
				System.exit(0);
			}
			
			else {
				ignored.add(arg);
			}
		}
		
		// check if resuired --lang option has been set
		if (lang == null) {
			throw new Exception("required --lang=(java|cpp) option not specified.");
		}
		
		// let IDLImporter parse its options
		ignored = importer.parseOptions(ignored);
		
		return ignored;
	}
	
	
	
	/**
	 * Print help message.
	 */
	public static void printHelp() {
		String pkg = Generator.class.getPackage().getName();
		System.out.println("Usage: java -jar " + pkg + ".jar [OPTION]... [IDL file]");
		System.out.println("Generate code for a set of selected homes from an IDL file.");
		System.out.println("");
		System.out.println("Options:");
		printOptions();
	}

	/**
	 * Print available options
	 */
	public static void printOptions() {
		System.out.println("  --home=(absName)[,session|,extension]  home (absolute name) to be considered for generation and container type (by default session)");
		System.out.println("                                         If no home is specified, code will be generated for all homes found in IDL.");
		System.out.println("  --lang=(java|cpp)                      language of generate code (required option)");
		System.out.println("  --lfc_interceptors=[true|false]        generated lifecycle interception code (default = true)");
		System.out.println("  --save-model                           save the IDL model file");
		IDLImporter.printOptions();
	}
		
		
	public boolean doImport() {
			// import IDL as model
			System.out.println("import IDL...");
			List<IStatus> status = importer.doImport();
			
			// check resulting status
			if (!status.isEmpty()) {
				IDLImporter.printErrorStatus(status);
				return false;
			}
			
			System.out.println("import IDL OK");
			
			if (saveModel) {
				try {
					importer.saveIdlModelFile();
					System.out.println("IDL model saved");
				} catch (Exception e) {
					System.err.println("Error saving IDL model file:\n " + e);
				}
			}
			return true;
	}
	
	public com.prismtech.idt.idlmm.Repository getModel()
	{
		return importer.getModel();
	}
	
	public void doGeneration() 
		throws Exception
	{
		// get imported model
		com.prismtech.idt.idlmm.Repository rep = importer.getModel();

		// init AcceleoDriver with appropriate scripts
		if (lang.equals("java")) {
			acceleoDriver = new AcceleoDriver(com.thalesgroup.ccm.codegenerator.container.java.GenerateAll.emtFile);
		} else if (lang.equals("cpp")) {
			acceleoDriver = new AcceleoDriver(com.thalesgroup.ccm.codegenerator.container.cpp.GenerateAll.emtFile);			
		} else {
			throw new Exception("Unkown language: " + lang);
		}
		// set output directory
		acceleoDriver.setOutputDir(importer.getOutputDir().getAbsolutePath());

/*
 * Following code search all applicable scripts (i.e. producing file) for all model elements.
 * This is not acceptable for us. We want to control for which model elements which files are produced.
 * 
		// start generation
		System.out.println("start generation...");		
		acceleoDriver.generate("GenerateAll", rep);

		System.out.println("generation OK");		
 *
 *
 */	
		
		/*
		 * This is a workaround for preceding issue.
		 * This code reproduces what the GenerateAll.mt script should do.
		 */
	
		// get selected homes (model elements + container type)
		Map<HomeDef, String> homes = getSelectedHomes(rep);
		
		// for each selected home
		for (Map.Entry<HomeDef, String> set : homes.entrySet()) {
			HomeDef home = set.getKey();
			
			if (lang.equals("java")) {
				generateJava(home, set.getValue());
			} else if (lang.equals("cpp")) {
				generateCpp(home, set.getValue());
			} else {
				throw new Exception("Unkown language: " + lang);
			}
		}
		
	}
	
	
	private void generateJava(HomeDef home, String containerType)
		throws Exception
	{
		System.out.println("Generate java " + containerType + 
				" code for " + home.getAbsoluteName());

		// generate home code
		acceleoDriver.generate("SessionHome", home);

		// generate component code 
		acceleoDriver.generate("SessionComponent", home.getManagedComponent());
		// generate context code 
		acceleoDriver.generate("SessionContext", home.getManagedComponent());

		// if Extension container required:
		if (containerType.toUpperCase().equals("EXTENSION")) {
			// generate home code (extension container)
			acceleoDriver.generate("ExtensionHome", home);
			// generate component code (extension container)
			acceleoDriver.generate("ExtensionComponent", home.getManagedComponent());
			// generate context code (extension container)
			acceleoDriver.generate("ExtensionContext", home.getManagedComponent());
		}

		// generate all facets code 
		for (ProvidesDef port : home.getManagedComponent().getAllProvidesPorts()) {
			acceleoDriver.generate("Facet", port.getInterface_type());
		}
		// generate all consumers code 
		for (ConsumesDef port : home.getManagedComponent().getAllConsumesPorts()) {
			acceleoDriver.generate("Consumer", port.getEvent());
		}

		// generate all facets code (with interception) 
		for (ProvidesDef port : home.getManagedComponent().getAllProvidesPorts()) {
			acceleoDriver.generate("COPIFacet", port.getInterface_type());
		}
		// generate all receptacle code (with interception) 
		for (UsesDef port : home.getManagedComponent().getAllUsesPorts()) {
			acceleoDriver.generate("COPIReceptacle", port.getInterface_type());
		}
	}
	

	private void generateCpp(HomeDef home, String containerType)
	throws Exception
	{
		System.out.println("Generate c++ " + containerType + 
				" code for " + home.getAbsoluteName());

		// generate home code
		acceleoDriver.generate("SessionHome_hpp", home);
		acceleoDriver.generate("SessionHome_cpp", home);

		// generate component code 
		acceleoDriver.generate("SessionComponent_hpp", home.getManagedComponent());
		acceleoDriver.generate("SessionComponent_cpp", home.getManagedComponent());
		// generate context code 
		acceleoDriver.generate("SessionContext_hpp", home.getManagedComponent());
		acceleoDriver.generate("SessionContext_cpp", home.getManagedComponent());

		// if Extension container required:
		if (containerType.toUpperCase().equals("EXTENSION")) {
			// generate home code (extension container)
			acceleoDriver.generate("ExtensionHome_hpp", home);
			acceleoDriver.generate("ExtensionHome_cpp", home);
			// generate component code (extension container)
			acceleoDriver.generate("ExtensionComponent_hpp", home.getManagedComponent());
			acceleoDriver.generate("ExtensionComponent_cpp", home.getManagedComponent());
			// generate context code (extension container)
			acceleoDriver.generate("ExtensionContext_hpp", home.getManagedComponent());
			acceleoDriver.generate("ExtensionContext_cpp", home.getManagedComponent());
		}

		// generate all facets code 
		for (ProvidesDef port : home.getManagedComponent().getAllProvidesPorts()) {
			acceleoDriver.generate("Facet_hpp", port.getInterface_type());
			acceleoDriver.generate("Facet_cpp", port.getInterface_type());
		}
		// generate all consumers code 
		for (ConsumesDef port : home.getManagedComponent().getAllConsumesPorts()) {
			acceleoDriver.generate("Consumer_hpp", port.getEvent());
			acceleoDriver.generate("Consumer_cpp", port.getEvent());
		}

		// generate all facets code (with interception) 
		for (ProvidesDef port : home.getManagedComponent().getAllProvidesPorts()) {
			acceleoDriver.generate("COPIFacet_hpp", port.getInterface_type());
			acceleoDriver.generate("COPIFacet_cpp", port.getInterface_type());
		}
		// generate all receptacle code (with interception) 
		for (UsesDef port : home.getManagedComponent().getAllUsesPorts()) {
			acceleoDriver.generate("COPIReceptacle_hpp", port.getInterface_type());
			acceleoDriver.generate("COPIReceptacle_cpp", port.getInterface_type());
		}
	}

	
	/*
	 * Returns a map of selected homes (model element for home + container type)
	 */
	private Map<HomeDef, String> getSelectedHomes(Repository rep) {
		// get all homes from model
		List<HomeDef> allHomes = new LinkedList<HomeDef>();
		TreeIterator<EObject> all = rep.eAllContents();
		while (all.hasNext()) {
			EObject object = (EObject) all.next();
			if (object instanceof HomeDef) {
				allHomes.add((HomeDef) object);
			}
		}
		
		Map<HomeDef, String> selectedHomes = new HashMap<HomeDef, String>();
		
		
		if (CommonServices.SelectedHomes.size() > 0) {
			// for each selected homes
			for (String homeName : CommonServices.SelectedHomes.keySet()) {
				// find it amongst all homes from model
				HomeDef homeFound = null;
				for (HomeDef homeDef : allHomes) {
					if (homeDef.getAbsoluteName().equals(homeName)) {
						homeFound = homeDef;
						break;
					}
				}
				if (homeFound !=  null) {
					selectedHomes.put(homeFound, CommonServices.SelectedHomes.get(homeName));
				} else {
					System.err.println("WARNING: Home '" + homeName + "' not found.\n");
				}
			}
		} else {
			// no home selected: select all homes and generate session container by default
			for (HomeDef homeDef : allHomes) {
				selectedHomes.put(homeDef, "container");
			}
		}
			
		return selectedHomes;
	}
		
	
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {

		if (args.length == 0) {
			printHelp();
			return;
		}
		
		try {
			Generator g = new Generator();

			List<String> argList = new Vector<String>();
			for (String arg : args) {
				argList.add(arg);
			}
			
			// parse command line
			argList = g.parseOptions(argList);
			
			if (!argList.isEmpty()) {
				throw new Exception("Unrecognized options: " + argList);
			}
			
			// import model from IDL file
			if (g.doImport()) {
				// do code generation
				g.doGeneration();
			}
			
		} catch (Exception e) {
			System.err.println(e.getMessage());
			System.err.println("");
			printHelp();
			System.exit(1);
		}
	}

}
