package com.thalesgroup.cdmw.codegenerator;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileWriter;
import java.io.InputStreamReader;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import com.prismtech.idt.idlmm.Repository;
import com.thalesgroup.ccm.codegenerator.AcceleoDriver;
import com.thalesgroup.ccm.codegenerator.Generator;
import com.thalesgroup.ccm.codegenerator.container.CommonServices;
import com.thalesgroup.cdmw.codegenerator.CdmwXMLFile.FactoryImplData;
import com.thalesgroup.cdmw.codegenerator.CdmwXMLFile.HomeData;
import com.thalesgroup.cdmw.codegenerator.cdmwinit.DatastoreTypesServices;
import com.thalesgroup.cdmw.codegenerator.cdmwinit.InitDTD;
import com.thalesgroup.cdmw.codegenerator.cdmwinit.DatastoreTypesServices.DatastoreTypeData;
import com.thalesgroup.cdmw.codegenerator.cdmwinit.cpp.DataStoreTypeTraits;
import com.thalesgroup.cdmw.codegenerator.cdmwinit.cpp.GlobalDataStores;

public class Main {

	private final static String EXEC_IDL_GENERATOR = "cdmw_idl_generator.sh";
	
	
	// options
	private String xmlFile;
	private String outDir;
	private String lang;
	private boolean validateOnly;
	private boolean genEcoreOnly;
	private boolean genIdl2Only;
	
	// Cdmw XML file data
	CdmwXMLFile cdmwFile;
	private List<String> filesToConvert;
	private List<String> idlSearchDirs;
	private List<String> factoriesIdlFiles;
	private List<String> datastoresIdlFiles;
	private String preprocFlags;
	private Map<String, HomeData> homesData;
	
	
	/**
	 * constructor
	 * @throws Exception
	 */
	public Main() 
	{
		xmlFile = null;
		outDir = null;
		lang = null;
		validateOnly = false;
		genEcoreOnly = false;
		genIdl2Only = false;
		
		cdmwFile = null;
		filesToConvert = null;
		idlSearchDirs = null;
		factoriesIdlFiles = null;
		datastoresIdlFiles = null;
	}
	

	/**
	 * Parse command line options.
	 * Returns the list of non-recognized options (not considered)
	 * Accepted options are:<ul>
     *     <li>--help                            display this screen
     *     <li>--xml=[FILE]               (*)    file name of the XML doc.
     *     <li>--dir=[DIR]                (*)    default output directory
     *     <li>--lang=[cpp|java]          (*)    implementation languages
     *     <li>--validate-only                   validate input XML and exit
     *     <li>--generate-ecore-only             export to ecore model and exit
     *     <li>--generate-idl2-only              export to IDL2 and exit
     * @param args The command line options
     * @return the ignored options
     */
    public List<String> parseOptions(List<String> args) 
        throws Exception
    {
        List<String> ignored = new Vector<String>();
        
        for (String arg : args) {

            if (arg.startsWith("--xml=")) {
                if (xmlFile != null) {
                    throw new Exception(
                    "'--xml=' option must be used only once");
                }
                xmlFile = arg.substring(6);
                if (xmlFile.length() == 0) {
                    throw new Exception(
                        "'--xml=' option specified without XML file name");
                }
            }

            else if (arg.startsWith("--dir=")) {
            	if (outDir != null) {
                    throw new Exception(
                    "'--dir=' option must be used only once");
                }
            	outDir = arg.substring(6);
                if (outDir.length() == 0) {
                    throw new Exception(
                        "'--dir=' option specified without directory name");
                }
            }

            else if (arg.startsWith("--lang=")) {
            	if (lang != null) {
                    throw new Exception(
                    "'--lang=' option must be used only once");
                }
            	lang = arg.substring(7);
                if (lang.length() == 0) {
                    throw new Exception(
                        "'--lang=' option specified without language");
                }
                if (!lang.equals("cpp") && !lang.equals("java")) {
                    throw new Exception(
                    "Unknown language for '--lang=' option: " + lang);
                }
            }

            else if (arg.equals("--validate-only")) {
            	validateOnly = true;
            }

            else if (arg.equals("--generate-ecore-only")) {
            	genEcoreOnly = true;
            }

            else if (arg.equals("--generate-idl2-only")) {
            	genIdl2Only = true;
            }

            else if (arg.equals("--help")) {
                printHelp();
                // exit program
                System.exit(0);
            }
            
            else {
                ignored.add(arg);
            }
        }

		// check if resuired --lang option has been set
		if (lang == null) {
			throw new Exception("required --lang=(java|cpp) option not specified.");
		}
        
        return ignored;
    }
    
    
	public void checkOutDir() 
		throws Exception
	{
        // check if outDir exists and is a directory
        // (create it if not existing)
        File dir = new File(outDir);
        
        if (!dir.exists()) {
        	dir.mkdirs();
        } else if (!dir.isDirectory()) {
        	throw new Exception(outDir + " is not a directory.");
        }
	}

    
    /**
     * Print help message.
     */
    public static void printHelp() {
        String pkg = Main.class.getPackage().getName();
        System.out.println("NAME");
        System.out.println("     " + pkg);
        System.out.println();        
        System.out.println("SYNOPSIS");
        System.out.println("     java -jar "+pkg+".jar --xml=<FILE> --dir=<DIR>");
        System.out.println("      [ --lang=<cpp|java> ] [ --generate-idl2-only ]");
        System.out.println("      [ --generate-ecore-only ] [ --validate-only ]");
        System.out.println("      [ --help ]");
        System.out.println();
        System.out.println("DESCRIPTION");
        System.out.println("     This program reads an XML file used for the CDMW code generation");
        System.out.println("     and generates source code in native language for the CDMW");
        System.out.println("     services and/or CIF using information gathered from the XML file.");
        System.out.println();
        System.out.println("OPTIONS");
        printOptions();
    }

    /**
     * Print available options
     */
    public static void printOptions() {
        System.out.println("     --help                            display this screen");
        System.out.println("     --xml=<FILE>               (*)    file name of the XML doc.");
        System.out.println("     --dir=<DIR>                (*)    default output directory");
        System.out.println("     --lang=<cpp|java>                 implementation languages");
        System.out.println("     --validate-only                   validate input XML and exit");
        System.out.println("     --generate-ecore-only             export to ecore model and exit");
        System.out.println("     --generate-idl2-only              export to IDL2 and exit");
        System.out.println();
        System.out.println("(*) = those options are mandatory");

    }
       
    /**
     * Load code generation XML file
     * @throws Exception if an error occurs
     */
    public void loadXMLFile() 
    	throws Exception
    {
    	String cdmwHome = System.getenv("CDMW_HOME");
    	if (cdmwHome == null || cdmwHome.length() == 0) {
    		throw new Exception("Environment variable $CDMW_HOME is not defined.");
    	}
    	
    	cdmwFile = new CdmwXMLFile(xmlFile, cdmwHome + "/share");
    	
    	// if only <factories> are specified, we just need to generate IDL2
    	if (cdmwFile.hasOnlyFactoriesDefs()) {
    		genIdl2Only = true;
    	}
    	
    	// get all IDL files to load in model
    	filesToConvert = cdmwFile.getIdlFilesToConvert();
    	factoriesIdlFiles = cdmwFile.getFactoriesIdlFilenames();
    	datastoresIdlFiles = cdmwFile.getDatastoresIdlFilenames();
    	// get search directories for IDL files
    	idlSearchDirs = cdmwFile.getIdlSearchDirs();
    	// get preprocessor flags
    	preprocFlags = cdmwFile.getPreprocessingFlags();
    	// add -DTAO to preprocFlags
    	preprocFlags = "-DTAO " + preprocFlags;
    	
    	homesData = cdmwFile.getHomesData();
    }
    
    /**
     * Do IDL2 generation.
     * @throws Exception if an error occurs
     */
    public void generateIDL2()
    	throws Exception
	{
    	List<String> filesToExclude = cdmwFile.getIdlFilesToExclude();
    	
    	// write files to exclude in a tmp file
    	File excludePropFile = File.createTempFile("ExcludedFiles", ".properties");
    	FileWriter fw = new FileWriter(excludePropFile);
    	for (String s : filesToExclude) {
			fw.write(s + "\n");
		}
    	fw.flush();
    	fw.close();
    	
    	// pre-build command
    	String precommand = EXEC_IDL_GENERATOR;
    	precommand += " " + preprocFlags;
    	for (String s : idlSearchDirs) {
    		precommand += " -I" + s;
    	}
    	precommand += " -DexcludedFile=" + excludePropFile.getAbsolutePath();
    	precommand += " -o " + outDir;
    	
    	// run command for each file to convert
    	for (String file : filesToConvert) {
    		String filePath = findFile(file);
    		String command = precommand + " " + filePath;
    		System.out.println("Generate IDL2 for " + filePath + " : ");
    		System.out.println(command);
    		runCommand(command);
    		System.out.println("Generation done");
    	}
    	
    	// delete tmp file
    	excludePropFile.delete();
	}
    
    /**
     * Find a file in idlSearchDirs, ans return its absolute path
     * @param file the file to find
     * @return the file's absolute path
     * @exception Exception if the file was not found
     */
    private String findFile(String file) 
    	throws Exception
    {
    	// first, try in current directory
    	File f = new File(file);
    	if (f.isFile())
    		return f.getAbsolutePath();
    	
    	// now look in all idlSearchDirs
    	for (String searchDir : idlSearchDirs) {
    		f = new File(searchDir + File.separator + file);
    		if (f.isFile())
        		return f.getAbsolutePath();
		}
    	
		throw new Exception("Cannot find file '" + file + "' in specified searrch directories.");
	}


	/**
     * Run a command and wait until completion
     * @param command The command to run
     * @throws Exception if an error occurs
     */
    private void runCommand(String command) 
    	throws Exception
    {
    	// start command
    	Process p = Runtime.getRuntime().exec(command);
    	
    	// wait for completion
    	int result = p.waitFor();
    	
    	// print command output & error streams
    	printStream(p.getInputStream());
    	printStream(p.getErrorStream());
    	
    	if (result != 0) {
    		throw new Exception("Error in '" + 
    				command.substring(0, command.indexOf(' ')) + 
    				"' result");
    	}
    }
    
    /**
     * Print content of an InputStream in standard output
     * @param stream The InputStream
     * @throws Exception if an error occurs
     */
    private void printStream(java.io.InputStream stream) 
    	throws Exception
    {
        BufferedReader inReader =
	            new BufferedReader(new InputStreamReader(stream));
	     String line;
	     while ((line = inReader.readLine()) != null) { 
	             System.out.println(line);
	     } 
    }
    
    
    
    /**
     * Create a dummy IDL file including all IDL files to be converted
     * @return the dummy IDL file
     * @exception Exception if an error occurs
     */
    private File createDummyIDL() 
    	throws Exception
    {
    	File dummy = File.createTempFile("dummy", ".idl");
    	FileWriter fw = new FileWriter(dummy);

    	//fw.write("#include <Components.idl>\n");
    	//fw.write("#include <CdmwLifeCycle.idl>\n");
    	
    	// include all files to convert
    	for (String file : filesToConvert) {
        	fw.write("#include <" + file + ">\n");
		}

    	// include all idl files referenced by XML file
    	for (String file : factoriesIdlFiles) {
        	fw.write("#include <" + file + ">\n");
		}
    	for (String file : datastoresIdlFiles) {
        	fw.write("#include <" + file + ">\n");
		}

    	// add a dummy module and interface for the IDL to be valid
    	fw.write("\nmodule Cdmw_dummy {\n");
    	fw.write("   interface dummy {\n");
    	fw.write("   };\n");
    	fw.write("};\n");

    	fw.flush();
    	fw.close();
    	
    	return dummy;
    }
    


	private void generateLifeCycle(Repository model) 
		throws Exception
	{
		if (cdmwFile.getFactoriesData().size() > 0) {
			System.out.println("Generate LifeCycle frameworks ");
			
			if (lang.equals("java")) {
				generateLifeCycleJava(model);
			} else if (lang.equals("cpp")) {
				generateLifeCycleCpp(model);
			} else {
				throw new Exception("Unkown language: " + lang);
			}
		}
	}

	private void generateLifeCycleJava(Repository model) 
		throws Exception
	{
		// LifeCycle generator
		AcceleoDriver lifeCycleDriver = new AcceleoDriver(
			com.thalesgroup.cdmw.codegenerator.lifecycle.java.LifeCycle.emtFile);
		lifeCycleDriver.setOutputDir(outDir);

		for (FactoryImplData factory : cdmwFile.getFactoriesData()) {
			// prepare Acceleo Service with all info
    		// for this factory-impl
			cdmwFile.prepareLifeCycleServices(factory.classname, model);
			
			// generate ServantProvider if required
			if (factory.hasServantProvider) {
				lifeCycleDriver.generate("ServantProvider", model);
			}
			
			// generate appropriate Factory
			if (factory.category.equals("ENTITY")) {
				lifeCycleDriver.generate("EntityFactory", model);
				lifeCycleDriver.generate("EntityFactoryAllocator", model);
			} else if (factory.category.equals("SERVICE")) {
				lifeCycleDriver.generate("ServiceFactory", model);
				lifeCycleDriver.generate("ServiceFactoryAllocator", model);
			} else if (factory.category.equals("SESSION")) {
				lifeCycleDriver.generate("SessionFactory", model);
				lifeCycleDriver.generate("SessionFactoryAllocator", model);
			} else if (factory.category.equals("UNMANAGED_ENTITY")) {
				lifeCycleDriver.generate("UnmanagedEntityFactory", model);
				lifeCycleDriver.generate("UnmanagedEntityFactoryAllocator", model);
			} else  {
				throw new Exception("Unkown category '"+ factory.category + 
						"' for <factory-impl> " + factory.classname);
			}
		}
	}

	private void generateLifeCycleCpp(Repository model) 
		throws Exception
	{
		// LifeCycle generator
		AcceleoDriver lifeCycleDriver = new AcceleoDriver(
				com.thalesgroup.cdmw.codegenerator.lifecycle.cpp.LifeCycle.emtFile);
		lifeCycleDriver.setOutputDir(outDir);

		for (FactoryImplData factory : cdmwFile.getFactoriesData()) {
			// prepare Acceleo Service with all info
    		// for this factory-impl
			cdmwFile.prepareLifeCycleServices(factory.classname, model);
			
			// generate ServantProvider if required
			if (factory.hasServantProvider) {
				lifeCycleDriver.generate("ServantProvider_hpp", model);
				lifeCycleDriver.generate("ServantProvider_cpp", model);
			}
			
			// generate appropriate Factory
			if (factory.category.equals("ENTITY")) {
				lifeCycleDriver.generate("EntityFactory_hpp", model);
				lifeCycleDriver.generate("EntityFactory_cpp", model);
				lifeCycleDriver.generate("EntityFactoryAllocator_hpp", model);
			} else if (factory.category.equals("SERVICE")) {
				lifeCycleDriver.generate("ServiceFactory_hpp", model);
				lifeCycleDriver.generate("ServiceFactory_cpp", model);
				lifeCycleDriver.generate("ServiceFactoryAllocator_hpp", model);
			} else if (factory.category.equals("SESSION")) {
				lifeCycleDriver.generate("SessionFactory_hpp", model);
				lifeCycleDriver.generate("SessionFactory_cpp", model);
				lifeCycleDriver.generate("SessionFactoryAllocator_hpp", model);
			} else if (factory.category.equals("UNMANAGED_ENTITY")) {
				lifeCycleDriver.generate("UnmanagedEntityFactory_hpp", model);
				lifeCycleDriver.generate("UnmanagedEntityFactory_cpp", model);
				lifeCycleDriver.generate("UnmanagedEntityFactoryAllocator_hpp", model);
			} else  {
				throw new Exception("Unkown category '"+ factory.category + 
						"' for <factory-impl> " + factory.classname);
			}
			
		}
	}
	
	
	private void generateDatastores(Repository model) 
		throws Exception
	{
		// prepare Acceleo Service with all info for <datastore-types>
		cdmwFile.prepareDatastoreTypesService(model);

		// generate only if some <datastore-type> have been found
		if (DatastoreTypesServices.datastoreTypes.size() > 0) {
			System.out.println("Generate Datastores TypeTraits ");

			if (lang.equals("java")) {
				generateDatastoresJava(model);
			} else if (lang.equals("cpp")) {
				generateDatastoresCpp(model);
			} else {
				throw new Exception("Unkown language: " + lang);
			}
		}
	}

	private void generateDatastoresJava(Repository model)
		throws Exception
	{
		// No datastore code generated for Java
	}
	
	private void generateDatastoresCpp(Repository model)
		throws Exception
	{
		// generator for GlobalDataStores file
    	AcceleoDriver globalDataStoresDriver = new AcceleoDriver(GlobalDataStores.emtFile);
    	globalDataStoresDriver.setOutputDir(outDir);
    	// generate GlobalDataStores file
    	globalDataStoresDriver.generate("GlobalDataStores", model);

    	// generator for DataStoreTypeTraits files
    	AcceleoDriver dataStoreTypeTraitsDriver = new AcceleoDriver(DataStoreTypeTraits.emtFile);
    	dataStoreTypeTraitsDriver.setOutputDir(outDir);

    	for (DatastoreTypeData datastore : DatastoreTypesServices.datastoreTypes) {
    		// set current datastore for file generation
    		DatastoreTypesServices.currentDatastoreTypename = datastore.typename;
    		
    		// generate DataStoreTypeTraits file for current datastore
    		dataStoreTypeTraitsDriver.generate("DataStoreTypeTraits", model);
    	}
	}
	

    private void generatePrograms(Repository model) 
		throws Exception
	{
		if (lang.equals("java")) {
			generateProgramsJava(model);
		} else if (lang.equals("cpp")) {
			generateProgramsCpp(model);
		} else {
			throw new Exception("Unkown language: " + lang);
		}
	}
    
    private void generateProgramsJava(Repository model) 
    	throws Exception
    {
    	// get all programs names
    	List<String> programs = cdmwFile.getProgramsFilenames();
    	
    	// generator for CdmwInit file
    	AcceleoDriver cdmwInitDriver = new AcceleoDriver(
    			com.thalesgroup.cdmw.codegenerator.cdmwinit.java.CdmwInit.emtFile);
    	cdmwInitDriver.setOutputDir(outDir);
    	
    	// generator for init DTD file
    	AcceleoDriver initDTDDriver = new AcceleoDriver(InitDTD.emtFile);
    	initDTDDriver.setOutputDir(outDir);
    	
    	for (String program : programs) {
    		// prepare Acceleo Service with all info
    		// for this program
			cdmwFile.prepareProgramServices(program, model);
			
			// generate Cdmw init code
			System.out.println("Generate Cdmw init code for program " + program);
			cdmwInitDriver.generate("CdmwInit", model);
			initDTDDriver.generate("InitDTD", model);
		}
	}
    
    private void generateProgramsCpp(Repository model) 
		throws Exception
	{
    	// get all programs names
    	List<String> programs = cdmwFile.getProgramsFilenames();
    	
    	// generator for CdmwInit file
    	AcceleoDriver cdmwInitDriver = new AcceleoDriver(
    			com.thalesgroup.cdmw.codegenerator.cdmwinit.cpp.CdmwInit.emtFile);
    	cdmwInitDriver.setOutputDir(outDir);
    	
    	// generator for init DTD file
    	AcceleoDriver initDTDDriver = new AcceleoDriver(InitDTD.emtFile);
    	initDTDDriver.setOutputDir(outDir);

    	for (String program : programs) {
    		// prepare Acceleo Service with all info
    		// for this program
			cdmwFile.prepareProgramServices(program, model);
			
			// generate Cdmw init code
			System.out.println("Generate Cdmw init code for program " + program);
			cdmwInitDriver.generate("CdmwInit", model);
			initDTDDriver.generate("InitDTD", model);
		}

	}
    
    
    /**
     * Do all required generations (IDL2, ecore file, container code and Cdmw_init)
     * @throws Exception if an error occurs
     */
    public void doGeneration() 
    	throws Exception
    {
    	if (validateOnly) {
    		// do nothing
    		System.out.println("XML file is valid");
    		return;
    	}
    	
    	if (!genEcoreOnly || genIdl2Only) {
    		// generate IDL2 files
    		generateIDL2();
    	}
    	
    	if (!genIdl2Only || genEcoreOnly) {
    		// create dummy.idl including all idl files
    		File dummy = createDummyIDL();
    		
    		// prepare options for container code generator
    		List<String> options = new Vector<String>();
    		options.add("--lang="+lang);
    		options.add("-O"+outDir);
    		for (String inclDir : idlSearchDirs) {
    			options.add("-I"+inclDir);
			}
    		if (preprocFlags.length() > 0) {
    			String[] flags = preprocFlags.split("\\p{Space}+");
    			for (String flag : flags) {
        			options.add(flag);
				}
    		}
    		for (HomeData homeData : homesData.values()) {
    			options.add("--home=" + homeData);
			}
    		if (genEcoreOnly) {
    			options.add("--save-model");
    		}
    		if (cdmwFile.hasContainerInterception()) {
    			System.setProperty(CommonServices.CONTAINER_INTERCEPTOR_PROPERTY, "true");
    		}
    		
    		options.add(dummy.getAbsolutePath());
    		
    		// create CCM container generator
    		Generator gen = new Generator();
    		System.out.println("Call CCM generator with options:\n"+options);
    		gen.parseOptions(options);
    		
    		// import idl (ecore model if saved if required)
    		if (!gen.doImport()) {
    			throw new Exception("Error importing IDL!");
    		}
    		
    		// delete dummy file    		
    		//dummy.delete();
    		
    		if (!genEcoreOnly) {
    			// generate CCM container code
    			gen.doGeneration();
    		
    			// generate LifeCycle code
    			generateLifeCycle(gen.getModel());
    			
	    		// generate Datastores code
    			generateDatastores(gen.getModel());
	    		
	    		// generate programs code
	    		generatePrograms(gen.getModel());
    		}
    	}
    }
    
    

	/**
     * @param args
     */
    public static void main(String[] args) {

        if (args.length == 0) {
            printHelp();
            return;
        }

        try {
            Main m = new Main();

            List<String> argList = new Vector<String>();
            // copy all args in list
            for (int i = 0; i < args.length; i++) {
            	// Because of MPC, C++ makefile cannot use
            	// "--xml=file" option, but only "--xml file" option.
            	// We need to convert such option here.
            	if (args[i].equals("--xml")) {
            		argList.add(args[i] + "=" + args[++i]);
            	} else {
            		argList.add(args[i]);
            	}
			}
            /*
            for (String arg : args) {
                argList.add(arg);
            }*/
            
            // parse command line
            argList = m.parseOptions(argList);
            
            if (!argList.isEmpty()) {
                throw new Exception("Unrecognized options: " + argList);
            }
            
            // check output directory
            m.checkOutDir();
            
            // load xml file
            m.loadXMLFile();
            
            // start generation
            m.doGeneration();
            
            
            
        } catch (Exception e) {
        	e.printStackTrace();
            System.err.println("\n" + e.getMessage());
            System.err.println("");
            System.exit(1);
        }
    }


}
