
header {
/*
 * COPYRIGHT NOTICE 
 *
 * The information contained herein and which follows is proprietary
 * information and is the property of PrismTech Corporation.  This
 * information is not to be divulged, released, copied, reproduced, or
 * conveyed to unauthorized personnel, companies, or other institutions
 * without the direct and expressed approval in writing of PrismTech
 * Corporation.
 *
 * DISCLAIMER
 * 
 * Neither PrismTech Limited nor PrismTech Corporation ("PrismTech")
 * are under any obligation to support the use of this code which is
 * provided without warranty of any kind. PrismTech shall (except as
 * otherwise prohibited by law) have no liability whatsoever in relation
 * to its use.
 */

package com.prismtech.idt.imports.idl.internal.importer;

import java.util.*;

import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.MultiStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.emf.ecore.*;

import com.prismtech.idt.idlmm.*;
import com.prismtech.idt.imports.idl.importer.*;

import com.prismtech.idt.imports.idl.parser.IdlParserAST;
}

class IdlTreeParser extends TreeParser;

options {
    importVocab=Idl;
    buildAST = true; //?
    ASTLabelType = "IdlParserAST";
}
{
   // ---------members
   private IdlmmBuilder m_builder;

   /**
    * This is a list of {@link IStatus} objects that contain the details of any
    * problems encountered during the {@link #parse(IdlParserAST) parse}
    */
   private final StatusHolder m_statusHolder = new StatusHolder();

   // ---------Java methods

   public IdlTreeParser(IIdlSourceComposer sourceAttrComposer, List<IIdlPragmaHandler> pragmaHandlers)
   {
      m_builder = new IdlmmBuilder(m_statusHolder, sourceAttrComposer, pragmaHandlers);
      tokenNames = _tokenNames;      
      this.getASTFactory().setASTNodeClass (IdlParserAST.class);
   } 
    
   /**
    * Returns the first object that was encountered when tree parsing was
    * started. For example, if {@link #specification(IdlParserAST)} is called as the
    * initial rule, then this will return a {@link Repository}.
    */
   public <T extends EObject> T getRootObject (Class<T> type)
   {
      return m_builder.getRootObject (type);
   }

   /**
    * Calls the {@link #specification(IdlParserAST)} rule to parse the specified {@link IdlParserAST}.
    * Any problems encountered will be returned in the {@link IStatus}
    * 
    * @return The details of any warnings or errors found during the
    *         parse/populate
    */
   public IStatus parse (IdlParserAST ast)
   {
      m_statusHolder.reset ();      
      try
      {
         specification (ast);
      }
      catch (RecognitionException e)
      {
         reportError(e);
      }
      return getStatus ();
   }

   /**
    * Removes the '_' escape character from ESCAPED_IDENT tokens.
    *
    * @return The modified token
    */
   private IdlParserAST removeEscape (IdlParserAST ast)
   {
      String text = ast.getText ();
      ast.setText (text.substring (1));
      return ast;
   }

   /**
    * 
    */
   public IStatus getStatus ()
   {
      List<IStatus> status = m_statusHolder.getStatus(); 
      switch (status.size ())
      {
         case 0:
         {
            return new Status (IStatus.OK, IdlImporterPlugin.PLUGIN_ID, 0, "", null);
         }
         case 1:
         {
            return status.get (0);
         }
         default:
         {
            MultiStatus st = new MultiStatus (IdlImporterPlugin.PLUGIN_ID, 0, "", null);
            for (IStatus s : status)
            {
               st.add (s);
            }
            return st;
         }
      }
   }

   @Override
   public void reportError (RecognitionException exception)
   {
      m_statusHolder.addStatusItem (new IdlImporterStatus(exception));
   }

   @Override
   public void reportError (String message)
   {
      m_statusHolder.addStatusItem (new IdlImporterStatus (this, IStatus.ERROR, message, null));
   }
}
//---------- ANTLR part -----------------------------
//----------------------------------------------------
specification :
 #(i1:SPECIFICATION
  {
   this.m_builder.startRepository (i1);
  }
  definition_list
  {
   this.m_builder.endRepository ();
  }
 );
 
//------------------------- definition
definition
 :
 interface_dcl
 | module
 | type_dcl
 | const_dcl
 | pragma_dcl
 | exception_dcl
 | value
 | event
 | component_dcl
 | home_dcl
 | type_id_dcl
 | type_prefix_dcl
 
;
//------------------------- definition list
definition_list :
 (
 {
  this.m_builder.startDefinition();
 }
    definition
    {
     this.m_builder.endDefinition();
    }
 )*
;
//----------------------------------------------------------- Module
module :
 #(module_literal:LITERAL_module (i1:IDENT | i2:ESCAPED_IDENT)
 {
     IdlParserAST ast = (i1 != null) ? i1 : this.removeEscape (i2);
  this.m_builder.startModule (ast, module_literal);
 }
 definition_list
 {
  this.m_builder.endModule();
 }
 );
 
//---------------------------   type_decl
type_dcl :
 (struct_type) => struct_type
 | (union_type) => union_type
 | typedef_type
 | enum_type
 | native_decl
 | constr_forward_decl
;

//-------------------------- type_spec
type_spec
 {
  this.m_builder.startTypeSpec();
 }
 :   simple_type_spec
 |   constr_type_spec
 {
  this.m_builder.endTypeSpec();
 }
;
//--------------------------- simple_type_spec
simple_type_spec
 {
  this.m_builder.startSimpleTypeSpec();
 }
 :   base_type_spec
 |   template_type_spec
 |   scoped_name 
 {
  this.m_builder.endSimpleTypeSpec();
 }
;
//-------------------------- base_type_spec - parsing simple types
base_type_spec 
 //parse types which start with "long" token
 : long_integer_type   
    | integer_type
 | floating_pt_type
 | char_type  
 | wide_char_type  
 | boolean_type 
 | octet_type
 | any_type
 | object_type
    | value_base_type
;
//------------------------- template_type_spec
template_type_spec
 : string_type
    |  wide_string_type 
//  | fixed_pt_type  //n.r. for the DCPS
 | sequence_type
;
//----------------------------------------- constr_type_spec
constr_type_spec
 {
  this.m_builder.startConstrTypeSpec();
 }
 
 :   struct_type
 |   union_type
 |   enum_type
 
 {
  this.m_builder.startConstrTypeSpec();
 }
 ;
 
//-------------------------------------------------------- CONST_DECL
const_dcl:
 {
  this.m_builder.startConstType();
 } 
 
 #(i1:LITERAL_const const_type (i2:IDENT | i3:ESCAPED_IDENT)
  {
      IdlParserAST ast = (i2 != null) ? i2 : this.removeEscape (i3);
   this.m_builder.constMetaPoint(ast, i1);
  }
  
 const_or_scoped_decl
 )
 {
  this.m_builder.endConstType();
 } 

;

const_type
    : long_integer_type
 | integer_type
 | char_type
 | wide_char_type
 | boolean_type
 | floating_pt_type
 | string_type
 | wide_string_type
// | fixed_pt_const_type
 | scoped_name
 | octet_type
 ;

//---------------------------------------------------------SCOPED NAME 
scoped_name_list
 :    scoped_name (scoped_name)*
 ;

scoped_name
 {
  this.m_builder.startScopedName();
  StringBuilder sb = new StringBuilder();
  boolean isMultiName = false;
  List<IdlParserAST> astList = new ArrayList<IdlParserAST>();
 }

 : (#(i1:IDENT (i2:SCOPEOP)? 
 {
  astList.add(i1); 
  if (i2 != null)
  {
   sb.append("::");
   astList.add(i2);
  }
  sb.append(i1.getText());

 }
 ((i4:IDENT | i5:ESCAPED_IDENT) 
 {
     IdlParserAST ast = (i4 != null) ? i4 : this.removeEscape (i5);
  sb.append("::");
  sb.append(ast.getText());
  astList.add(ast);
 }
 )*)
 |
 #(i6:ESCAPED_IDENT (i7:SCOPEOP)? 
 {
     IdlParserAST ast = this.removeEscape (i6);
  astList.add(ast); 
  if (i7 != null)
  {
   sb.append("::");
   astList.add(i7);
  }
  sb.append(ast.getText());

 }
 ((i9:IDENT | i10:ESCAPED_IDENT) 
 {
     IdlParserAST ast = (i9 != null) ? i9 : this.removeEscape (i10);
  sb.append("::");
  sb.append(ast.getText());
  astList.add(ast);
 }
 )*))
  
 {
  this.m_builder.endScopedName(sb, astList.toArray(new IdlParserAST[astList.size()]));
 }
;

//---------------------------------------------------------- "SIMPLE TYPES"
//-------------------------- long int
long_integer_type
    : (#(LITERAL_long LITERAL_double)) => long_double_type //long double
    | (LITERAL_long LITERAL_long) => signed_longlong_int //long long
    | (LITERAL_long) => signed_long_int

 //parse types which start with "unsigned" token
   | (LITERAL_unsigned LITERAL_long LITERAL_long) => unsigned_longlong_int
   | (LITERAL_unsigned LITERAL_long) => unsigned_long_int
   | (LITERAL_unsigned LITERAL_short) => unsigned_short_int
;
//-------------------------- float
floating_pt_type
 : i1:LITERAL_float
 {
  this.m_builder.startFloatType(i1);
 }
 
 {
  this.m_builder.endFloatType();
 }
 
 | i2:LITERAL_double
 {
  this.m_builder.startDoubleType(i2);
 }
 
 {
  this.m_builder.endDoubleType();
 }
;
//-------------------------- 
integer_type
 :  signed_int
;
//-------------------------- 
signed_int
 :  signed_short_int
 ;
//-------------------------- 
signed_short_int
 :  i1:LITERAL_short 
 {
  this.m_builder.startShortType(i1);
 }
 
 {
  this.m_builder.endShortType();
 }
 ;
//-------------------------- 
signed_long_int
 :  i1:LITERAL_long 
 {
  this.m_builder.startLongType(i1);
 }
 
 {
  this.m_builder.endLongType();
 }
 ;
//-------------------------- 
signed_longlong_int
 :  i1:LITERAL_long i2:LITERAL_long
 {
  this.m_builder.startLongLongType(i1, i2); 
 }
 
 {
  this.m_builder.endLongLongType();
 }
;
//-------------------------- 
unsigned_short_int
 :  i1:LITERAL_unsigned i2:LITERAL_short 
 {
  this.m_builder.startUShortType(i1, i2);
 }
 
 {
  this.m_builder.endUShortType();
 }
;
//-------------------------- 
unsigned_long_int
 :  i1:LITERAL_unsigned i2:LITERAL_long 
 {
  this.m_builder.startULongType(i1, i2);
 }
 
 {
  this.m_builder.endULongType();
 }
;
//-------------------------- 
unsigned_longlong_int
 : i1:LITERAL_unsigned i2:LITERAL_long i3:LITERAL_long
 {
  this.m_builder.startULongLongType(i1, i2, i3);
 }
 
 {
  this.m_builder.endULongLongType();
 }
;
//-------------------------- 
char_type
 : i1:LITERAL_char
 {
  this.m_builder.startCharType(i1);
 }
 
 {
  this.m_builder.endCharType();
 }
;
//-------------------------- 
wide_char_type
 : i1:LITERAL_wchar
 {
  this.m_builder.startWCharType(i1);
 }
 
 {
  this.m_builder.endWCharType();
 }
;
//-------------------------- 
boolean_type
 : i1:LITERAL_boolean
 {
  this.m_builder.startBooleanType(i1);
 }
 
 {
  this.m_builder.endBooleanType();
 }
;
//-------------------------- 
octet_type
 : i1:LITERAL_octet
 {
  this.m_builder.startOctetType(i1);
 }
 
 {
  this.m_builder.endOctetType();
 }
;
//-------------------------- 
any_type
 : i1:LITERAL_any
 {
  this.m_builder.startAnyType(i1);
 }
 
 {
  this.m_builder.endAnyType();
 }
;
//-------------------------- 
object_type
 : i1:LITERAL_Object
 {
  this.m_builder.startObjectType(i1); 
 }
 
 {
  this.m_builder.endObjectType();
 }
;
//---------------------------
value_base_type
 : i1:LITERAL_ValueBase
 {
  this.m_builder.startValueBaseType(i1); 
 }
 
 {
  this.m_builder.endValueBaseType(); 
 }
;
//---------------------------
long_double_type
 : #(i1:LITERAL_long i2:LITERAL_double)
 {
  this.m_builder.startLongDoubleType(i1, i2); 
 }
 
 {
  this.m_builder.endLongDoubleType(); 
 }
;
//------------------------------------------------------------string
string_type
 :
 #(i1:LITERAL_string  
  {
   this.m_builder.startStringType(i1);
  }
  
  (const_or_scoped_decl)?
 )
 {
  this.m_builder.endStringType(); 
 }
;
//------------------------------------------------------------wstring
wide_string_type
 :
 #(i1:LITERAL_wstring 
  {
   this.m_builder.startWStringType(i1);
  }
 (const_or_scoped_decl)?)
 {
  this.m_builder.endWStringType(); 
 }
;

//------------------------------------------------------------- END of "SIMPLE TYPES" 
//------------------------------------------------------------- FIXED - n.r. for the DCPS
/*fixed_pt_type :
  #(LITERAL_fixed i1:positive_int_const i2:positive_int_const)
  {
   this.m_builder.startFixedType(i1,i2);
   }
   (i3:IDENT | i4:ESCAPED_IDENT)
   {
      IdlParserAST ast = (i3 != null) ? i3 : this.removeEscape (i4);
    this.m_builder.startDeclarator(i2); 
    this.m_builder.endDeclarator(); 
   }
   {
    this.m_builder.endFixedType();
   }
;*/

//---------------------------------------------------------------------- SEQUENCE
sequence_type
 {
  this.m_builder.startSequenceType();
 }
 :
 #(i1:LITERAL_sequence simple_type_spec
  {
  }
 
  {
  this.m_builder.sequenceMetaPoint(i1);
  }
 
  (const_or_scoped_decl)?
  {
   this.m_builder.endSequenceType();
  }
 )
 
;
//---------------------------------------------------------------------- TYPEDEF
typedef_type:
 {
  this.m_builder.startTypedefType();
 }
 #(i1:LITERAL_typedef type_spec           
 {
  this.m_builder.typedefMetaPoint(i1);
 }
 declarators)
 
 {
  this.m_builder.endTypedefType();
 }
;

//----------------------------------------------------------------------STRUCT
struct_type:
 #(struct_literal:LITERAL_struct (i1:IDENT | i2: ESCAPED_IDENT)
 {
     IdlParserAST ast = (i1 != null) ? i1 : this.removeEscape (i2);
  this.m_builder.startStruct(ast, struct_literal);
 }
 
 member_list
 
 {
  this.m_builder.endStruct();
 }
);
//------------------------------------------------
member_list:
   (member)+
;

//--------------------------------------- member
member:
 {
  this.m_builder.startMember();
 }
 type_spec declarators semi:SEMI
 {
  this.m_builder.endMember(semi); 
 }
;


//-------------------------------------------------------------------- UNION
union_type:
 #(literal_union:LITERAL_union (i2:IDENT | i3:ESCAPED_IDENT)
 {
     IdlParserAST ast = (i2 != null) ? i2 : this.removeEscape (i3);
  this.m_builder.startUnionType(ast, literal_union);
 }
 switch_type_spec switch_body
 )
 
 {
  this.m_builder.endUnionType();
 }
;

//--------------------------------------------------------------------- ENUM
enum_type:
 #(literal_enum:LITERAL_enum (i1:IDENT | i2:ESCAPED_IDENT)
 {
     IdlParserAST ast = (i1 != null) ? i1 : this.removeEscape (i2);
  this.m_builder.startEnumType(ast, literal_enum);
 }

 enumerator_list

 )
 {
  this.m_builder.endEnumType();
 }
;

enumerator_list
 :    enumerator (enumerator)*
;

enumerator:
 {
  this.m_builder.startEnumerator();
 }
 (i1:IDENT | i2:ESCAPED_IDENT)
 {
     IdlParserAST ast = (i1 != null) ? i1 : this.removeEscape (i2);
  this.m_builder.endEnumerator(ast);
 }
;

//-------------------------------------------------------------------- switch_type_spec
switch_type_spec
{
 this.m_builder.startSwitchTypeSpec();
}
 : t1:(long_integer_type
 | integer_type
 | char_type
 | boolean_type
 | enum_type
 | scoped_name)
{
 this.m_builder.endSwitchTypeSpec(t1);
}
;
//---------------------------------------------------------------------- switch_body
switch_body
 {
  this.m_builder.startSwitchBody();
 }
 :   case_stmt_list
 {
  this.m_builder.endSwitchBody();
 }
;

//-------------------------------------------------------------------- case_stmt_list
case_stmt_list
 :
 (
 {
  this.m_builder.startUnionMember();
 }
 case_stmt
 {
  this.m_builder.endUnionMember();
 }
 )+
 ;

//-------------------------------------------------------------------- case_stmt
case_stmt
 :   
 (
  #(i1:LITERAL_case (case_stmt)? const_or_scoped_decl
  {
   this.m_builder.startCaseStmt(i1);
  }
  (element_spec)?)
 
 |
 #(i2:LITERAL_default
 {
  this.m_builder.startCaseStmt(i2);
 }
 element_spec
 )
 )
 
 {
  this.m_builder.endCaseStmt();
 }
 
;
 
 
//---------------------------------------------------------------------- element_spec
element_spec
 {
  this.m_builder.startElementSpec();
 }
 :   type_spec declarator
 {
  this.m_builder.endElementSpec();
 }
;

//-------------------------------------------------------------------------- VALUE

value :
    (value_dcl) => value_dcl
    | (value_abs_dcl) => value_abs_dcl
    | (value_box_dcl) => value_box_dcl
    | value_forward_dcl
;

value_dcl:
    #(value_literal:LITERAL_valuetype (cst:LITERAL_custom)? (id1:IDENT | id2:ESCAPED_IDENT) 
    {
        IdlParserAST id = (id1 != null) ? id1 : this.removeEscape (id2);
     this.m_builder.startValueDcl(cst, id, value_literal);
    }
    (value_inheritance_spec)? LCURLY (value_element)* RCURLY
    )
    {
        this.m_builder.endValueDcl();
    }
;

value_abs_dcl:
    #(value_literal:LITERAL_valuetype LITERAL_abstract (id1:IDENT | id2:ESCAPED_IDENT)  
    {
        IdlParserAST id = (id1 != null) ? id1 : this.removeEscape (id2);
        this.m_builder.startValueAbsDcl(id, value_literal);
    }
    (value_inheritance_spec)? LCURLY (export_dcl)* RCURLY
    )
    {
        this.m_builder.endValueAbsDcl();
    }
;

value_box_dcl:
    #(value_literal:LITERAL_valuetype (id1:IDENT | id2:ESCAPED_IDENT) type_spec)
    {
        IdlParserAST id = (id1 != null) ? id1 : this.removeEscape (id2);
     this.m_builder.endValueBoxDcl(id, value_literal);
    }
;

value_forward_dcl:
    #(value_literal:LITERAL_valuetype
    {
        this.m_builder.startValueForwardDcl();
    }
    (abs:LITERAL_abstract)? (id1:IDENT | id2:ESCAPED_IDENT)
    {
        IdlParserAST id = (id1 != null) ? id1 : this.removeEscape (id2);
        this.m_builder.endValueForwardDcl(abs, id, value_literal);
    }
    )
;

value_inheritance_spec:
    (value_bases)? (value_supported_interfaces)?
;

value_bases:
    {
     this.m_builder.startValueBases();  
    }
    #(COLON (trn:LITERAL_truncatable)? scoped_name_list)
    {
     this.m_builder.endValueBases(trn);  
    }
;

value_supported_interfaces:
    {
     this.m_builder.startValueSupports();  
    }
    #(LITERAL_supports scoped_name_list)
    {
     this.m_builder.endValueSupports();  
    }
;

value_element:
    export_dcl
    | state_member
    | init_dcl
;

state_member:
 {
  this.m_builder.startMember();
 }
    acc:access type_spec declarators semi:SEMI
 {
  this.m_builder.endMember(acc);
 }
;

access:
    LITERAL_private 
    | LITERAL_public
;

init_dcl:
    #(LITERAL_factory (id1:IDENT | id2:ESCAPED_IDENT) 
    {
        IdlParserAST id = (id1 != null) ? id1 : this.removeEscape (id2);
        this.m_builder.startInitDcl(id);
    }
    (init_param_decls)? (raises_expr)?)
    {
        this.m_builder.endInitDcl();
    }
;

init_param_decls:
    init_param_decl (init_param_decl)*
;

init_param_decl:
    {
        this.m_builder.startInitParamDcl();
    }
    LITERAL_in
    param_type_spec
 dec:simple_declarator
    {
        this.m_builder.endInitParamDcl(dec);
    }
;

//-------------------------------------------------------------------------- INTERFACE

interface_dcl:
 (interface_def) => interface_def 
 | interface_forward_dcl
 ;
 
 
interface_forward_dcl:
 #(interface_literal:LITERAL_interface
  {
   this.m_builder.startInterfaceForwardDecl();
  }
    (i1:interf_modifier_decl)? (i2:IDENT | i3:ESCAPED_IDENT)
  {
      IdlParserAST ast = (i2 != null) ? i2 : this.removeEscape (i3);
   this.m_builder.endInterfaceForwardDecl(i1, ast, interface_literal);
  }
 )
;

interface_def
:
 #(interface_literal:LITERAL_interface 
     (i1:interf_modifier_decl)?
     (i2:IDENT | i3:ESCAPED_IDENT)
    {
        IdlParserAST ast = (i2 != null) ? i2 : this.removeEscape (i3);
     this.m_builder.startInterfaceType(i1, ast, interface_literal);
    }
  
    interface_header LCURLY interface_body RCURLY
  )
  {
     this.m_builder.endInterfaceType();
  }
;

interf_modifier_decl:
 LITERAL_abstract
 | LITERAL_local
;

//---------------------------------------------------- interface_header
interface_header :
{
 this.m_builder.startInterfaceHeader();  
}

 (interf_inheritance_spec)?

{
 this.m_builder.endInterfaceHeader();
}
;

//--------------------------------------------------inheritance_spec
interf_inheritance_spec:
{
 this.m_builder.startInterfaceInherSpec();  
}

 #(COLON scoped_name_list)
  
{
 this.m_builder.endInterfaceInherSpec();  
}

;


//--------------------------------------------------- interface_body
interface_body :
 {
  this.m_builder.startInterfaceBody();
 }

 (export_dcl)*
 
 {
  this.m_builder.endInterfaceBody();
 }  
;

export_dcl
 {
  this.m_builder.startExportDcl();
 }
:

   (
      type_dcl
      | const_dcl
      | exception_dcl
      | attr_dcl
      | op_dcl
      | type_id_dcl
      | type_prefix_dcl
      | pragma_dcl
 )
 {
  this.m_builder.endExportDcl();
 }
;

//----------------------------------------------------------------------- INTERFACE TYPES
//---------------------------------------------------------- type_id_dcl
type_id_dcl:
{
 this.m_builder.startTypeIdDcl();
}
 #(i1:LITERAL_typeid scoped_name string_literal)
{
 this.m_builder.endTypeIdDcl(i1);
}
;

//---------------------------------------------------------- type_prefix_dcl
type_prefix_dcl:
{
 this.m_builder.startTypePrefixDcl();
}

 #(i1:LITERAL_typeprefix scoped_name string_literal)

{
 this.m_builder.endTypePrefixDcl(i1);
}
;

//------------------------------------------------ native_decl
native_decl:
 {
  this.m_builder.startNativeDecl();
 }  

  #(i1:LITERAL_native i2:simple_declarator)

 {
 this.m_builder.endNativeDecl(i1, i2);
 }  
;

//-------------------------------------------------- constr_forward_decl
constr_forward_decl:
 struct_forward_decl
 | union_forward_decl
;

//-------------------------------------------------- struct_forward_decl
struct_forward_decl:
 {
  this.m_builder.startStructForwardDecl();
 }

  #(LITERAL_struct (i1:IDENT | i2:ESCAPED_IDENT))

 {
     IdlParserAST ast = (i1 != null) ? i1 : this.removeEscape (i2);
  this.m_builder.endStructForwardDecl(ast);
 }
;

//-------------------------------------------------- union_forward_decl
union_forward_decl:
 {
  this.m_builder.startUnionForwardDecl();
 }

 #(LITERAL_union (i1:IDENT | i2:ESCAPED_IDENT))

 {
     IdlParserAST ast = (i1 != null) ? i1 : this.removeEscape (i2);
  this.m_builder.endUnionForwardDecl(ast);
 }
;

//--------------------------------------------------------------- operation
op_dcl
{
 this.m_builder.startOperationDecl();
}
:

 (#(i1:IDENT
  (i2:op_attribute)?
  {
   this.m_builder.endOperationDeclAttribute(i1, i2);
  }
  
  op_type_spec
  parameter_dcls
     (raises_expr)?
     (context_expr)?
  
 )
 |
 #(i3:ESCAPED_IDENT
  (i4:op_attribute)?
  {
   this.m_builder.endOperationDeclAttribute(this.removeEscape (i3), i4);
  }
  
  op_type_spec
  parameter_dcls
     (raises_expr)?
     (context_expr)?
  
 ))
{
 this.m_builder.endOperationDecl();
}
;

//------------------------------------------------------------- op_attribute
op_attribute
{
 this.m_builder.startOpAttribute();
}
 : LITERAL_oneway
{
 this.m_builder.endOpAttribute();
}
;

//------------------------------------------------------------- op_type_spec
op_type_spec
{
 this.m_builder.startOpTypeSpec();
}
:  
 (
  param_type_spec
  | void_literal
 )
 
{
 this.m_builder.endOpTypeSpec();
}
;

void_literal:
{
 this.m_builder.startVoidLiteral();
}
 i1:LITERAL_void
{
 this.m_builder.endVoidLiteral(i1);
}
;

//---------------------------------------------------------------- ATTRIBUTE
attr_dcl:
{
 this.m_builder.startAttrDecl();
}
 attr_spec
{
 this.m_builder.endAttrDecl();
}
; 
 
//----------------------------------------------------- attr_spec
attr_spec:
{
 this.m_builder.startAttrSpec();
}

 (#(LITERAL_attribute LITERAL_readonly))=> readonly_attr_spec
 | normal_attr_spec
{
 this.m_builder.endAttrSpec();
}
; 
//----------------------------------------------------- readonly_attr_spec
readonly_attr_spec:
 #(i1:LITERAL_attribute i2:LITERAL_readonly 
  {
   this.m_builder.startReadonlyAttrSpec(i1, i2);
  }
 
  param_type_spec
  {
   this.m_builder.endAttrParamSpec();
  }
  readonly_attr_declarator
 )
{
 this.m_builder.endReadonlyAttrSpec();
}
; 

//-------------------------------------------------------- normal_attr_spec
normal_attr_spec
:
 #(i1:LITERAL_attribute
  {
   this.m_builder.startNormalAttrSpec(i1);
  }

  param_type_spec
  {
   this.m_builder.endAttrParamSpec();
  }
  
  attr_declarator
 )
{
 this.m_builder.endNormalAttrSpec();
}
; 

//----------------------------------------------------------- context_expr
context_expr:
 #(LITERAL_context
  {
   this.m_builder.startContextExpr();
  }  
  (string_literal)+
 )
{
 this.m_builder.endContextExpr();
}
;
//------------------------------------------------------------- parameter_decls
parameter_dcls
:
 (param_dcl_list)?
;

//------------------------------------------------------------- param_dcl_list
param_dcl_list:
 (param_dcl)+ 
;

//------------------------------------------------------------- param_dcl
param_dcl
{
 this.m_builder.startParamDcl();
 IdlParserAST i7 = null, i8 = null;
}
:
 (
  #(i1:LITERAL_in
   param_type_spec
   i4:simple_declarator
  )
  |
  #(i2:LITERAL_out
   param_type_spec
   i5:simple_declarator
  )
  |
  #(i3:LITERAL_inout
   param_type_spec
   i6:simple_declarator
  )
 )
 //select node to pass from i1, i2, i3 and i4, i5, i6 nodes
 {
  if (i1 != null)
  {
   i7 = i1;
   i8 = i4;
  }
  else if (i2 != null)
  {
   i7 = i2;
   i8 = i5;
  }
  else if (i3 != null)
  {
   i7 = i3;
   i8 = i6;
  }
 }
 
{
 this.m_builder.endParamDcl(i7, i8);
}
;

//----------------------------------------------------------------------- exception
exception_dcl:
 #(exception_literal:LITERAL_exception (i1:IDENT | i2:ESCAPED_IDENT)
 
  {
  IdlParserAST ast = (i1 != null) ? i1 : this.removeEscape (i2);
  this.m_builder.startExceptionDecl(ast, exception_literal);
  }
  
  exception_member_list
 
{
 this.m_builder.endExceptionDecl();
}
);

//------------------------------------------------ 
exception_member_list:
   (member)*
;

//----------------------------------------------------------------------- PREPROCESSOR
//------------------------------------------------------------------------ PRAGMA
pragma_dcl
 :
 {
  this.m_builder.startPragmaType();
 }
 //#(LITERAL_pragma declarators)
 i1:PRAGMA
 {
  
  this.m_builder.endPragmaType(i1);
 }
;

 
//------------------------------------------------------ param_type_spec
param_type_spec:
{
 this.m_builder.startParamTypeSpec();
}

 base_type_spec
 | string_type
 | wide_string_type
 | scoped_name

{
 this.m_builder.endParamTypeSpec();
}
; 
 
//--------------------------------------------------------------------- DECLARATORS
declarators
 :   declarator (declarator)*
;
//---------------------------
declarator
 : (#(IDENT fixed_array_size)) => complex_declarator
 | (#(ESCAPED_IDENT fixed_array_size)) => complex_declarator
    | simple_declarator
;
//--------------------------- 
simple_declarator
 :   (i1:IDENT | i2:ESCAPED_IDENT)
 {
     IdlParserAST ast = (i1 != null) ? i1 : this.removeEscape (i2);
  this.m_builder.startSimpleDeclarator(ast);
    
  this.m_builder.endSimpleDeclarator();  
 }
;
//-------------------------------------------------------- complex_decl
complex_declarator
 :   array_declarator
;
//-------------------------------------------------------- array_decl
array_declarator
 :   
 (#(i1:IDENT 
  {
   this.m_builder.startArrayDeclarator(i1);
  }
  
  (const_or_scoped_decl)+
 )
 |
 #(i2:ESCAPED_IDENT 
  {
   this.m_builder.startArrayDeclarator(this.removeEscape (i2));
  }
  
  (const_or_scoped_decl)+
 ))
 
 {
  this.m_builder.endArrayDeclarator();
 }
;
//------------------------------------------------------- readonly_attr_declarator
readonly_attr_declarator:
{
 this.m_builder.startReadonlyAttrDeclarator();
}
 simple_declarator
 (
  raises_expr  
  | (simple_declarator)*
 )
  
{
 this.m_builder.endReadonlyAttrDeclarator();
}
; 

//------------------------------------------------------- attr_declarator
attr_declarator:
{
 this.m_builder.startAttrDeclarator();
}
 simple_declarator
 (
  (#(LITERAL_getraises scoped_name_list)) => attr_raises_expr
  |(#(LITERAL_setraises scoped_name_list)) => attr_raises_expr   
  |(simple_declarator)*
 )
{
 this.m_builder.endAttrDeclarator();
}
;

//------------------------------------------------------------- raises_expr
raises_expr:
{
 this.m_builder.startRaisesExpr();
}

 #(i1:LITERAL_raises scoped_name_list) 
{
 this.m_builder.endRaisesExpr(i1);
}
;

//------------------------------------------------------------- attr_raises_exp
attr_raises_expr
:   (get_excep_expr)?
    (set_excep_expr)?
;

//-----------------------------------------------
get_excep_expr
{
 this.m_builder.startGetRaisedException();
}
:
   #(i1:LITERAL_getraises scoped_name_list)
   
{
 this.m_builder.endGetRaisedException(i1);
}   
;

set_excep_expr
{
 this.m_builder.startSetRaisedException();
}
:
   #(i1:LITERAL_setraises scoped_name_list)
   
{
 this.m_builder.endSetRaisedException(i1);
}
   
;

//------------------------------------------------
fixed_array_size
 :   const_or_scoped_decl
 ;
//-------------------------------------------------------------- END of "DECLARATORS"

//--------------------------------------------------------------- "CONSTANTS"
positive_int_const
 :    
 {
  this.m_builder.startPositiveIntConst();
 }
 
 const_exp
 
 {
  this.m_builder.endPositiveIntConst();
 }
 ;
 
opt_pos_int
 : (positive_int_const)?
 ;
//=================================================================================== EXPRESSIONS
const_or_scoped_decl
  : 
  i1:(const_exp
  | scoped_name
  )   
;
 
const_exp
 : or_expr
 | xor_expr
 | and_expr
 | rshift_expr
 | lshift_expr
 | plus_expr
 | minus_expr
 | tilde_expr
 | star_expr
 | div_expr
 | mod_expr
 | literal 
 | scoped_name //test
;

//------------------------------------------------------ or
or_expr:
 #(i1:OR 
  {
   this.m_builder.startOrExpr(i1);
  }

  const_exp const_exp
 )
 
 {
  this.m_builder.endOrExpr();
 }
;

//------------------------------------------------------ xor
xor_expr:
 #(i1:XOR
  {
   this.m_builder.startXorExpr(i1);
  }
  const_exp const_exp)
 {
  this.m_builder.endXorExpr();
 }
;

//------------------------------------------------------ and
and_expr:
 #(i1:AND 
  {
   this.m_builder.startAndExpr(i1);
  }
 const_exp const_exp)
 
 {
  this.m_builder.endAndExpr();
 }
;

//------------------------------------------------------ rshift
rshift_expr:
 #(i1:RSHIFT 
  {
   this.m_builder.startRShiftExpr(i1);
  }
 const_exp const_exp)
 
 {
  this.m_builder.endRShiftExpr();
 }
;

//------------------------------------------------------ lshift
lshift_expr:
 #(i1:LSHIFT 
  {
   this.m_builder.startLShiftExpr(i1);
  }
 const_exp const_exp)
 
 {
  this.m_builder.endLShiftExpr();
 }
;


//------------------------------------------------------ plus
plus_expr:
 #(i1:PLUS 
  {
   this.m_builder.startPlusExpr(i1);
  }
 const_exp 
 (const_exp)? //plus operator can be unary
 )
 
 {
  this.m_builder.endPlusExpr();
 }
;

//------------------------------------------------------ minus
minus_expr:
 #(i1:MINUS 
  {
   this.m_builder.startMinusExpr(i1);
  }
 const_exp 
 (const_exp)? //minus operator can be unary
 )
 
 {
  this.m_builder.endMinusExpr();
 }
;

//------------------------------------------------------ tilde
tilde_expr:
 #(i1:TILDE
  {
   this.m_builder.startTildeExpr(i1);
  }
 const_exp 
 //const_exp
 )
 {
  this.m_builder.endTildeExpr();
 }
;

//------------------------------------------------------ star
star_expr:
 #(i1:STAR
  {
   this.m_builder.startStarExpr(i1);
  }
 const_exp const_exp)
 
 {
  this.m_builder.endStarExpr();
 }
;


//------------------------------------------------------ div
div_expr:
 #(i1:DIV
  {
   this.m_builder.startDivExpr(i1);
  }
  const_exp const_exp)
  
 {
  this.m_builder.endDivExpr();
 }
;

//------------------------------------------------------ mod
mod_expr:
 #(i1:MOD
  {
   this.m_builder.startModExpr(i1);
  }
 const_exp const_exp)
 
 {
  this.m_builder.endModExpr();
 }
;

//--------------------------------------------------------"LITERALS"
literal
 :   integer_literal
 |   string_literal
 |   wide_string_literal
 |   character_literal
 |   wide_character_literal
 |   fixed_pt_literal
 |   floating_pt_literal
 |   boolean_literal
 ;

boolean_literal
 {
  this.m_builder.startBooleanLiteral();
 }
 :   i1: boolean_literal_inner
 {
  this.m_builder.endBooleanLiteral(i1);
 }
 ;

boolean_literal_inner
 :   "TRUE"
 |   "FALSE"
;


integer_literal
 {
  this.m_builder.startIntegerLiteral();
 }
 :   i1: integer_literal_inner
 {
  this.m_builder.endIntegerLiteral(i1);
 }
;

integer_literal_inner:
 INT | OCTAL | HEX
;

string_literal
 {
  this.m_builder.startStringLiteral();
 }
 : 
 (  
  i1:string_literal_inner
  {
   this.m_builder.endStringLiteral(i1);
  }
 )+
;

string_literal_inner
 :  STRING_LITERAL
 ;


wide_string_literal
 {
  this.m_builder.startWStringLiteral();
 }
 :   i1: wide_string_literal_inner
 {
  this.m_builder.endWStringLiteral(i1);
 }
;

wide_string_literal_inner
 :  (WIDE_STRING_LITERAL)+
 ;
 
character_literal
 {
  this.m_builder.startCharLiteral();
 }
 :   i1: character_literal_inner
 {
  this.m_builder.endCharLiteral(i1);
 }
;
character_literal_inner
 :  CHAR_LITERAL
 ;


wide_character_literal
 {
  this.m_builder.startWCharLiteral();
 }
 :   i1: wide_character_literal_inner
 {
  this.m_builder.endWCharLiteral(i1);
 }
;
wide_character_literal_inner
 :  WIDE_CHAR_LITERAL
 ;

fixed_pt_literal
 :  FIXED
 ;

floating_pt_literal
 {
  this.m_builder.startFloatLiteral();
 }
 :   i1: floating_pt_literal_inner
 {
  this.m_builder.endFloatLiteral(i1);
 }
;
 
floating_pt_literal_inner
 :   FLOAT
 ; 
 
//-------------------------------------------------------------------------- EVENT

event :
    (event_dcl) => event_dcl
    | (event_abs_dcl) => event_abs_dcl
    | event_forward_dcl
;

event_dcl:
    #(event_literal:LITERAL_eventtype (cst:LITERAL_custom)? id:IDENT 
    {
     this.m_builder.startEventDcl(cst, id, event_literal);
    }
    (value_inheritance_spec)? LCURLY (value_element)* RCURLY
    )
    {
        this.m_builder.endEventDcl();
    }
;

event_abs_dcl:
    #(event_literal:LITERAL_eventtype LITERAL_abstract id:IDENT 
    {
        this.m_builder.startEventAbsDcl(id, event_literal);
    }
    (value_inheritance_spec)? LCURLY (export_dcl)* RCURLY
    )
    {
        this.m_builder.endEventAbsDcl();
    }
;

event_forward_dcl:
    #(value_literal:LITERAL_valuetype
    {
        this.m_builder.startEventForwardDcl();
    }
    (abs:LITERAL_abstract)? id:IDENT
    {
        this.m_builder.endEventForwardDcl(abs, id, value_literal);
    }
    )
;


//-------------------------------------------------------------------------- COMPONENT

component_dcl:
 (component_def) => component_def 
 | component_forward_dcl
 ;
 
 
component_forward_dcl:
 #(component_literal:LITERAL_component
  {
   this.m_builder.startComponentForwardDecl();
  }
    i1:IDENT
  {
   this.m_builder.endComponentForwardDecl(i1, component_literal);
  }
 )
;

component_def
:
 #(component_literal:LITERAL_component 
     i1:IDENT
    {
     this.m_builder.startComponentType(i1, component_literal);
    }
  
    component_header LCURLY component_body RCURLY
  )
  {
     this.m_builder.endComponentType();
  }
;


//---------------------------------------------------- component_header
component_header :
{
 this.m_builder.startComponentHeader();  
}

 (component_inheritance_spec)?
 (component_supported_spec)?

{
 this.m_builder.endComponentHeader();
}
;

//--------------------------------------------------inheritance_spec
component_inheritance_spec:
{
 this.m_builder.startComponentInherSpec();  
}

 #(COLON scoped_name)
  
{
 this.m_builder.endComponentInherSpec();  
}
;

//--------------------------------------------------supported_spec
component_supported_spec:
{
 this.m_builder.startComponentSupportSpec();  
}

 #(LITERAL_supports scoped_name_list)
  
{
 this.m_builder.endComponentSupportSpec();  
}
;


//--------------------------------------------------- component_body
component_body :
 {
  this.m_builder.startComponentBody();
 }

 (component_export_dcl)*
 
 {
  this.m_builder.endComponentBody();
 }  
;

component_export_dcl
 {
  this.m_builder.startComponentExportDcl();
 }
:

   (
        provides_dcl
      | uses_dcl
      | emits_dcl
      | publishes_dcl
      | consumes_dcl
      | attr_dcl
 )
 {
  this.m_builder.endComponentExportDcl();
 }
;

provides_dcl:
 #(provides_literal:LITERAL_provides
  {
   this.m_builder.startProvidesDcl(provides_literal);
  }

  param:param_type_spec
  {
   this.m_builder.endProvidesParamSpec(param);
  }
  
  simple_declarator
  {
   this.m_builder.endProvidesDcl();
  }
 )
;

uses_dcl:
 (#(LITERAL_uses LITERAL_multiple))=> multiple_uses_dcl
 | single_uses_dcl
; 

multiple_uses_dcl:
 #(uses_literal:LITERAL_uses LITERAL_multiple
    {
     this.m_builder.startUsesDcl(uses_literal, true);
    }

  param:param_type_spec
  {
   this.m_builder.endUsesParamSpec(param);
  }
  
  simple_declarator
  {
      this.m_builder.endUsesDcl();
    }    
 )
;

single_uses_dcl:
 #(uses_literal:LITERAL_uses
    {
     this.m_builder.startUsesDcl(uses_literal, false);
    }

  param:param_type_spec
  {
   this.m_builder.endUsesParamSpec(param);
  }
  
  simple_declarator
  {
      this.m_builder.endUsesDcl();
    }    
 )
;

emits_dcl:
 #(emits_literal:LITERAL_emits
    {
     this.m_builder.startEmitsDcl(emits_literal);
    }

  param:param_type_spec
  {
   this.m_builder.endEmitsParamSpec(param);
  }
  
  simple_declarator
     {
      this.m_builder.endEmitsDcl();
    }    
 )
;

publishes_dcl:
 #(publishes_literal:LITERAL_publishes
    {
     this.m_builder.startPublishesDcl(publishes_literal);
    }

  param:param_type_spec
  {
   this.m_builder.endPublishesParamSpec(param);
  }
  
  simple_declarator
     {
      this.m_builder.endPublishesDcl();
    }    
 )
;

consumes_dcl:
 #(consumes_literal:LITERAL_consumes
    {
     this.m_builder.startConsumesDcl(consumes_literal);
    }

  param:param_type_spec
  {
   this.m_builder.endConsumesParamSpec(param);
  }
  
  simple_declarator
     {
      this.m_builder.endConsumesDcl();
    }    
 )
;



//-------------------------------------------------------------------------- HOME

home_dcl:
 #(home_literal:LITERAL_home 
     i1:IDENT
    {
     this.m_builder.startHomeType(i1, home_literal);
    }
  
    home_header LCURLY home_body RCURLY
  )
  {
     this.m_builder.endHomeType();
  }
;


//---------------------------------------------------- home_header
home_header :
{
 this.m_builder.startHomeHeader();  
}

 (home_inheritance_spec)?
 (home_supported_spec)?
 (home_manages_spec)
 (home_primary_key_spec)?

{
 this.m_builder.endHomeHeader();
}
;

//--------------------------------------------------inheritance_spec
home_inheritance_spec:
{
 this.m_builder.startHomeInherSpec();  
}

 #(COLON scoped_name)
  
{
 this.m_builder.endHomeInherSpec();  
}
;

//--------------------------------------------------supported_spec
home_supported_spec:
{
 this.m_builder.startHomeSupportSpec();  
}

 #(LITERAL_supports scoped_name_list)
  
{
 this.m_builder.endHomeSupportSpec();  
}
;

//--------------------------------------------------manages_spec
home_manages_spec:
{
 this.m_builder.startHomeManagesSpec();  
}

 #(LITERAL_manages scoped_name)
  
{
 this.m_builder.endHomeManagesSpec();  
}
;

//--------------------------------------------------primary_key_spec
home_primary_key_spec:
{
 this.m_builder.startHomePrimaryKeySpec();  
}

 #(LITERAL_primarykey scoped_name)
{
 this.m_builder.endHomePrimaryKeySpec();  
}
;


//--------------------------------------------------- home_body
home_body :
 {
  this.m_builder.startHomeBody();
 }

 (home_export_dcl)*
 
 {
  this.m_builder.endHomeBody();
 }  
;

home_export_dcl
 {
  this.m_builder.startHomeExportDcl();
 }
:
   (
        type_dcl
      | const_dcl
      | exception_dcl
      | attr_dcl
      | op_dcl
      | type_id_dcl
      | type_prefix_dcl
      | pragma_dcl
      | factory_dcl
      | finder_dcl
 )
 {
  this.m_builder.endHomeExportDcl();
 }
;

factory_dcl:
    #(LITERAL_factory id:IDENT 
    {
        this.m_builder.startFactoryDcl(id);
    }
    (init_param_decls)? (raises_expr)?)
    {
        this.m_builder.endFactoryDcl();
    }
;

finder_dcl:
    #(LITERAL_finder id:IDENT 
    {
        this.m_builder.startFinderDcl(id);
    }
    (init_param_decls)? (raises_expr)?)
    {
        this.m_builder.endFinderDcl();
    }
;
