/*
 * Copyright (c) 2005-2008 Obeo.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Obeo - initial API and implementation
 */

package fr.obeo.acceleo.gen.template.eval;

import fr.obeo.acceleo.gen.phantom.SyncElement;

import java.text.NumberFormat;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.acceleo.engine.AcceleoEngineMessages;
import org.acceleo.engine.template.impl.spec.IEvaluable;
import org.acceleo.template.core.ASTNode;
import org.acceleo.tools.factory.EFactory;
import org.acceleo.tools.factory.ETools;
import org.acceleo.tools.log.EvalLog;
import org.acceleo.tools.strings.Int2;
import org.acceleo.tools.strings.TextSearch;
import org.eclipse.emf.common.util.Enumerator;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EcoreFactory;
import org.eclipse.emf.ecore.EcorePackage;

/**
 * Node of generation. Result of a template evaluation.
 * <p>
 * Encapsulates a generation value :
 * <li> EObject</li>
 * <li> ENodeList</li>
 * <li> String</li>
 * <li> boolean</li>
 * <li> int</li>
 * <li> double</li>
 * <li> null</li>
 * 
 * @author www.obeo.fr
 * 
 */
public class ENode implements SyncElement, Comparable<ENode> {
	/**
	 * Used to create empty ENode.
	 */
	public static final EObject EMPTY = null;

	/** Constant for EObject string type. */
	protected static final String T_EOBJECT = "EObject"; //$NON-NLS-1$

	/** Constant for ENodeList string type. */
	protected static final String T_ENODELIST = "ENodeList"; //$NON-NLS-1$

	/** Constant for String string type. */
	protected static final String T_STRING = "String"; //$NON-NLS-1$

	/** Constant for boolean string type. */
	protected static final String T_BOOLEAN = "boolean"; //$NON-NLS-1$

	/** Constant for int string type. */
	protected static final String T_INT = "int"; //$NON-NLS-1$

	/** Constant for double string type. */
	protected static final String T_DOUBLE = "double"; //$NON-NLS-1$

	/** Constant for null string type. */
	protected static final String T_NULL = "null"; //$NON-NLS-1$

	/**
	 * Constant for ENodeCastException localization message.
	 */
	private static final String ENODE_CAST_EXCEPTION_MESSAGE = "ENode.ENodeCastExceptionMessage"; //$NON-NLS-1$

	/**
	 * Value of this node.
	 * <p>
	 * It's an EObject, an ENodeList, a String, a Boolean, an Integer, or null.
	 */
	protected Object value;

	/**
	 * Type of this node.
	 * <p>
	 * <li>type == T_EObject : the value is an EObject</li>
	 * <li>type == T_ENodeList : the value is an ENodeList</li>
	 * <li>type == T_String : the value is a String</li>
	 * <li>type == T_boolean : the value is a Boolean</li>
	 * <li>type == T_int : the value is an Integer</li>
	 * <li>type == T_double : the value is a Double</li>
	 * <li>type == T_null : the value is null</li>
	 */
	protected String type;

	/**
	 * Indicates if it's a containment node.
	 */
	protected boolean containment = true;

	/**
	 * Class of this node.
	 * <p>
	 * <li>type == T_EObject : the class is EObject.class</li>
	 * <li>type == T_ENodeList : the class is ENodeList.class</li>
	 * <li>type == T_String : the class is String.class</li>
	 * <li>type == T_boolean : the class is boolean.class</li>
	 * <li>type == T_int : the class is int.class</li>
	 * <li>type == T_double : the class is double.class</li>
	 * <li>type == T_null : the class is null</li>
	 */
	protected Class<?> typeClass;

	/**
	 * It's the model element used to generate this node.
	 */
	protected EObject containerEObject;

	/**
	 * It's the template element used to generate this node.
	 */
	protected ASTNode containerTemplateElement;

	/**
	 * Correspondences between the model and the text value, can be null.
	 */
	protected TextModelMapping textModelMapping;

	/**
	 * Correspondences between the template element and the text value, can be
	 * null.
	 */
	protected TextTemplateElementMapping textTemplateElementMapping;

	/**
	 * Evaluation errors and warnings.
	 */
	protected EvalLog log = new EvalLog();

	/**
	 * Indicates if it's an optional node.
	 */
	protected boolean optional;

	/**
	 * Activates model/text mapping.
	 */
	private boolean synchronize = true;

	/**
	 * Creates a root node. The value is an EObject. The model element used to
	 * generate this node is the element itself because the parent node is
	 * unknown.
	 * <p>
	 * Precondition : object != EMPTY && object != null
	 * <p>
	 * Remark : eContainer() == getValue().
	 * 
	 * @param object
	 *            is the value
	 * @deprecated
	 */
	@Deprecated
	public ENode(EObject object) {
		this(object, (IEvaluable) null, true);
	}

	/**
	 * Creates a root node. The value is an EObject. The model element used to
	 * generate this node is the element itself because the parent node is
	 * unknown.
	 * <p>
	 * Precondition : object != EMPTY && object != null
	 * <p>
	 * Remark : eContainer() == getValue().
	 * 
	 * @param object
	 *            is the value
	 * @param element
	 *            is the template element
	 * @param synchronize
	 *            activates model/text mapping
	 */
	public ENode(EObject object, IEvaluable element, boolean synchronize) {
		containerEObject = object;
		containerTemplateElement = element;
		this.synchronize = synchronize;
		init(object, T_EOBJECT, EObject.class);
	}

	/**
	 * Creates a child node. The value is an EObject. The model element used to
	 * generate this node is the same element as for the parent.
	 * <p>
	 * Remark : eContainer() == parent.eContainer().
	 * 
	 * @param object
	 *            is the value
	 * @param parent
	 *            is the parent node
	 */
	public ENode(EObject object, ENode parent) {
		init(parent);
		init(object, T_EOBJECT, EObject.class);
	}

	/**
	 * Creates a child node. The value is an EObject. The model element used to
	 * generate this node is the parent.
	 * <p>
	 * Precondition : parent != EMPTY && parent != null
	 * <p>
	 * Equivalent : new ENode(object, new ENode(parent)).
	 * 
	 * @param object
	 *            is the value
	 * @param parent
	 *            is the parent object
	 * @param element
	 *            is the template element
	 * @param synchronize
	 *            activates model/text mapping
	 */
	public ENode(EObject object, EObject parent, IEvaluable element, boolean synchronize) {
		this(object, new ENode(parent, element, synchronize));
	}

	/**
	 * Creates a child node. The value is an ENodeList. The model element used
	 * to generate this node is the same element as for the parent.
	 * <p>
	 * Remark : eContainer() == parent.eContainer().
	 * 
	 * @param l
	 *            is the value
	 * @param parent
	 *            is the parent node
	 */
	public ENode(ENodeList l, ENode parent) {
		init(parent);
		init(l, T_ENODELIST, ENodeList.class);
	}

	/**
	 * Creates a child node. The value is a String. The model element used to
	 * generate this node is the same element as for the parent.
	 * <p>
	 * Remark : eContainer() == parent.eContainer().
	 * 
	 * @param s
	 *            is the value
	 * @param parent
	 *            is the parent node
	 */
	public ENode(String s, ENode parent) {
		init(parent);
		init(s, T_STRING, String.class);
	}

	/**
	 * Creates a child node. The value is a boolean. The model element used to
	 * generate this node is the same element as for the parent.
	 * <p>
	 * Remark : eContainer() == parent.eContainer().
	 * 
	 * @param b
	 *            is the value
	 * @param parent
	 *            is the parent node
	 */
	public ENode(boolean b, ENode parent) {
		init(parent);
		init(new Boolean(b), T_BOOLEAN, boolean.class);
	}

	/**
	 * Creates a child node. The value is an int. The model element used to
	 * generate this node is the same element as for the parent.
	 * <p>
	 * Remark : eContainer() == parent.eContainer().
	 * 
	 * @param i
	 *            is the value
	 * @param parent
	 *            is the parent node
	 */
	public ENode(int i, ENode parent) {
		init(parent);
		init(new Integer(i), T_INT, int.class);
	}

	/**
	 * Creates a child node. The value is a double. The model element used to
	 * generate this node is the same element as for the parent.
	 * <p>
	 * Remark : eContainer() == parent.eContainer().
	 * 
	 * @param i
	 *            is the value
	 * @param parent
	 *            is the parent node
	 */
	public ENode(double i, ENode parent) {
		init(parent);
		init(new Double(i), T_DOUBLE, double.class);
	}

	/**
	 * Getter for the value of the node.
	 * 
	 * @return the value of this node : an EObject, an ENodeList, a String, a
	 *         Boolean, an Integer, or null.
	 */
	public Object getValue() {
		if (isString()) {
			return value.toString();
		} else {
			return value;
		}
	}

	/**
	 * Getter for the type of this node.
	 * 
	 * @return the type of this node : "EObject", "ENodeList", "String",
	 *         "boolean", "int", "double", or null.
	 */
	public String getType() {
		return type;
	}

	/**
	 * Getter for the class of this node.
	 * 
	 * @return the class of this node : EObject.class, ENodeList.class,
	 *         String.class, boolean.class, int.class, double.class, or null.
	 */
	public Class<?> getTypeClass() {
		return typeClass;
	}

	/**
	 * Getter for the model element used to generate this node.
	 * 
	 * @return the model element used to generate this node
	 */
	public EObject getContainerEObject() {
		return containerEObject;
	}

	/**
	 * Getter for the template element used to generate this node.
	 * 
	 * @return the template element used to generate this node
	 */
	public ASTNode getContainerTemplateElement() {
		return containerTemplateElement;
	}

	/**
	 * Getter for correspondences between the model and the text value.
	 * 
	 * @return correspondences between the model and the text value, can be null
	 */
	public TextModelMapping getTextModelMapping() {
		return textModelMapping;
	}

	/**
	 * Getter for correspondences between the template element and the text
	 * value.
	 * 
	 * @return correspondences between the template element and the text value,
	 *         can be null
	 */
	public TextTemplateElementMapping getTextTemplateElementMapping() {
		return textTemplateElementMapping;
	}

	/**
	 * Getter for evaluation errors and warnings.
	 * 
	 * @return evaluation errors and warnings
	 */
	public EvalLog log() {
		return log;
	}

	/**
	 * Updates container and log.
	 * 
	 * @param parent
	 *            is the parent node
	 */
	protected void init(ENode parent) {
		if (parent != null) {
			synchronize = parent.synchronize;
			containerEObject = parent.containerEObject;
			containerTemplateElement = parent.containerTemplateElement;
			log.getAll(parent.log, false);
		}
	}

	/**
	 * Updates value, type, and class.
	 * 
	 * @param val
	 *            is the value
	 * @param t
	 *            is the type of the value
	 * @param clazz
	 *            is the class of the value
	 */
	protected void init(Object val, String t, Class<?> clazz) {
		if (val != null) {
			if (val instanceof String) {
				this.value = new StringBuilder((String) val);
				this.type = t;
				this.typeClass = clazz;
			} else {
				this.value = val;
				this.type = t;
				this.typeClass = clazz;
			}
		} else {
			this.value = null;
			this.type = T_NULL;
			this.typeClass = null;
		}
		if (synchronize) {
			this.textModelMapping = createTextModelMapping();
			this.textTemplateElementMapping = createTextTemplateElementMapping();
		}
	}

	/**
	 * Updates text-model mapping, for String only.
	 * 
	 * @return a new mapping if the value is a String, null if not
	 */
	protected TextModelMapping createTextModelMapping() {
		if (isString() && containerEObject != null) {
			final TextModelMapping mapping = new TextModelMapping(containerEObject, !synchronize);
			mapping.shift(((StringBuilder) value).length());
			return mapping;
		} else {
			return null;
		}
	}

	/**
	 * Updates text-template mapping, for String only.
	 * 
	 * @return a new mapping
	 */
	protected TextTemplateElementMapping createTextTemplateElementMapping() {
		if (isString() && containerTemplateElement != null) {
			final TextTemplateElementMapping mapping = new TextTemplateElementMapping(containerTemplateElement, !synchronize);
			mapping.shift(((StringBuilder) value).length());
			return mapping;
		} else {
			return null;
		}
	}

	/**
	 * Is the node contains an EObject?
	 * 
	 * @return true if it's an EObject, false if not
	 */
	public boolean isEObject() {
		return type == T_EOBJECT;
	}

	/**
	 * Is the node contains an ENodeList?
	 * 
	 * @return true if it's an ENodeList, false if not
	 */
	public boolean isList() {
		return type == T_ENODELIST;
	}

	/**
	 * Is the node contains a String.
	 * 
	 * @return true if it's a String, false if not
	 */
	public boolean isString() {
		return type == T_STRING;
	}

	/**
	 * Is the node contains a boolean.
	 * 
	 * @return true if it's a boolean, false if not
	 */
	public boolean isBoolean() {
		return type == T_BOOLEAN;
	}

	/**
	 * Is the node contains a int.
	 * 
	 * @return true if it's an int, false if not
	 */
	public boolean isInt() {
		return type == T_INT;
	}

	/**
	 * Is the node contains a double.
	 * 
	 * @return true if it's a double, false if not
	 */
	public boolean isDouble() {
		return type == T_DOUBLE;
	}

	/**
	 * Is the node contains a null value.
	 * 
	 * @return true if it's null, false if not
	 */
	public boolean isNull() {
		return type == T_NULL;
	}

	/**
	 * Returns the adaptive type for the given class.
	 * 
	 * @param c
	 *            is the class
	 * @return the adaptive type
	 */
	public static Class<?> getAdapterType(Class<?> c) {
		Class<?> res;
		boolean canBeHandled = c == ENode.class || c == EObject.class || c == ENodeList.class;
		canBeHandled = canBeHandled || c == String.class || c == boolean.class || c == int.class;
		canBeHandled = canBeHandled || c == double.class || c == null;
		if (canBeHandled) {
			res = c;
		} else if (EObject.class.isAssignableFrom(c)) {
			res = EObject.class;
		} else if (c == List.class) { // List fix
			res = List.class;
		} else if (c == Collection.class) { // Collection fix
			res = List.class;
		} else {
			res = null;
		}
		return res;
	}

	/**
	 * Returns the adaptive value for the given class.
	 * 
	 * @param c
	 *            is the class
	 * @return the adaptive value
	 * @throws ENodeCastException
	 *             if no adapter could be found.
	 */
	public Object getAdapterValue(Class<?> c) throws ENodeCastException {
		Object adaptiveValue;
		if (c == EObject.class) {
			adaptiveValue = adaptToEObject().getValue();
		} else if (c == ENodeList.class) {
			adaptiveValue = toList().getValue();
		} else if (c == String.class) {
			adaptiveValue = adaptToString().getValue();
		} else if (c == boolean.class) {
			adaptiveValue = adaptToBoolean().getValue();
		} else if (c == int.class) {
			adaptiveValue = adaptToInt().getValue();
		} else if (c == double.class) {
			adaptiveValue = adaptToDouble().getValue();
		} else if (c == null) {
			adaptiveValue = adaptToNull().getValue();
		} else if (c == List.class) { // List fix
			adaptiveValue = toList().getList().asList();
		} else if (c == Collection.class) { // Collection fix
			adaptiveValue = toList().getList().asList();
		} else if (c == ENode.class) {
			adaptiveValue = this;
		} else {
			throw new ENodeCastException(AcceleoEngineMessages.getString("ENode.InvalidAdapterType", new Object[] { c.getClass().getName(), })); //$NON-NLS-1$
		}
		return adaptiveValue;
	}

	/**
	 * Adapt the value to EObject.
	 * 
	 * @return a new ENode casted to EObject
	 * @throws ENodeCastException
	 *             if it can't be casted
	 */
	private ENode adaptToEObject() throws ENodeCastException {
		ENode res;
		if (typeClass == EObject.class) {
			res = this;
		} else if (typeClass == ENodeList.class) {
			if (getList().size() > 0) {
				res = getList().get(0);
			} else {
				res = new ENode(EMPTY, this);
			}
		} else if (typeClass == String.class) {
			final Object result = EcoreFactory.eINSTANCE.createFromString(EcorePackage.eINSTANCE.getEString(), getString());
			if (result instanceof EObject) {
				res = new ENode((EObject) result, this);
			} else {
				res = new ENode(EMPTY, this);
			}
		} else if (typeClass == boolean.class) {
			final Object result = EcoreFactory.eINSTANCE.createFromString(EcorePackage.eINSTANCE.getEBoolean(), toString());
			if (result instanceof EObject) {
				res = new ENode((EObject) result, this);
			} else {
				res = new ENode(EMPTY, this);
			}
		} else if (typeClass == int.class) {
			final Object result = EcoreFactory.eINSTANCE.createFromString(EcorePackage.eINSTANCE.getEInt(), toString());
			if (result instanceof EObject) {
				res = new ENode((EObject) result, this);
			} else {
				res = new ENode(EMPTY, this);
			}
		} else if (typeClass == double.class) {
			Object result;
			result = EcoreFactory.eINSTANCE.createFromString(EcorePackage.eINSTANCE.getEDouble(), value.toString());
			if (result instanceof EObject) {
				res = new ENode((EObject) result, this);
			} else {
				res = new ENode(EMPTY, this);
			}
		} else {
			res = new ENode(EMPTY, this);
		}
		return res;
	}

	/**
	 * Adapt the value to ENodeList.
	 * 
	 * @return a new ENode casted to ENodeList
	 */
	private ENode toList() {
		ENode res;
		if (typeClass == ENodeList.class) {
			res = this;
		} else {
			final ENodeList result = new ENodeList();
			result.add(this);
			res = new ENode(result, this);
		}
		return res;
	}

	/**
	 * Adapt the value to String.
	 * 
	 * @return a new ENode casted to String
	 * @throws ENodeCastException
	 *             if it can't be casted
	 */
	private ENode adaptToString() throws ENodeCastException {
		ENode res;
		if (typeClass == String.class) {
			res = this;
		} else if (typeClass == ENodeList.class) {
			final ENode result = new ENode("", this); //$NON-NLS-1$
			final ENodeIterator it = getList().iterator();
			while (it.hasNext()) {
				result.append(it.next());
			}
			res = result;
		} else {
			res = new ENode(toString(), this);
		}
		return res;
	}

	/**
	 * Adapt the value to Boolean.
	 * 
	 * @return a new ENode casted to Boolean
	 * @throws ENodeCastException
	 *             if it can't be casted
	 */
	private ENode adaptToBoolean() throws ENodeCastException {
		ENode res;
		if (typeClass == EObject.class) {
			res = new ENode(true, this);
		} else if (typeClass == ENodeList.class) {
			res = new ENode(getList().size() > 0, this);
		} else if (typeClass == String.class) {
			res = new ENode("true".equalsIgnoreCase(getString().trim()), this); //$NON-NLS-1$
		} else if (typeClass == boolean.class) {
			res = this;
		} else if (typeClass == int.class) {
			if (getInt() > 0) {
				res = new ENode(true, this);
			} else {
				res = new ENode(false, this);
			}
		} else if (typeClass == double.class) {
			if (getDouble() > 0) {
				res = new ENode(true, this);
			} else {
				res = new ENode(false, this);
			}
		} else {
			res = new ENode(false, this);
		}
		return res;
	}

	/**
	 * Adapt the value to int.
	 * 
	 * @return a new ENode casted to int
	 * @throws ENodeCastException
	 *             if it can't be casted
	 */
	private ENode adaptToInt() throws ENodeCastException {
		ENode res;
		if (typeClass == EObject.class) {
			res = new ENode(1, this);
		} else if (typeClass == ENodeList.class) {
			res = new ENode(getList().size(), this);
		} else if (typeClass == String.class) {
			try {
				res = new ENode(Integer.parseInt(getString().trim()), this);
			} catch (final NumberFormatException e) {
				throw new ENodeCastException(AcceleoEngineMessages.getString("ENode.AdapterNotFound", new Object[] { "Integer", toString() + " [" + getType() + ']', })); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
			}
		} else if (typeClass == boolean.class) {
			if (getBoolean()) {
				res = new ENode(1, this);
			} else {
				res = new ENode(0, this);
			}
		} else if (typeClass == int.class) {
			res = this;
		} else if (typeClass == double.class) {
			res = new ENode(((Double) value).intValue(), this);
		} else {
			return new ENode(0, this);
		}
		return res;
	}

	/**
	 * Adapt the value to double.
	 * 
	 * @return a new ENode casted to double
	 * @throws ENodeCastException
	 *             if it can't be casted
	 */
	private ENode adaptToDouble() throws ENodeCastException {
		ENode res;
		if (typeClass == EObject.class) {
			res = new ENode(1.0, this);
		} else if (typeClass == ENodeList.class) {
			res = new ENode(new Double(getList().size()).doubleValue(), this);
		} else if (typeClass == String.class) {
			try {
				res = new ENode(Double.parseDouble(getString().trim()), this);
			} catch (final NumberFormatException e) {
				throw new ENodeCastException(AcceleoEngineMessages.getString("ENode.AdapterNotFound", new Object[] { "Double", toString() + " [" + getType() + ']', })); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
			}
		} else if (typeClass == boolean.class) {
			if (getBoolean()) {
				res = new ENode(1.0, this);
			} else {
				res = new ENode(0.0, this);
			}
		} else if (typeClass == int.class) {
			res = new ENode(new Double(getInt()).doubleValue(), this);
		} else if (typeClass == double.class) {
			res = this;
		} else {
			res = new ENode(0.0, this);
		}
		return res;
	}

	/**
	 * Adapt the value to null.
	 * 
	 * @return a new empty ENode
	 */
	private ENode adaptToNull() {
		return new ENode(EMPTY, this);
	}

	/**
	 * Getter for the EObject value if it's an EObject.
	 * 
	 * @return the EObject value if it's an EObject
	 * @throws ENodeCastException
	 *             if it isn't an EObject
	 */
	public EObject getEObject() throws ENodeCastException {
		if (isEObject()) {
			return (EObject) value;
		} else {
			throw new ENodeCastException(AcceleoEngineMessages.getString(ENODE_CAST_EXCEPTION_MESSAGE, new Object[] { type, T_EOBJECT, }));
		}
	}

	/**
	 * Getter for the ENodeList value if it's an ENodeList.
	 * 
	 * @return the ENodeList value if it's an ENodeList
	 * @throws ENodeCastException
	 *             if it isn't an ENodeList
	 */
	public ENodeList getList() throws ENodeCastException {
		if (isList()) {
			return (ENodeList) value;
		} else {
			throw new ENodeCastException(AcceleoEngineMessages.getString(ENODE_CAST_EXCEPTION_MESSAGE, new Object[] { type, T_ENODELIST, }));
		}
	}

	/**
	 * Getter for the String value if it's a String.
	 * 
	 * @return the String value if it's a String
	 * @throws ENodeCastException
	 *             if it isn't a String
	 */
	public String getString() throws ENodeCastException {
		if (isString()) {
			return ((StringBuilder) value).toString();
		} else {
			throw new ENodeCastException(AcceleoEngineMessages.getString(ENODE_CAST_EXCEPTION_MESSAGE, new Object[] { type, T_STRING, }));
		}
	}

	/**
	 * Getter for the boolean value if it's a boolean.
	 * 
	 * @return the boolean value if it's a boolean
	 * @throws ENodeCastException
	 *             if it isn't a boolean
	 */
	public boolean getBoolean() throws ENodeCastException {
		if (isBoolean()) {
			return ((Boolean) value).booleanValue();
		} else {
			throw new ENodeCastException(AcceleoEngineMessages.getString(ENODE_CAST_EXCEPTION_MESSAGE, new Object[] { type, T_BOOLEAN, }));
		}
	}

	/**
	 * Getter for the int value if it's an int.
	 * 
	 * @return the int value if it's an int
	 * @throws ENodeCastException
	 *             if it isn't an int
	 */
	public int getInt() throws ENodeCastException {
		if (isInt()) {
			return ((Integer) value).intValue();
		} else {
			throw new ENodeCastException(AcceleoEngineMessages.getString(ENODE_CAST_EXCEPTION_MESSAGE, new Object[] { type, T_INT, }));
		}
	}

	/**
	 * Getter for the double value if it's a double.
	 * 
	 * @return the double value if it's a double
	 * @throws ENodeCastException
	 *             if it isn't a double
	 */
	public double getDouble() throws ENodeCastException {
		if (isDouble()) {
			return ((Double) value).doubleValue();
		} else {
			throw new ENodeCastException(AcceleoEngineMessages.getString(ENODE_CAST_EXCEPTION_MESSAGE, new Object[] { type, T_DOUBLE, }));
		}
	}

	/**
	 * It transforms this node into String.
	 * 
	 * @return the new String value.
	 */
	public String asString() {
		if (isList()) {
			try {
				final ENodeIterator it = getList().iterator();
				init("", T_STRING, String.class); //$NON-NLS-1$
				while (it.hasNext()) {
					final ENode child = it.next();
					append(child);
				}
			} catch (final ENodeCastException e) {
				// Never catch
			}
		} else if (!isString()) {
			final String v = toString();
			init(v, T_STRING, String.class);
		}
		return ((StringBuilder) value).toString();
	}

	/**
	 * Appends an other node to this current node.
	 * <p>
	 * If current is null, it updates current with other node settings.
	 * <p>
	 * If current isn't null, it transforms current into String, it transforms
	 * other into String, and it appends the two node.
	 * 
	 * @param other
	 *            is the other node
	 */
	public void append(ENode other) {
		if (isNull() && !other.isString()) {
			init(other.value, other.type, other.typeClass);
			log.getAll(other.log, false);
		} else if (other.isList()) {
			try {
				final ENodeIterator it = other.getList().iterator();
				while (it.hasNext()) {
					final ENode child = it.next();
					append(child);
				}
			} catch (final ENodeCastException e) {
				// Never catch
			}
		} else {
			if (!other.isString()) {
				other.asString();
			}
			if (other.synchronize) {
				if (other.textModelMapping != null) {
					other.textModelMapping.commit();
				}
				if (other.textTemplateElementMapping != null) {
					other.textTemplateElementMapping.commit();
				}
			}
			if (!isString()) {
				asString();
			}
			final int size = size();
			((StringBuilder) value).append((StringBuilder) other.value);
			if (synchronize) {
				if (textModelMapping != null && other.textModelMapping != null) {
					textModelMapping.from(other.textModelMapping);
				}
				if (textTemplateElementMapping != null && other.textTemplateElementMapping != null) {
					textTemplateElementMapping.from(other.textTemplateElementMapping);
				}
			}
			other.log().shiftPosition(size);
			log.getAll(other.log, false);
		}
	}

	/**
	 * Appends text to this node.
	 * <p>
	 * The text will have the default highlight
	 * TextModelMapping.HIGHLIGHTED_DEFAULT.
	 * 
	 * @param text
	 *            is the text to be added
	 */
	public void append(String text) {
		append(text, TextModelMapping.HIGHLIGHTED_DEFAULT);
	}

	/**
	 * Appends text to this node.
	 * <p>
	 * The text will have the given highlight.
	 * 
	 * @param text
	 *            is the text to be added
	 * @param highlightedType
	 *            is the highlight type.
	 */
	public void append(String text, int highlightedType) {
		if (!isString()) {
			asString();
		}
		((StringBuilder) value).append(text);
		if (synchronize) {
			if (textModelMapping != null) {
				textModelMapping.shift(text.length(), highlightedType);
			}
			if (textTemplateElementMapping != null) {
				textTemplateElementMapping.shift(text.length(), highlightedType);
			}
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean equals(Object other) {
		boolean res;
		if (other != null && other instanceof ENode) {
			try {
				res = equals((ENode) other);
			} catch (final ENodeCastException e) {
				res = false;
			}
		} else {
			if (value == null && other == null) {
				res = true;
			} else if (value == null || other == null) {
				res = false;
			} else {
				res = value.equals(other);
			}
		}
		return res;
	}

	/**
	 * Indicates if nodes are equals.
	 * 
	 * @param other
	 *            is the other node
	 * @return true if nodes are equals
	 * @throws ENodeCastException
	 *             if the types of the nodes are not compatible
	 */
	public boolean equals(ENode other) throws ENodeCastException {
		boolean res;
		if (isNull() && other.isNull()) {
			res = true;
		} else if (isNull() || other.isNull()) {
			res = toString().equals(other.toString());
		} else if (type == other.type) {
			if (isString()) {
				res = value.toString().equals(other.value.toString());
			} else {
				res = value.equals(other.value);
			}
		} else if (other.isBoolean()) {
			res = other.value.equals(getAdapterValue(other.getTypeClass()));
		} else {
			if (value.equals(other.getAdapterValue(getTypeClass()))) {
				res = true;
			} else {
				res = toString().equals(other.toString());
			}
		}
		// It never throws : new ENodeCastException(type + " can't be equals
		// with " + other.type);
		return res;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public int hashCode() {
		return super.hashCode();
	}

	/**
	 * Returns the size of the node.
	 * <li>isEObject() : return 1</li>
	 * <li>isList() : return getList().size()</li>
	 * <li>isString() : return getString().length()</li>
	 * <li>isBoolean() : return 1</li>
	 * <li>isInt() : return 1</li>
	 * <li>isDouble() : return 1</li>
	 * <li>isNull() : return 0</li>
	 * 
	 * @return value size
	 */
	public int size() {
		int res;
		if (isEObject()) {
			res = 1;
		} else if (isList()) {
			res = ((ENodeList) value).size();
		} else if (isString()) {
			res = ((StringBuilder) value).length();
		} else if (isBoolean()) {
			res = 1;
		} else if (isInt()) {
			res = 1;
		} else if (isDouble()) {
			res = 1;
		} else {
			// ASSERT isNull()
			res = 0;
		}
		return res;
	}

	/**
	 * {@inheritDoc}
	 */
	@SuppressWarnings("unchecked")
	public int compareTo(ENode other) {
		int result = 0;
		if (getValue().getClass() == other.getValue().getClass() && getValue() instanceof Comparable && other.getValue() instanceof Comparable) {
			result = ((Comparable) getValue()).compareTo(other.getValue());
		} else {
			result = toString().compareTo(other.toString());
		}
		if (result != 0) {
			return result;
		} else {
			return 1;
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public String toString() {
		String res;
		if (isNull()) {
			res = ""; //$NON-NLS-1$
		} else if (isEObject()) {
			final EObject object = (EObject) value;
			final String buffer = object.toString();
			final int iProperties = buffer.indexOf("("); //$NON-NLS-1$
			if (iProperties > -1) {
				res = ETools.getEClassifierShortPath(object.eClass()) + ' ' + buffer.substring(iProperties);
			} else {
				res = value.toString();
			}
		} else if (isDouble()) {
			res = NumberFormat.getInstance().format(((Double) value).doubleValue());
		} else {
			res = value.toString();
		}
		return res;
	}

	/**
	 * Is the node a containment node or not?
	 * 
	 * @return true if it's a containment node, false if not
	 */
	public boolean isContainment() {
		return containment;
	}

	/**
	 * Set the node as containment node or not.
	 * 
	 * @param containment
	 *            indicates if it's a containment node
	 */
	public void setContainment(boolean containment) {
		this.containment = containment;
	}

	/**
	 * Is the node an optional node or not?
	 * 
	 * @return true if it's an optional node, false if not
	 */
	public boolean isOptional() {
		return optional;
	}

	/**
	 * Set the node as an optional node or not.
	 * 
	 * @param optional
	 *            indicates if it's an optional node
	 */
	public void setOptional(boolean optional) {
		this.optional = optional;
	}

	/**
	 * Indicates if this node is instance of the class whose name is given.
	 * <li>getType() == name => true</li>
	 * <li>isEObject() => EFactory.eInstanceOf(getEObject(),name)</li>
	 * <p>
	 * Samples :
	 * <li>Return true if isString() and name == "String"</li>
	 * <li>An instance of the EObject java.resources.Folder</li>
	 * return true if name equals "Folder" or "resources.Folder".
	 * <li>An instance of the EObject java.resources.Folder return true if name
	 * equals "File" and Folder inherits File.</li>
	 * 
	 * @param name
	 *            is the class name
	 * @return true if this node is instance of the class whose name is given
	 */
	public boolean eInstanceof(String name) {
		boolean res;
		if (getType().equals(name)) {
			res = true;
		} else {
			try {
				if (EFactory.eInstanceOf(getEObject(), name)) {
					res = true;
				} else {
					res = false;
				}
			} catch (final ENodeCastException e) {
				res = false;
			}
		}
		return res;
	}

	/**
	 * It returns a new ENode if value is valid (EObject, ENodeList, String,
	 * Boolean, Integer, or null), else it returns null.
	 * 
	 * @param value
	 *            is the value
	 * @param parent
	 *            is the parent node
	 * @return the new ENode or null
	 */
	public static ENode createTry(Object value, ENode parent) {
		ENode res;
		if (value == null) {
			res = new ENode(ENode.EMPTY, parent);
		} else if (value instanceof ENode) {
			final ENode node = (ENode) value;
			if (node.containerEObject == null && parent.containerEObject != null) {
				final ENode result = new ENode(EMPTY, parent);
				result.append(node);
				res = result;
			} else {
				res = node;
			}
		} else if (value instanceof EObject) {
			if (value instanceof Enumerator) {
				res = new ENode(((Enumerator) value).getLiteral(), parent);
			} else {
				res = new ENode((EObject) value, parent);
			}
		} else if (value instanceof String) {
			res = new ENode((String) value, parent);
		} else if (value instanceof ENodeList) {
			res = new ENode((ENodeList) value, parent);
		} else if (value instanceof Collection) {
			final ENodeList result = new ENodeList();
			for (final Object obj : (Collection<?>) value) {
				final ENode element = createTry(obj, parent);
				if (element != null) {
					result.add(element);
				}
			}
			res = new ENode(result, parent);
		} else if (value instanceof Object[]) {
			final ENodeList result = new ENodeList();
			final Object[] values = (Object[]) value;
			for (final Object element2 : values) {
				final ENode element = createTry(element2, parent);
				if (element != null) {
					result.add(element);
				}
			}
			res = new ENode(result, parent);
		} else if (value instanceof Boolean) {
			res = new ENode(((Boolean) value).booleanValue(), parent);
		} else if (value instanceof Integer) {
			res = new ENode(((Integer) value).intValue(), parent);
		} else if (value instanceof Short) {
			res = new ENode(((Short) value).intValue(), parent);
		} else if (value instanceof Long) {
			res = new ENode(((Long) value).intValue(), parent);
		} else if (value instanceof Double) {
			res = new ENode(((Double) value).doubleValue(), parent);
		} else if (value instanceof Float) {
			res = new ENode(((Float) value).doubleValue(), parent);
		} else if (value instanceof Character) {
			res = new ENode(String.valueOf(value), parent);
		} else {
			res = new ENode(value.toString(), parent);
		}
		return res;
	}

	/**
	 * {@inheritDoc}
	 */
	public Map<String, Int2[]> inOutMapping() {
		if (getTextModelMapping() != null) {
			return getTextModelMapping().uri2positionsSerializableMap();
		} else {
			return new HashMap<String, Int2[]>();
		}
	}

	/**
	 * {@inheritDoc}
	 */
	public Map<String, Int2[]> templateOutMapping() {
		if (getTextTemplateElementMapping() != null) {
			return getTextTemplateElementMapping().uri2positionsSerializableMap();
		} else {
			return new HashMap<String, Int2[]>();
		}
	}

	/**
	 * {@inheritDoc}
	 */
	public Map<Int2, String> outInMapping() {
		if (getTextModelMapping() != null) {
			return getTextModelMapping().position2uriSerializableMap();
		} else {
			return new HashMap<Int2, String>();
		}
	}

	/**
	 * {@inheritDoc}
	 */
	public Map<Int2, String> outTemplateMapping() {
		if (getTextTemplateElementMapping() != null) {
			return getTextTemplateElementMapping().position2uriSerializableMap();
		} else {
			return new HashMap<Int2, String>();
		}
	}

	/**
	 * {@inheritDoc}
	 */
	public Object getContent() {
		return toString();
	}

	/**
	 * String method call.
	 * 
	 * @param name
	 *            is the name of the method to call
	 * @param begin
	 *            is the beginning index, or -1
	 * @param end
	 *            is the ending index, or -1
	 */
	public void stringCall(String name, int begin, int end) {
		if (isList()) {
			try {
				final ENodeIterator it = getList().iterator();
				init(new ENodeList(), T_ENODELIST, ENodeList.class);
				while (it.hasNext()) {
					final ENode child = it.next();
					child.stringCall(name, begin, end);
					getList().add(child);
				}
			} catch (final ENodeCastException e) {
				// Never catch
			}
		} else {
			if (!isString()) {
				final String v = toString();
				init(v, T_STRING, String.class);
			}
			// isString
			isStringCallString(name, begin, end);
		}
	}

	/**
	 * String method call for a string node.
	 * 
	 * @param name
	 *            is the name of the method to call
	 * @param begin
	 *            is the beginning index, or -1
	 * @param end
	 *            is the ending index, or -1
	 */
	private void isStringCallString(String name, int begin, int end) {
		final String text = value.toString();
		if (text.length() > 0) {
			if ("trim".equals(name)) { //$NON-NLS-1$
				final Int2 range = TextSearch.getDefaultSearch().trim(text, 0, text.length());
				if (range.b() > 0 || range.e() < text.length()) {
					range(range);
					value = new StringBuilder(substring(text, range.b(), range.e()));
				}
			} else if ("substring".equals(name)) { //$NON-NLS-1$
				Int2 range;
				if (end > -1) {
					range = new Int2(begin, end);
				} else {
					range = new Int2(begin, text.length());
				}
				if (range.b() > 0 || range.e() < text.length()) {
					range(range);
					value = new StringBuilder(substring(text, range.b(), range.e()));
				}
			} else if ("toLowerCase".equals(name)) { //$NON-NLS-1$
				value = new StringBuilder(text.toLowerCase());
			} else if ("toUpperCase".equals(name)) { //$NON-NLS-1$
				value = new StringBuilder(text.toUpperCase());
			} else if ("toU1Case".equals(name)) { //$NON-NLS-1$
				value = new StringBuilder(Character.toUpperCase(text.charAt(0)) + text.substring(1));
			} else if ("toL1Case".equals(name)) { //$NON-NLS-1$
				value = new StringBuilder(Character.toLowerCase(text.charAt(0)) + text.substring(1));
			} else if ("indentSpace".equals(name) || "indentTab".equals(name)) { //$NON-NLS-1$ //$NON-NLS-2$
				char c;
				if ("indentSpace".equals(name)) { //$NON-NLS-1$
					c = ' ';
				} else {
					c = '\t';
				}

				final Int2[] lines = TextSearch.getDefaultSearch().splitPositionsOf(text, new String[] { "\n" }, false); //$NON-NLS-1$
				indent(lines);
				for (int i = lines.length - 1; i >= 0; i--) {
					((StringBuilder) value).insert(lines[i].b(), c);
				}
			} else if (name != null && name.startsWith("internalIndent:")) { //$NON-NLS-1$
				final String indent = name.substring("internalIndent:".length()); //$NON-NLS-1$
				final Int2[] lines = TextSearch.getDefaultSearch().splitPositionsOf(text, new String[] { "\n" }, false); //$NON-NLS-1$
				for (int i = 0; i < indent.length(); i++) {
					indent(lines);
				}
				for (int i = lines.length - 1; i >= 0; i--) {
					((StringBuilder) value).insert(lines[i].b(), indent);
				}
				stringCall("substring", indent.length(), -1); //$NON-NLS-1$
			}
		}
	}

	/**
	 * Moves the bounds of the errors into the given range.
	 * 
	 * @param range
	 *            the range
	 */
	private void range(Int2 range) {
		if (textModelMapping != null) {
			textModelMapping.range(range);
		}
		if (textTemplateElementMapping != null) {
			textTemplateElementMapping.range(range);
		}
		log.range(range);
	}

	/**
	 * Compute the sub string of s begining at position begin and ending at the
	 * position end.
	 * 
	 * @param s
	 *            the String to work on
	 * @param begin
	 *            the begin position
	 * @param end
	 *            the ending position
	 * @return the sub string of s begining at position begin and ending at the
	 *         position end
	 */
	private String substring(String s, int begin, int end) {
		String res;
		if (begin < 0 || end < 0 || begin >= s.length() || end <= begin) {
			res = ""; //$NON-NLS-1$
		} else if (end >= s.length()) {
			res = s.substring(begin);
		} else {
			res = s.substring(begin, end);
		}
		return res;
	}

	/**
	 * Indent text/model mapping.
	 * 
	 * @param lines
	 *            lines to indent
	 */
	private void indent(Int2[] lines) {
		if (synchronize) {
			if (textModelMapping != null) {
				textModelMapping.indent(lines);
			}
			if (textTemplateElementMapping != null) {
				textTemplateElementMapping.indent(lines);
			}
		}
		log.indent(lines);
	}

	/**
	 * Creates a copy.
	 * 
	 * @return the copy
	 */
	public ENode copy() {
		ENode res;
		if (value == null) {
			res = new ENode(ENode.EMPTY, this);
		} else if (value instanceof EObject) {
			res = new ENode((EObject) value, this);
		} else if (value instanceof StringBuilder) {
			res = new ENode(((StringBuilder) value).toString(), this);
		} else if (value instanceof ENodeList) {
			res = new ENode((ENodeList) value, this);
		} else if (value instanceof Boolean) {
			res = new ENode(((Boolean) value).booleanValue(), this);
		} else if (value instanceof Integer) {
			res = new ENode(((Integer) value).intValue(), this);
		} else if (value instanceof Double) {
			res = new ENode(((Double) value).doubleValue(), this);
		} else {
			res = new ENode(value.toString(), this);
		}
		return res;
	}
}
