/*
 * Copyright (c) 2005-2008 Obeo.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Obeo - initial API and implementation
 */

package fr.obeo.acceleo.gen.template.eval;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Map.Entry;

import org.acceleo.tools.factory.ETools;
import org.acceleo.tools.strings.Int2;
import org.eclipse.emf.ecore.EObject;

/**
 * Mapping between the model and the text value.
 * 
 * @author www.obeo.fr
 * 
 */
public class TextModelMapping {

	/** Default highlight. */
	public static final int HIGHLIGHTED_DEFAULT = -1;

	/** Highlight for the text which is not in the models. */
	public static final int HIGHLIGHTED_STATIC_TEXT = 0;

	/** Highlight for comments. */
	public static final int HIGHLIGHTED_COMMENT = 1;

	/**
	 * The element that compares positions (Int2).
	 */
	private static class InversePosComparator implements Comparator<Int2>, Serializable {

		/**
		 * The default serial version ID.
		 */
		private static final long serialVersionUID = 1L;

		/**
		 * {@inheritDoc}
		 */
		public int compare(Int2 pos0, Int2 pos1) {
			int res;
			if (pos0.b() < pos1.b()) {
				res = 1;
			} else if (pos0.b() > pos1.b()) {
				res = -1;
			} else {
				if (pos0.e() < pos1.e()) {
					res = -1;
				} else if (pos0.e() > pos1.e()) {
					res = 1;
				} else {
					res = 0;
				}
			}
			return res;
		}
	}

	/**
	 * The element that compares positions (Int2).
	 */
	private static class PosComparator implements Comparator<Int2> {
		/**
		 * {@inheritDoc}
		 */
		public int compare(Int2 pos0, Int2 pos1) {
			int res;
			if (pos0.b() < pos1.b()) {
				res = -1;
			} else if (pos0.b() > pos1.b()) {
				res = 1;
			} else {
				if (pos0.e() < pos1.e()) {
					res = 1;
				} else if (pos0.e() > pos1.e()) {
					res = -1;
				} else {
					res = 0;
				}
			}
			return res;
		}
	}

	/**
	 * Current size of the text used for the mapping.
	 */
	protected int shift;

	/**
	 * Indicates if the changes were validated.
	 * <p>
	 * It is impossible to change mappings when commit is true.
	 */
	protected boolean commit;

	/**
	 * An object that maps text position to EObject.
	 */
	protected Map<Int2, EObject> pos2EObject = new TreeMap<Int2, EObject>(new InversePosComparator());

	/**
	 * An object that maps EObject to text position.
	 */
	protected Map<EObject, List<Int2>> eObject2Positions = new HashMap<EObject, List<Int2>>();

	/**
	 * An object that maps EObject to comment text position.
	 */
	protected Map<EObject, List<Int2>> eObject2CommentPositions = new HashMap<EObject, List<Int2>>();

	/**
	 * An object that maps text position to an EObject declaration. It is used
	 * for an open link action.
	 */
	protected Map<Int2, EObject> pos2LinkEObject = new TreeMap<Int2, EObject>(new PosComparator());

	/**
	 * Open link result for the next added text.
	 */
	protected EObject linkEObject;

	/**
	 * Highlight positions.
	 * <p>
	 * <li>highlightedPos[HIGHLIGHTED_STATIC_TEXT] is a list of positions for
	 * the text which is not in the models. </li>
	 * <li>highlightedPos[HIGHLIGHTED_COMMENT] is a list of positions for
	 * comments. </li>
	 */
	@SuppressWarnings("unchecked")
	protected List<Int2>[] highlightedPos = new List[2];

	/**
	 * Default object for text mapping.
	 */
	protected EObject object;

	/**
	 * Mapping between index and model's object.
	 */
	private Map<Integer, EObject> index2EObject = new HashMap<Integer, EObject>();

	/**
	 * Mapping between index and link model's element.
	 */
	private Map<Integer, EObject> index2LinkEObject = new HashMap<Integer, EObject>();

	/**
	 * Constructor.
	 * 
	 * @param object
	 *            is the default object for text mapping
	 * @param freeze
	 *            is used to freeze the mappings, this helps to improve the
	 *            performance
	 */
	public TextModelMapping(EObject object, boolean freeze) {
		this.object = object;
		highlightedPos[0] = new ArrayList<Int2>();
		highlightedPos[1] = new ArrayList<Int2>();
		if (freeze) {
			commit = true;
		} else {
			reset();
		}
	}

	/**
	 * Reset all informations.
	 */
	protected void reset() {
		if (!commit) {
			this.shift = 0;
			pos2EObject.clear();
			index2EObject.clear();
			eObject2Positions.clear();
			eObject2CommentPositions.clear();
			pos2LinkEObject.clear();
			index2LinkEObject.clear();
			for (int i = 0; i < highlightedPos.length; i++) {
				if (highlightedPos[i] != null) {
					highlightedPos[i].clear();
				} else {
					highlightedPos[i] = new ArrayList<Int2>();
				}
			}
		}
	}

	/**
	 * Appends other mappings to current mappings.
	 * 
	 * @param other
	 *            is the other mappings between the model and the text value
	 */
	public void from(TextModelMapping other) {
		if (!commit && other != null) {
			// mapping pos2EObject, eObject2Pos
			for (Entry<Int2, EObject> entry : other.pos2EObject.entrySet()) {
				int b = shift + entry.getKey().b();
				int e = shift + entry.getKey().e();
				addMapping(entry.getValue(), b, e);
			}
			// mapping eObject2CommentPositions
			for (Entry<EObject, List<Int2>> entry : other.eObject2CommentPositions.entrySet()) {
				for (Int2 pos : entry.getValue()) {
					int b = shift + pos.b();
					int e = shift + pos.e();
					addCommentMapping(entry.getKey(), b, e);
				}
			}
			// mapping pos2LinkEObject
			for (Entry<Int2, EObject> entry : other.pos2LinkEObject.entrySet()) {
				int b = shift + entry.getKey().b();
				int e = shift + entry.getKey().e();
				pos2LinkEObject.put(new Int2(b, e), entry.getValue());
			}
			// highlightedPos
			for (int i = 0; i < other.highlightedPos.length; i++) {
				for (Int2 pos : other.highlightedPos[i]) {
					highlightedPos[i].add(new Int2(shift + pos.b(), shift + pos.e()));
				}
			}
			// shift
			shift += other.shift;
		}
	}

	/**
	 * Shifts text positions used for the mapping and puts the default
	 * highlight.
	 * 
	 * @param size
	 *            is the size of the shift
	 */
	public void shift(int size) {
		shift(size, HIGHLIGHTED_DEFAULT);
	}

	/**
	 * Shifts text positions used for the mapping and puts the given highlight.
	 * 
	 * @param size
	 *            is the size of the shift
	 * @param highlightedType
	 *            is the text highlight
	 */
	public void shift(int size, int highlightedType) {
		if (!commit) {
			if (linkEObject != null) {
				Int2 pos = new Int2(shift, shift + size);
				pos2LinkEObject.put(pos, linkEObject);
			}
			if (highlightedType == HIGHLIGHTED_STATIC_TEXT) {
				Int2 pos = new Int2(shift, shift + size);
				highlightedPos[HIGHLIGHTED_STATIC_TEXT].add(pos);
			} else if (highlightedType == HIGHLIGHTED_COMMENT) {
				Int2 pos = new Int2(shift, shift + size);
				highlightedPos[HIGHLIGHTED_COMMENT].add(pos);
				addCommentMapping(object, pos.b(), pos.e());
			}
			shift += size;
		}
	}

	/**
	 * Adds a mapping between the model object and the text begins at the
	 * specified begin and extends to the character at index end - 1.
	 * 
	 * @param eObj
	 *            is the model object
	 * @param begin
	 *            is the beginning index, inclusive
	 * @param end
	 *            is the ending index, exclusive
	 */
	protected void addMapping(EObject eObj, int begin, int end) {
		if (begin > -1 && end > -1) {
			// Mapping pos2EObject
			pos2EObject.put(new Int2(begin, end), eObj);
			// Mapping eObject2Pos
			List<Int2> objectPositions = eObject2Positions.get(eObj);
			if (objectPositions == null) {
				objectPositions = new ArrayList<Int2>();
				eObject2Positions.put(eObj, objectPositions);
			}
			add(objectPositions, begin, end);
		}
	}

	/**
	 * Adds a comment mapping between the model object and the text begins at
	 * the specified begin and extends to the character at index end - 1.
	 * 
	 * @param eObj
	 *            is the model object
	 * @param begin
	 *            is the beginning index, inclusive
	 * @param end
	 *            is the ending index, exclusive
	 */
	protected void addCommentMapping(EObject eObj, int begin, int end) {
		if (begin > -1 && end > -1) {
			List<Int2> commentPositions = eObject2CommentPositions.get(eObj);
			if (commentPositions == null) {
				commentPositions = new ArrayList<Int2>();
				eObject2CommentPositions.put(eObj, commentPositions);
			}
			add(commentPositions, begin, end);
		}
	}

	/**
	 * Amalgamates the new position and the other positions.
	 * 
	 * @param positions
	 *            is the list of positions
	 * @param begin
	 *            is the beginning index of the new position, inclusive
	 * @param end
	 *            is the ending index of the new position, exclusive
	 */
	private void add(List<Int2> positions, int begin, int end) {
		boolean insert = false;
		ListIterator<Int2> it = positions.listIterator();
		while (!insert && it.hasNext()) {
			Int2 pos = it.next();
			if (begin <= pos.e() && begin >= pos.b()) {
				if (end > pos.e()) {
					it.remove();
					add(positions, pos.b(), end);
				}
				insert = true;
			} else if (end >= pos.b() && end <= pos.e()) {
				if (begin < pos.b()) {
					it.remove();
					add(positions, begin, pos.e());
				}
				insert = true;
			} else if (begin < pos.b() && end > pos.e()) {
				it.remove();
				add(positions, begin, end);
				insert = true;
			}
		}
		if (!insert) {
			it.add(new Int2(begin, end));
		}
	}

	/**
	 * Marks the begin of an open link action.
	 * 
	 * @param linkEObj
	 *            is the linked object
	 */
	public void linkBegin(EObject linkEObj) {
		this.linkEObject = linkEObj;
	}

	/**
	 * Marks the end of the current open link action.
	 */
	public void linkEnd() {
		this.linkEObject = null;
	}

	/**
	 * Validate the changes.
	 * <p>
	 * It is impossible to change mappings now.
	 */
	protected void commit() {
		if (!commit) {
			commit = true;
			addMapping(object, 0, shift);
		}
	}

	/**
	 * Returns the object at the given index.
	 * 
	 * @param index
	 *            is the position in the text
	 * @return the object at the index.
	 */
	public EObject index2EObject(int index) {
		EObject eObj = index2EObject.get(new Integer(index));
		if (eObj == null) {
			if (!commit) {
				commit();
			}
			Iterator<Entry<Int2, EObject>> it = pos2EObject.entrySet().iterator();
			while (eObj == null && it.hasNext()) {
				Entry<Int2, EObject> entry = it.next();
				int b = entry.getKey().b();
				if (index >= b) {
					int e = entry.getKey().e();
					if (index < e) {
						eObj = entry.getValue();
					}
				}
			}
			index2EObject.put(new Integer(index), eObj);
		}
		return eObj;
	}

	/**
	 * Gets a serializable object that maps the ranges of the text and the URIs
	 * of the model's objects.
	 * 
	 * @return a serializable map
	 * @deprecated
	 */
	public Map<Int2, String> position2uriSerializableMap() {
		Map<Int2, String> result = new TreeMap<Int2, String>(new InversePosComparator());
		if (!commit) {
			commit();
		}
		for (Entry<Int2, EObject> entry : pos2EObject.entrySet()) {
			Int2 pos = entry.getKey();
			EObject eObj = entry.getValue();
			String uriFragment = ETools.getURI(eObj);
			result.put(pos, uriFragment);
		}
		return result;
	}

	/**
	 * Returns the positions of the given object.
	 * 
	 * @param eObj
	 *            is the model object
	 * @return the positions of the object
	 */
	public Int2[] eObject2Positions(EObject eObj) {
		Int2[] res;
		if (eObj == null) {
			res = new Int2[] {};
		} else {
			if (!commit) {
				commit();
			}
			List<Int2> positions = eObject2Positions.get(eObj);
			if (positions != null) {
				res = positions.toArray(new Int2[positions.size()]);
			} else {
				res = new Int2[] {};
			}
		}
		return res;
	}

	/**
	 * Gets a serializable object that maps the URIs of the model's objects and
	 * the ranges of the text.
	 * 
	 * @return a serializable map
	 * @deprecated
	 */
	public Map<String, Int2[]> uri2positionsSerializableMap() {
		Map<String, Int2[]> result = new TreeMap<String, Int2[]>(new StringComparator());
		if (!commit) {
			commit();
		}
		for (Entry<EObject, List<Int2>> entry : eObject2Positions.entrySet()) {
			EObject eObj = entry.getKey();
			String uriFragment = ETools.getURI(eObj);
			Set<Int2> positions = new TreeSet<Int2>(new PosComparator());
			positions.addAll(entry.getValue());
			result.put(uriFragment, positions.toArray(new Int2[positions.size()]));
		}
		return result;
	}

	/**
	 * Returns the first comment position of the given object.
	 * 
	 * @param eObj
	 *            is the model object
	 * @param limits
	 *            delimits the part of the text where the comment can be
	 *            searched
	 * @return the first comment position of the object
	 */
	public Int2 eObject2CommentPositionIn(EObject eObj, Int2 limits) {
		Int2 res = Int2.NOT_FOUND;
		if (eObj == null) {
			res = Int2.NOT_FOUND;
		}
		if (!commit) {
			commit();
		} else {
			List<Int2> positions = eObject2CommentPositions.get(eObj);
			if (positions != null) {
				for (Int2 pos : positions) {
					if (pos.b() >= limits.b() && pos.e() <= limits.e()) {
						res = pos;
					}
				}
			}
		}
		return res;
	}

	/**
	 * Returns the linked object at the given index. It is used to make an open
	 * link action.
	 * 
	 * @param index
	 *            is the position in the text
	 * @return the linked object at the index
	 */
	public EObject index2LinkEObject(int index) {
		EObject eObj = index2LinkEObject.get(new Integer(index));
		if (eObj == null) {
			if (!commit) {
				commit();
			}
			for (Entry<Int2, EObject> entry : pos2LinkEObject.entrySet()) {
				int b = entry.getKey().b();
				if (index >= b) {
					int e = entry.getKey().e();
					if (index < e) {
						eObj = entry.getValue();
					}
				}
			}
			index2LinkEObject.put(new Integer(index), eObj);
		}
		return eObj;
	}

	/**
	 * Returns all positions for the given highlight.
	 * 
	 * @param highlightedType
	 *            is the text highlight
	 * @return all positions for the given highlight
	 */
	public Int2[] getHighlightedPos(int highlightedType) {
		Int2[] res;
		if (highlightedType == HIGHLIGHTED_STATIC_TEXT) {
			res = highlightedPos[HIGHLIGHTED_STATIC_TEXT].toArray(new Int2[] {});
		} else if (highlightedType == HIGHLIGHTED_COMMENT) {
			res = highlightedPos[HIGHLIGHTED_COMMENT].toArray(new Int2[] {});
		} else {
			res = new Int2[] {};
		}
		return res;
	}

	/**
	 * Moves the positions into the given range.
	 * 
	 * @param range
	 *            are the new bounds
	 */
	public void range(Int2 range) {
		if (!commit) {
			shift = range.e() - range.b();
			// mapping pos2EObject
			Iterator<Entry<Int2, EObject>> it;
			it = pos2EObject.entrySet().iterator();
			while (it.hasNext()) {
				Entry<Int2, EObject> entry = it.next();
				entry.getKey().range(range);
				if (entry.getKey().b() == -1) {
					it.remove();
				}
			}
			// mapping eObject2Positions
			for (Entry<EObject, List<Int2>> entry : eObject2Positions.entrySet()) {
				List<Int2> value = entry.getValue();
				Iterator<Int2> positions = value.iterator();
				while (positions.hasNext()) {
					Int2 pos = positions.next();
					pos.range(range);
					if (pos.b() == -1) {
						positions.remove();
					}
				}
			}
			// mapping eObject2CommentPositions
			for (Entry<EObject, List<Int2>> entry : eObject2CommentPositions.entrySet()) {
				List<Int2> value = entry.getValue();
				Iterator<Int2> positions = value.iterator();
				while (positions.hasNext()) {
					Int2 pos = positions.next();
					pos.range(range);
					if (pos.b() == -1) {
						positions.remove();
					}
				}
			}
			// mapping pos2LinkEObject
			it = pos2LinkEObject.entrySet().iterator();
			while (it.hasNext()) {
				Entry<Int2, EObject> entry = it.next();
				entry.getKey().range(range);
				if (entry.getKey().b() == -1) {
					it.remove();
				}
			}
			// highlightedPos
			for (int i = 0; i < highlightedPos.length; i++) {
				Iterator<Int2> itPos = highlightedPos[i].iterator();
				while (itPos.hasNext()) {
					Int2 pos = itPos.next();
					pos.range(range);
					if (pos.b() == -1) {
						itPos.remove();
					}
				}
			}
		}
	}

	/**
	 * Applies the indent strategy to the positions (each line adds one
	 * character).
	 * 
	 * @param lines
	 *            are the positions of the lines
	 */
	public void indent(Int2[] lines) {
		if (!commit) {
			shift += lines.length;
			// mapping pos2EObject
			for (Entry<Int2, EObject> entry : pos2EObject.entrySet()) {
				entry.getKey().indent(lines);
			}
			// mapping eObject2Positions
			for (Entry<EObject, List<Int2>> entry : eObject2Positions.entrySet()) {
				for (Int2 pos : entry.getValue()) {
					pos.indent(lines);
				}
			}
			// mapping eObject2CommentPositions
			for (Entry<EObject, List<Int2>> entry : eObject2CommentPositions.entrySet()) {
				for (Int2 pos : entry.getValue()) {
					pos.indent(lines);
				}
			}
			// mapping pos2LinkEObject
			for (Entry<Int2, EObject> entry : pos2LinkEObject.entrySet()) {
				entry.getKey().indent(lines);
			}
			// highlightedPos
			for (int i = 0; i < highlightedPos.length; i++) {
				for (Int2 pos : highlightedPos[i]) {
					pos.indent(lines);
				}
			}
		}
	}
}
