/*
 * Copyright (c) 2005-2008 Obeo.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Obeo - initial API and implementation
 */

package fr.obeo.acceleo.gen.template.eval;

import org.acceleo.template.core.ASTNode;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Map.Entry;

import org.acceleo.tools.strings.Int2;

/**
 * Mapping between the template and the text value.
 * 
 * @author www.obeo.fr
 * 
 */
public class TextTemplateElementMapping {
	/** Default highlight. */
	public static final int HIGHLIGHTED_DEFAULT = -1;

	/** Highlight for the text which is not in the models. */
	public static final int HIGHLIGHTED_STATIC_TEXT = 0;

	/** Highlight for comments. */
	public static final int HIGHLIGHTED_COMMENT = 1;

	/**
	 * The element that compares positions inverse order (Int2).
	 */
	private static class InversePosComparator implements Comparator<Int2>, Serializable {

		/**
		 * The default serial version ID.
		 */
		private static final long serialVersionUID = 1L;

		/**
		 * {@inheritDoc}
		 */
		public int compare(Int2 pos0, Int2 pos1) {
			int res;
			if (pos0.b() < pos1.b()) {
				res = 1;
			} else if (pos0.b() > pos1.b()) {
				res = -1;
			} else {
				if (pos0.e() < pos1.e()) {
					res = -1;
				} else if (pos0.e() > pos1.e()) {
					res = 1;
				} else {
					res = 0;
				}
			}
			return res;
		}
	}

	/**
	 * The element that compares positions (Int2).
	 */
	private static class PosComparator implements Comparator<Int2> {
		/**
		 * {@inheritDoc}
		 */
		public int compare(Int2 pos0, Int2 pos1) {
			int res;
			if (pos0.b() < pos1.b()) {
				res = -1;
			} else if (pos0.b() > pos1.b()) {
				res = 1;
			} else {
				if (pos0.e() < pos1.e()) {
					res = 1;
				} else if (pos0.e() > pos1.e()) {
					res = -1;
				} else {
					res = 0;
				}
			}
			return res;
		}
	}

	/**
	 * Maps text position to EObject.
	 */
	protected Map<Int2, ASTNode> pos2TemplateElement = new TreeMap<Int2, ASTNode>(new InversePosComparator());

	/**
	 * Maps TemplateElement to text position.
	 */
	protected Map<ASTNode, List<Int2>> template2Positions = new HashMap<ASTNode, List<Int2>>();

	/**
	 * Maps TemplateElement to comment text position.
	 */
	protected Map<ASTNode, List<Int2>> template2CommentPositions = new HashMap<ASTNode, List<Int2>>();

	/**
	 * Maps text position to an TemplateElement declaration. It is used for an
	 * open link action.
	 */
	protected Map<Int2, ASTNode> pos2LinkTemplateElement = new TreeMap<Int2, ASTNode>(new PosComparator());

	/**
	 * Highlight positions.
	 * <p>
	 * <li>highlightedPos[HIGHLIGHTED_STATIC_TEXT] is a list of positions for
	 * the text which is not in the models. </li>
	 * <li>highlightedPos[HIGHLIGHTED_COMMENT] is a list of positions for
	 * comments. </li>
	 */
	@SuppressWarnings("unchecked")
	protected List<Int2>[] highlightedPos = new List[2];

	/**
	 * Default Script for text mapping.
	 */
	protected ASTNode script;

	/**
	 * Current size of the text used for the mapping.
	 */
	protected int shift;

	/**
	 * Indicates if the changes were validated.
	 * <p>
	 * It is impossible to change mappings when commit is true.
	 */
	protected boolean commit;

	/**
	 * Open link result for the next added text.
	 */
	protected ASTNode linkScriptElement;

	/**
	 * Mapping of the index to template element.
	 */
	private Map<Integer, ASTNode> index2TemplateElement = new HashMap<Integer, ASTNode>();

	/**
	 * Mapping between index and template element.
	 */
	private Map<Integer, ASTNode> index2LinkTemplateElement = new HashMap<Integer, ASTNode>();

	/**
	 * Constructor.
	 * 
	 * @param script
	 *            is the default script for text mapping
	 * @param freeze
	 *            is used to freeze the mappings, this helps to improve the
	 *            performance
	 */
	public TextTemplateElementMapping(ASTNode script, boolean freeze) {
		this.script = script;
		for (int i = 0; i < highlightedPos.length; i++) {
			if (highlightedPos[i] == null) {
				highlightedPos[i] = new ArrayList<Int2>();
			}
		}
		if (freeze) {
			commit = true;
		} else {
			reset();
		}
	}

	/**
	 * Reset all informations.
	 */
	protected void reset() {
		if (!commit) {
			this.shift = 0;
			pos2TemplateElement.clear();
			index2TemplateElement.clear();
			template2Positions.clear();
			template2CommentPositions.clear();
			pos2LinkTemplateElement.clear();
			index2LinkTemplateElement.clear();
			for (int i = 0; i < highlightedPos.length; i++) {
				if (highlightedPos[i] != null) {
					highlightedPos[i].clear();
				} else {
					highlightedPos[i] = new ArrayList<Int2>();
				}
			}
		}
	}

	/**
	 * Appends other mappings to current mappings.
	 * 
	 * @param other
	 *            is the other mappings between the model and the text value
	 */
	public void from(TextTemplateElementMapping other) {
		if (!commit && other != null) {
			// mapping pos2TemplateElement, template2Pos
			for (Entry<Int2, ASTNode> entry : other.pos2TemplateElement.entrySet()) {
				int b = shift + entry.getKey().b();
				int e = shift + entry.getKey().e();
				ASTNode astNode = entry.getValue();
				addMapping(astNode, b, e);
			}
			// mapping template2CommentPositions
			for (Entry<ASTNode, List<Int2>> entry : other.template2CommentPositions.entrySet()) {
				List<Int2> value = entry.getValue();
				for (Int2 pos : value) {
					int b = shift + pos.b();
					int e = shift + pos.e();
					ASTNode template = entry.getKey();
					addCommentMapping(template, b, e);
				}
			}
			// mapping pos2LinkTemplateElement
			for (Entry<Int2, ASTNode> entry : other.pos2LinkTemplateElement.entrySet()) {
				int b = shift + entry.getKey().b();
				int e = shift + entry.getKey().e();
				ASTNode template = entry.getValue();
				pos2LinkTemplateElement.put(new Int2(b, e), template);
			}
			// highlightedPos
			for (int i = 0; i < other.highlightedPos.length; i++) {
				for (Int2 pos : other.highlightedPos[i]) {
					highlightedPos[i].add(new Int2(shift + pos.b(), shift + pos.e()));
				}
			}
			// shift
			shift += other.shift;
		}
	}

	/**
	 * Shifts text positions used for the mapping and puts the default
	 * highlight.
	 * 
	 * @param size
	 *            is the size of the shift
	 */
	public void shift(int size) {
		shift(size, HIGHLIGHTED_DEFAULT);
	}

	/**
	 * Shifts text positions used for the mapping and puts the given highlight.
	 * 
	 * @param size
	 *            is the size of the shift
	 * @param highlightedType
	 *            is the text highlight
	 */
	public void shift(int size, int highlightedType) {
		if (!commit) {
			if (linkScriptElement != null) {
				Int2 pos = new Int2(shift, shift + size);
				pos2LinkTemplateElement.put(pos, linkScriptElement);
			}
			if (highlightedType == HIGHLIGHTED_STATIC_TEXT) {
				Int2 pos = new Int2(shift, shift + size);
				highlightedPos[HIGHLIGHTED_STATIC_TEXT].add(pos);
			} else if (highlightedType == HIGHLIGHTED_COMMENT) {
				Int2 pos = new Int2(shift, shift + size);
				highlightedPos[HIGHLIGHTED_COMMENT].add(pos);
				addCommentMapping(script, pos.b(), pos.e());
			}
			shift += size;
		}
	}

	/**
	 * Adds a mapping between the script and the text begins at the specified
	 * begin and extends to the character at index end - 1.
	 * 
	 * @param astNode
	 *            is the script element
	 * @param begin
	 *            is the beginning index, inclusive
	 * @param end
	 *            is the ending index, exclusive
	 */
	protected void addMapping(ASTNode astNode, int begin, int end) {
		if (begin > -1 && end > -1) {
			// Mapping pos2TemplateElement
			pos2TemplateElement.put(new Int2(begin, end), astNode);
			// Mapping template2Pos
			List<Int2> templatePositions = template2Positions.get(astNode);
			if (templatePositions == null) {
				templatePositions = new ArrayList<Int2>();
				template2Positions.put(astNode, templatePositions);
			}
			add(templatePositions, begin, end);
		}
	}

	/**
	 * Adds a comment mapping between the script and the text begins at the
	 * specified begin and extends to the character at index end - 1.
	 * 
	 * @param astNode
	 *            is the script
	 * @param begin
	 *            is the beginning index, inclusive
	 * @param end
	 *            is the ending index, exclusive
	 */
	protected void addCommentMapping(ASTNode astNode, int begin, int end) {
		if (begin > -1 && end > -1) {
			List<Int2> commentPositions = template2CommentPositions.get(astNode);
			if (commentPositions == null) {
				commentPositions = new ArrayList<Int2>();
				template2CommentPositions.put(astNode, commentPositions);
			}
			add(commentPositions, begin, end);
		}
	}

	/**
	 * Amalgamates the new position and the other positions.
	 * 
	 * @param positions
	 *            is the list of positions
	 * @param begin
	 *            is the beginning index of the new position, inclusive
	 * @param end
	 *            is the ending index of the new position, exclusive
	 */
	private void add(List<Int2> positions, int begin, int end) {
		boolean insert = false;
		List<Int2> toAdd = new ArrayList<Int2>();
		ListIterator<Int2> it = positions.listIterator();
		while (it.hasNext()) {
			Int2 pos = it.next();
			if (begin <= pos.e() && begin >= pos.b()) {
				if (end > pos.e()) {
					it.remove();
					toAdd.add(new Int2(pos.b(), end));
				}
				insert = true;
			} else if (end >= pos.b() && end <= pos.e()) {
				if (begin < pos.b()) {
					it.remove();
					toAdd.add(new Int2(begin, pos.e()));
				}
				insert = true;
			} else if (begin < pos.b() && end > pos.e()) {
				it.remove();
				toAdd.add(new Int2(begin, end));
				insert = true;
			}
		}
		for (Int2 pos : toAdd) {
			add(positions, pos.b(), pos.e());
		}

		if (!insert) {
			positions.add(new Int2(begin, end));
		}
	}

	/**
	 * Marks the begin of an open link action.
	 * 
	 * @param astNode
	 *            is the linked template
	 */
	public void linkBegin(ASTNode astNode) {
		this.linkScriptElement = astNode;
	}

	/**
	 * Marks the end of the current open link action.
	 */
	public void linkEnd() {
		this.linkScriptElement = null;
	}

	/**
	 * Validate the changes.
	 * <p>
	 * It is impossible to change mappings now.
	 */
	protected void commit() {
		if (!commit) {
			commit = true;
			addMapping(script, 0, shift);
		}
	}

	/**
	 * Returns the template at the given index.
	 * 
	 * @param index
	 *            is the position in the text
	 * @return the template at the index.
	 */
	public ASTNode index2TemplateElement(int index) {
		ASTNode template = index2TemplateElement.get(new Integer(index));
		if (template == null) {
			if (!commit) {
				commit();
			}
			Iterator<Entry<Int2, ASTNode>> it = pos2TemplateElement.entrySet().iterator();
			while (template == null && it.hasNext()) {
				Entry<Int2, ASTNode> entry = it.next();
				int b = entry.getKey().b();
				if (index >= b) {
					int e = entry.getKey().e();
					if (index < e) {
						template = entry.getValue();
					}
				}
			}
			index2TemplateElement.put(new Integer(index), template);
		}
		return template;
	}

	/**
	 * Gets a serializable template that maps the ranges of the text and the
	 * URIs of the templates.
	 * 
	 * @return a serializable map
	 * @deprecated
	 */
	public Map<Int2, String> position2uriSerializableMap() {
		Map<Int2, String> result = new TreeMap<Int2, String>(new InversePosComparator());
		if (!commit) {
			commit();
		}
		for (Entry<Int2, ASTNode> entry : pos2TemplateElement.entrySet()) {
			Int2 pos = entry.getKey();
			ASTNode astNode = entry.getValue();
			// String uriFragment = script.getURIFragment();
			// TODO check if it gives the same behavior
			String uriFragment;
			if (astNode.eResource() != null) {
				uriFragment = astNode.eResource().getURIFragment(astNode);
			} else {
				uriFragment = ""; //$NON-NLS-1$
			}
			result.put(pos, uriFragment);
		}
		return result;
	}

	/**
	 * Returns the positions of the given template.
	 * 
	 * @param astNode
	 *            is the script
	 * @return the positions of the template
	 */
	public Int2[] template2Positions(ASTNode astNode) {
		Int2[] res;
		if (astNode == null) {
			res = new Int2[] {};
		}
		if (!commit) {
			commit();
		}
		List<Int2> positions = template2Positions.get(astNode);
		if (positions != null) {
			res = positions.toArray(new Int2[positions.size()]);
		} else {
			res = new Int2[] {};
		}
		return res;
	}

	/**
	 * Gets a serializable template that maps the URIs of the templates and the
	 * ranges of the text.
	 * 
	 * @return a serializable map
	 * @deprecated
	 */
	public Map<String, Int2[]> uri2positionsSerializableMap() {
		Map<String, Int2[]> result = new TreeMap<String, Int2[]>(new StringComparator());
		if (!commit) {
			commit();
		}
		for (Entry<ASTNode, List<Int2>> entry : template2Positions.entrySet()) {
			ASTNode astNode = entry.getKey();
			// String uriFragment = script.getURIFragment();
			// TODO check if it gives the same behavior
			String uriFragment;
			if (astNode.eResource() != null) {
				uriFragment = astNode.eResource().getURIFragment(astNode);
			} else {
				uriFragment = ""; //$NON-NLS-1$
			}
			Set<Int2> positions = new TreeSet<Int2>(new PosComparator());
			positions.addAll(entry.getValue());
			result.put(uriFragment, positions.toArray(new Int2[positions.size()]));
		}
		return result;
	}

	/**
	 * Returns the first comment position of the given template.
	 * 
	 * @param astNode
	 *            is the script
	 * @param limits
	 *            delimits the part of the text where the comment can be
	 *            searched
	 * @return the first comment position of the template
	 */
	public Int2 template2CommentPositionIn(ASTNode astNode, Int2 limits) {
		Int2 res = Int2.NOT_FOUND;
		if (astNode == null) {
			res = Int2.NOT_FOUND;
		} else {
			if (!commit) {
				commit();
			}
			List<Int2> positions = template2CommentPositions.get(astNode);
			if (positions != null) {
				for (Int2 pos : positions) {
					if (pos.b() >= limits.b() && pos.e() <= limits.e()) {
						res = pos;
					}
				}
			}
		}
		return res;
	}

	/**
	 * Returns the linked template at the given index. It is used to make an
	 * open link action.
	 * 
	 * @param index
	 *            is the position in the text
	 * @return the linked template at the index
	 */
	public ASTNode index2LinkTemplateElement(int index) {
		ASTNode template = index2LinkTemplateElement.get(new Integer(index));
		if (template == null) {
			if (!commit) {
				commit();
			}
			Iterator<Entry<Int2, ASTNode>> it = pos2LinkTemplateElement.entrySet().iterator();
			while (template == null && it.hasNext()) {
				Entry<Int2, ASTNode> entry = it.next();
				int b = entry.getKey().b();
				if (index >= b) {
					int e = entry.getKey().e();
					if (index < e) {
						template = entry.getValue();
					}
				}
			}
			index2LinkTemplateElement.put(new Integer(index), template);
		}
		return template;
	}

	/**
	 * Returns all positions for the given highlight.
	 * 
	 * @param highlightedType
	 *            is the text highlight
	 * @return all positions for the given highlight
	 */
	public Int2[] getHighlightedPos(int highlightedType) {
		Int2[] res;
		if (highlightedType == HIGHLIGHTED_STATIC_TEXT) {
			res = highlightedPos[HIGHLIGHTED_STATIC_TEXT].toArray(new Int2[] {});
		} else if (highlightedType == HIGHLIGHTED_COMMENT) {
			res = highlightedPos[HIGHLIGHTED_COMMENT].toArray(new Int2[] {});
		} else {
			res = new Int2[] {};
		}
		return res;
	}

	/**
	 * Moves the bounds of the errors into the given range.
	 * 
	 * @param range
	 *            are the new bounds
	 */
	public void range(Int2 range) {
		if (!commit) {
			shift = range.e() - range.b();
			// mapping pos2TemplateElement
			Iterator<Entry<Int2, ASTNode>> itPos2TemplateElement = pos2TemplateElement.entrySet().iterator();
			while (itPos2TemplateElement.hasNext()) {
				Entry<Int2, ASTNode> entry = itPos2TemplateElement.next();
				entry.getKey().range(range);
				if (entry.getKey().b() == -1) {
					itPos2TemplateElement.remove();
				}
			}
			// mapping template2Positions
			for (Entry<ASTNode, List<Int2>> entry : template2Positions.entrySet()) {
				List<Int2> value = entry.getValue();
				Iterator<Int2> positions = value.iterator();
				while (positions.hasNext()) {
					Int2 pos = positions.next();
					pos.range(range);
					if (pos.b() == -1) {
						positions.remove();
					}
				}
			}
			// mapping template2CommentPositions
			for (Entry<ASTNode, List<Int2>> entry : template2CommentPositions.entrySet()) {
				List<Int2> value = entry.getValue();
				Iterator<Int2> positions = value.iterator();
				while (positions.hasNext()) {
					Int2 pos = positions.next();
					pos.range(range);
					if (pos.b() == -1) {
						positions.remove();
					}
				}
			}
			// mapping pos2LinkTemplateElement
			Iterator<Entry<Int2, ASTNode>> itPos2LinkTemplateElement = pos2LinkTemplateElement.entrySet().iterator();
			while (itPos2LinkTemplateElement.hasNext()) {
				Entry<Int2, ASTNode> entry = itPos2LinkTemplateElement.next();
				entry.getKey().range(range);
				if (entry.getKey().b() == -1) {
					itPos2LinkTemplateElement.remove();
				}
			}
			// highlightedPos
			for (int i = 0; i < highlightedPos.length; i++) {
				Iterator<Int2> it = highlightedPos[i].iterator();
				while (it.hasNext()) {
					Int2 pos = it.next();
					pos.range(range);
					if (pos.b() == -1) {
						it.remove();
					}
				}
			}
		}
	}

	/**
	 * Applies the indent strategy to the positions (each line adds one
	 * character).
	 * 
	 * @param lines
	 *            are the positions of the lines
	 */
	public void indent(Int2[] lines) {
		if (!commit) {
			shift += lines.length;
			// mapping pos2TemplateElement
			for (Entry<Int2, ASTNode> entry : pos2TemplateElement.entrySet()) {
				entry.getKey().indent(lines);
			}
			// mapping template2Positions
			for (Entry<ASTNode, List<Int2>> entry : template2Positions.entrySet()) {
				for (Int2 pos : entry.getValue()) {
					pos.indent(lines);
				}
			}
			// mapping template2CommentPositions
			for (Entry<ASTNode, List<Int2>> entry : template2CommentPositions.entrySet()) {
				for (Int2 pos : entry.getValue()) {
					pos.indent(lines);
				}
			}
			// mapping pos2LinkTemplateElement
			for (Entry<Int2, ASTNode> entry : pos2LinkTemplateElement.entrySet()) {
				entry.getKey().indent(lines);
			}
			// highlightedPos
			for (int i = 0; i < highlightedPos.length; i++) {
				for (Int2 pos : highlightedPos[i]) {
					pos.indent(lines);
				}
			}
		}
	}
}
