/*
 * Copyright (c) 2008 Obeo.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Obeo - initial API and implementation
 */
package org.acceleo.engine;

import fr.obeo.acceleo.gen.template.eval.ENode;
import fr.obeo.acceleo.gen.template.eval.ENodeCastException;
import fr.obeo.acceleo.gen.template.eval.ENodeException;
import fr.obeo.acceleo.gen.template.eval.TextModelMapping;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;

import org.acceleo.engine.context.EvalContext;
import org.acceleo.engine.template.core.impl.spec.JavaServiceNotFoundException;
import org.acceleo.engine.template.core.impl.spec.TService;
import org.acceleo.engine.template.core.impl.spec.TTemplate;
import org.acceleo.engine.template.impl.spec.IEvaluable;
import org.acceleo.engine.template.impl.spec.TResourceSet;
import org.acceleo.template.Resource;
import org.acceleo.template.ResourceSet;
import org.acceleo.template.TemplatePackage;
import org.acceleo.template.core.Metamodel;
import org.acceleo.template.core.Script;
import org.acceleo.template.core.ScriptDescriptor;
import org.acceleo.template.core.Template;
import org.acceleo.tools.factory.EFactory;
import org.acceleo.tools.factory.ETools;
import org.acceleo.tools.factory.FactoryException;
import org.acceleo.tools.log.EvalLog;
import org.acceleo.tools.template.TemplateConstants;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;

/**
 * Utility class for EMT model handling.
 * 
 * @author <a href="mailto:david.bonneau@nomail.none">David Bonneau</a>
 * @author www.obeo.fr
 * 
 */
public class EMT {

	/** Resources (Templates, Metamodel, Services) used during the generation. */
	private final ResourceSet resourceSet;

	/** Generation source metamodel. */
	private Metamodel metamodel;

	/** Storage for generation logs. */
	private EvalLog logs = new EvalLog();

	/** Number of threads to use for generation. */
	private int threadsNumber = 1;

	/** Number of EOject to generate per thread. */
	private int eObjectsPerThread = 1;

	/**
	 * Mapping between EClass and scripts that can be applied on objects of this
	 * type.
	 */
	private final Map<EClass, List<Script>> compatibleScripts = new ConcurrentHashMap<EClass, List<Script>>();

	/** List templates. */
	private List<Template> templates;

	/**
	 * Constructor.
	 * 
	 * @param resourceSet
	 *            context ResourceSet
	 * @param file
	 *            is the XMI representation of EMT model
	 */
	public EMT(org.eclipse.emf.ecore.resource.ResourceSet resourceSet, URI templateURI, ClassLoader loader) {
		// Register the EMT metamodel.
		TemplatePackage.eINSTANCE.eClass();

		// TODO Refactored this part
		final File file = new File(templateURI.path());
		final Map<String, List<File>> propertyContainer = new HashMap<String, List<File>>();
		final List<File> currentList = new ArrayList<File>();
		currentList.add(file.getParentFile());
		propertyContainer.put(STANDALONE_EXECUTION_CONTEXT, currentList);

		// Demand load the resource for this file.
		this.resourceSet = (ResourceSet) resourceSet.getResource(templateURI, true).getContents().get(0);

		initializeResources(propertyContainer, loader);
	}

	/** Standalone context constant */
	public final static String STANDALONE_EXECUTION_CONTEXT = "STANDALONE_EXECUTION_CONTEXT"; //$NON-NLS-1$

	/**
	 * Constructor.
	 * 
	 * @param templateModel
	 *            in-memory EMT
	 */
	public EMT(ResourceSet templateModel, Map<String, List<File>> propertyContainer, ClassLoader loader) {
		// Register the EMT metamodel
		TemplatePackage.eINSTANCE.eClass();

		resourceSet = templateModel;

		initializeResources(propertyContainer, loader);
	}

	private void initializeResources(Map<String, List<File>> propertyContainer, ClassLoader loader) {
		for (final Resource resource : resourceSet.getResources()) {
			if (resource instanceof TTemplate) {
				final TTemplate template = (TTemplate) resource;
				template.init(propertyContainer);
			} else if (resource instanceof TService) {
				try {
					((TService) resource).init(loader);
				} catch (final JavaServiceNotFoundException e) {
					// TODO exception - throw the exception when old services
					// form Acceleo 2.x will be removed from the EMT
					// (fr.obeo.acceleo...)
				}
			}
		}
	}

	/**
	 * Returns metamodel of the source model.
	 * 
	 * @return metamodel of source model
	 */
	public Metamodel getMetamodel() {
		if (metamodel == null) {
			for (int j = 0; j < resourceSet.getResources().size() && metamodel == null; ++j) {
				if (resourceSet.getResources().get(j) instanceof Metamodel) {
					metamodel = (Metamodel) resourceSet.getResources().get(j);
				}
			}
		}
		return metamodel;
	}

	/**
	 * Generate all file names and their content.
	 * 
	 * @param model
	 *            the model to generate
	 * @param recursive
	 *            is recursive or not
	 * @return a map filename -> generated content
	 * @throws ENodeCastException
	 *             if generation failed
	 * @throws ENodeException
	 *             if generation failed
	 * @throws FactoryException
	 *             if the EMT metamodel's specific implementation is not
	 *             reachable if the metamodel of the EMT and the model aren't
	 *             the same if generation failed
	 */
	// TODO use FileContentMap ?
	// TODO manage Acceleo 2.x IGenFilter
	public Map<String, String> gen(IModel model, boolean recursive) throws FactoryException, ENodeException, ENodeCastException {
		final Map<String, String> ret = new HashMap<String, String>();
		final List<Future<Map<String, String>>> futureResults = new ArrayList<Future<Map<String, String>>>();
		// Checking of specific code access.
		if (metamodelSpecificImplFound()) {
			// Checking of metamodel compatibility.
			// TODO notify using a warning if nsURI are different
			// if (getMetamodel().getName().equals(model.getNsURI())) {
			final ExecutorService pool = Executors.newFixedThreadPool(threadsNumber);
			List<EObject> eObjects = new ArrayList<EObject>(eObjectsPerThread);
			for (final EObject root : model.getRoots()) {
				eObjects.add(root);
				if (recursive) {
					final Iterator<EObject> it = root.eAllContents();
					while (it.hasNext()) {
						eObjects.add(it.next());
						if (eObjects.size() == eObjectsPerThread) {
							futureResults.add(pool.submit(new EMTCallableToString(this, eObjects)));
							eObjects = new ArrayList<EObject>(eObjectsPerThread);
						}
					}
				}
			}

			if (eObjects.size() > 0) {
				futureResults.add(pool.submit(new EMTCallableToString(this, eObjects)));
			}

			for (final Future<Map<String, String>> future : futureResults) {
				try {
					ret.putAll(future.get());
				} catch (final InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				} catch (final ExecutionException e) {
					threadExceptionManamement(e);
				}
			}
			pool.shutdown();

			// TODO notify using a warning if nsURI are different
			// } else {
			// throw new
			// FactoryException(AcceleoEngineMessages.getString("EMT.IncompatibleMetamodel",
			// new Object[] { getMetamodel().getName(), model.getNsURI(), }));
			// //$NON-NLS-1$
			// }
		} else {
			throw new FactoryException(AcceleoEngineMessages.getString("EMT.MetamodelSpecificImplNotFound")); //$NON-NLS-1$
		}
		return ret;
	}

	/**
	 * Tell if we have access to the specific implementation of the EMT
	 * metamodel.
	 * 
	 * @return true if the EMT metamodel is reachable
	 */
	public boolean metamodelSpecificImplFound() {
		return resourceSet instanceof TResourceSet;
	}

	/**
	 * Generate files name and files content.
	 * 
	 * @param root
	 *            the EObject to use for the generation
	 * @param recursive
	 *            should we explore objects contained in the root object?
	 * @return the mapping beetwen files name and file content
	 * @throws ENodeCastException
	 *             if generation failed
	 * @throws ENodeException
	 *             if generation failed
	 * @throws FactoryException
	 *             if generation failed
	 */
	public Map<String, String> gen(EObject root, boolean recursive) throws FactoryException, ENodeException, ENodeCastException {
		final Map<String, String> ret = new HashMap<String, String>();
		final List<Future<Map<String, String>>> futureResults = new ArrayList<Future<Map<String, String>>>();

		final ExecutorService pool = Executors.newFixedThreadPool(threadsNumber);
		List<EObject> eObjects = new ArrayList<EObject>(eObjectsPerThread);
		eObjects.add(root);
		if (recursive) {
			final Iterator<EObject> it = root.eAllContents();
			while (it.hasNext()) {
				eObjects.add(it.next());
				if (eObjects.size() == eObjectsPerThread) {
					futureResults.add(pool.submit(new EMTCallableToString(this, eObjects)));
					eObjects = new ArrayList<EObject>(eObjectsPerThread);
				}
			}
		}

		if (eObjects.size() > 0) {
			futureResults.add(pool.submit(new EMTCallableToString(this, eObjects)));
		}

		for (final Future<Map<String, String>> future : futureResults) {
			try {
				ret.putAll(future.get());
			} catch (final InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (final ExecutionException e) {
				threadExceptionManamement(e);
			}
		}
		pool.shutdown();
		return ret;
	}

	private void threadExceptionManamement(Exception e) throws FactoryException, ENodeException, ENodeCastException {
		final Throwable cause = e.getCause();
		if (cause != null) {
			if (cause instanceof FactoryException) {
				throw (FactoryException) cause;
			} else if (cause instanceof ENodeException) {
				throw (ENodeException) cause;
			} else if (cause instanceof ENodeCastException) {
				throw (ENodeCastException) cause;
			} else {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
	}

	/**
	 * Generate a list of file name that can be generated from the given
	 * EObject.
	 * 
	 * @param root
	 *            the EObject to generate
	 * @return the file name list
	 * @throws ENodeException
	 *             if the generation of the file name fail.
	 * @throws FactoryException
	 *             if the generation of the file name fail. if metamodel
	 *             specific implementation can't be found.
	 * @throws ENodeCastException
	 *             if the generation of the file name fail. if the generated
	 *             file name can't be casted to a string.
	 */
	public List<String> getFileNames(EObject root, EvalContext evalContext) throws FactoryException, ENodeException {
		final List<String> res = new LinkedList<String>();
		final Map<String, ENode> generated = genENodes(root, false, evalContext);
		res.addAll(generated.keySet());
		Iterator<EObject> it = root.eAllContents();
		while(it.hasNext()) {
			res.addAll(genENodes(it.next(), false, evalContext).keySet());
		}
		return res;
	}

	/**
	 * Generate a map with file name as key and file content as value.
	 * 
	 * @param root
	 *            the EObject to generate
	 * @return the file name to content mapping
	 * @throws ENodeException
	 *             if the generation of the file name fail.
	 * @throws FactoryException
	 *             if the generation of the file name fail. if metamodel
	 *             specific implementation can't be found.
	 * @throws ENodeCastException
	 *             if the generation of the file name fail. if the generated
	 *             file name can't be casted to a string.
	 */
	public Map<String, String> genString(EObject root, EvalContext evalContext) throws FactoryException, ENodeException {
		final Map<String, String> res = new HashMap<String, String>();
		final Map<String, ENode> generated = genENodes(root, evalContext);

		for (final Entry<String, ENode> entry : generated.entrySet()) {
			res.put(entry.getKey(), entry.getValue().asString());
		}
		return res;
	}

	/**
	 * Generate a map with file name as key and the generated ENode as value.
	 * 
	 * @param root
	 *            the EObject to generate
	 * @return the file name to ENode mapping
	 * @throws ENodeException
	 *             if the generation of the file name fail.
	 * @throws FactoryException
	 *             if the generation of the file name fail. if metamodel
	 *             specific implementation can't be found.
	 * @throws ENodeCastException
	 *             if the generation of the file name fail. if the generated
	 *             file name can't be casted to a string.
	 */
	public Map<String, ENode> genENodes(EObject root, EvalContext evalContext) throws FactoryException, ENodeException {
		return genENodes(root, true, evalContext);
	}

	/**
	 * Generate a map with file name as key and the generated ENode as value.
	 * 
	 * @param root
	 *            the EObject to generate
	 * @param genContent
	 *            generate file content or not
	 * @param withComment
	 *            should we generate comments to or not
	 * @param evalContext
	 *            the current evaluation context
	 * @return the file name to ENode mapping
	 * @throws ENodeException
	 *             if the generation of the file name fail.
	 * @throws FactoryException
	 *             if the generation of the file name fail. if metamodel
	 *             specific implementation can't be found.
	 */
	private Map<String, ENode> genENodes(EObject root, boolean genContent, EvalContext evalContext) throws FactoryException, ENodeException {
		if (metamodelSpecificImplFound()) {
			final Map<String, ENode> ret = new HashMap<String, ENode>();
			// Checking of metamodel compatibility.
			final List<Script> scripts = getApplicableScripts(root);
			// Default initialization for runtime parsing
			TemplateConstants.initConstants();
			for (final Script script : scripts) {
				// A script that generates files is found.
				if (script.getDescriptor().getFile() != null) {
					List<Template> execPath = getExecutionPath(script.getTemplate());
					for (Template template : execPath) {
						evalContext.startTemplate(template);
					}
					final ENode fileName = ((IEvaluable) script.getDescriptor().getFile()).gen(new ENode(root, (IEvaluable) script.getDescriptor().getFile(), evalContext.getLaunchManager()
							.isSynchronize()), evalContext);
					// File generation.
					ENode node = null;
					if (genContent) {
						node = new ENode(root, (IEvaluable) script, evalContext.getLaunchManager().isSynchronize());
						if (script.getDescriptor().getFile() == null) {
							String comment = eGetAsString(root, "comment"); //$NON-NLS-1$
							if (comment.length() > 0) {
								node.append(comment, TextModelMapping.HIGHLIGHTED_COMMENT);
							}
						}
						node = ((IEvaluable) script).gen(node, evalContext);
						if (script.getDescriptor().getFile() == null) {
							String comment = eGetAsString(root, "endLineComment"); //$NON-NLS-1$
							if (comment.length() > 0) {
								node.append(comment, TextModelMapping.HIGHLIGHTED_COMMENT);
							}
						}
					} else {
						node = new ENode(ENode.EMPTY, null);
					}

					logs.getAll(node.log(), false);

					if (!"".equals(fileName.asString())) { //$NON-NLS-1$
						ret.put(fileName.asString().trim(), node);
					}

					for (int i = 0; i < execPath.size(); ++i) {
						evalContext.endTemplate();
					}
					break; // generate only the first script with file template
				}
			}
			return ret;
		} else {
			throw new FactoryException(AcceleoEngineMessages.getString("EMT.MetamodelSpecificImplNotFound")); //$NON-NLS-1$
		}
	}

	/**
	 * Gets the value of the given feature of the object, as a String. If any
	 * error happened this method return an empty string.
	 * 
	 * @param eObj
	 *            is the object
	 * @param name
	 *            is the feature name
	 * @return the value or an empty string if error happened
	 */
	private String eGetAsString(EObject eObj, String feature) {
		try {
			String result = EFactory.eGetAsString(eObj, feature);
			if (result == null) {
				result = ""; //$NON-NLS-1$
			}
			return result;
		} catch (final FactoryException e) {
			return ""; //$NON-NLS-1$
		}
	}

	/**
	 * Get the execution path to reach the given template.
	 * 
	 * @param template
	 *            the template containing the script to run
	 * @return the path of templates to reach the script to run
	 */
	private List<Template> getExecutionPath(Template template) {
		List<Template> res = new ArrayList<Template>();
		boolean found = false;
		Template rootTemplate = getTemplates().get(0);
		res.add(rootTemplate);

		Iterator<Resource> it = rootTemplate.getImports().iterator();
		while (it.hasNext() && !found) {
			Resource resource = it.next();
			if (resource == template) {
				found = true;
				res.add((Template) resource);
			}
		}

		// if (!found) {
		// it = rootTemplate.getImports().iterator();
		// while (it.hasNext() && !found) {
		// Resource resource = it.next();
		// if (resource instanceof Template) {
		// found = subGetExecutionPath((Template) resource, template, res);
		// if (found) {
		// res.add((Template) resource);
		// }
		// }
		// }
		// }
		return res;
	}

	// /**
	// * Sub method for recursive search of the path.
	// *
	// * @param template
	// * the current template of the path
	// * @param endTemplate
	// * the template to reach
	// * @param res
	// * the resulting list
	// * @return true if the template to reach is found
	// */
	// private boolean subGetExecutionPath(Template template, Template
	// endTemplate,
	// List<Template> res) {
	// boolean found = false;
	//
	// Iterator<Resource> it = template.getImports().iterator();
	// while (it.hasNext() && !found) {
	// Resource resource = it.next();
	// if (resource == template) {
	// found = true;
	// res.add((Template) resource);
	// }
	// }
	//
	// if (!found) {
	// it = template.getImports().iterator();
	// while (it.hasNext() && !found) {
	// Resource resource = it.next();
	// if (resource instanceof Template) {
	// found = subGetExecutionPath((Template) resource, template, res);
	// if (found) {
	// res.add((Template) resource);
	// }
	// }
	// }
	// }
	//
	// return found;
	// }

	/**
	 * Getter for the current EMT resource set.
	 * 
	 * @return the current resource set
	 */
	public ResourceSet getResourceSet() {
		return resourceSet;
	}

	/**
	 * Getter for the logger.
	 * 
	 * @return the current logger
	 */
	public EvalLog getLogs() {
		return logs;
	}

	public void resetLogs() {
		logs = new EvalLog();
	}

	public void setThreadsNumber(int threadsNumber) {
		this.threadsNumber = threadsNumber;
	}

	public void setEObjectsPerThread(int objectsPerThread) {
		eObjectsPerThread = objectsPerThread;
	}

	/**
	 * Computes available templates.
	 * 
	 * @return a map with the available templates
	 */
	public List<Template> getTemplates() {
		if (templates == null) {
			templates = new ArrayList<Template>();
			for (int j = 0; j < resourceSet.getResources().size(); ++j) {
				if (resourceSet.getResources().get(j) instanceof Template) {
					templates.add((Template) resourceSet.getResources().get(j));
				}
			}
		}
		return templates;
	}

	/**
	 * Returns the list of applicable scripts for the object.
	 * 
	 * @param object
	 *            current eObject for which applicable scripts will be found.
	 * @return the list of applicable scripts
	 */
	public List<Script> getApplicableScripts(EObject object) {
		final List<Script> ret = new ArrayList<Script>();
		final List<EClass> compat = new ArrayList<EClass>();
		// let's get all compatible types
		compat.add(object.eClass());
		compat.addAll(object.eClass().getEAllSuperTypes());
		for (final EClass cls : compat) {
			ret.addAll(subGetApplicableScripts(cls));
		}
		return ret;
	}

	/**
	 * Sub-method of getApplicableScripts(EObject object). It returns the list
	 * of applicable scripts for the current EClass.
	 * 
	 * @param eClass
	 *            current eClass for which applicable scripts will be found.
	 * @return the list of applicable scripts for the current EClass.
	 */
	private List<Script> subGetApplicableScripts(EClass eClass) {
		List<Script> ret = compatibleScripts.get(eClass);

		if (ret == null) {
			ret = new ArrayList<Script>();
			String type;

			// get EObject type as String
			type = ETools.getEClassifierPath(eClass);
			// Look for matching scripts in the top level template
			ret.addAll(getMatchingTypeScripts(getTemplates().get(0), type));

			// look for matching script in directly imported templates
			for (final Resource resource : getTemplates().get(0).getImports()) {
				if (resource instanceof Template) {
					ret.addAll(getMatchingTypeScripts((Template) resource, type));
				}
			}

			compatibleScripts.put(eClass, ret);
		}
		return ret;
	}

	/**
	 * Get scripts for the given type. This method scan imported templates as
	 * well.
	 * 
	 * @param template
	 *            the template to scan
	 * @param type
	 *            the type of the EObject in a string representation
	 * @return a list of applicable scripts.
	 */
	private List<Script> getMatchingTypeScripts(Template template, String type) {
		List<Script> ret = new ArrayList<Script>();
		for (final Script script : template.getScripts()) {
			if (script.getDescriptor() != null) {
				final ScriptDescriptor descriptor = script.getDescriptor();
				if (descriptor.getType().equals(type) || type.endsWith("." + descriptor.getType())) { //$NON-NLS-1$
					ret.add(script);
				}
			}
		}
		return ret;
	}
}
