/*
 * Copyright (c) 2008 Obeo.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Obeo - initial API and implementation
 */
package org.acceleo.engine.context;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import org.acceleo.engine.EMT;
import org.acceleo.template.core.Template;
import org.eclipse.core.runtime.CoreException;

/**
 * Reprise du code contenu dans Specific Script.
 * 
 * @author David Bonneau
 * @author www.obeo.fr
 * 
 */
public class ScriptProperties {

	/** Mapping between name of the property and the property itself. */
	private Map<String, Properties> name2Properties = new HashMap<String, Properties>();

	/** Array of all properties. */
	private Properties[] allProperties;

	/** Current template. */
	private Template template;

	/** System resource (.emt) containing the script. */
	private Map<String, List<File>> propertyContainer;

	/**
	 * Constructor.
	 * 
	 * @param script
	 *            the current script
	 * @param file
	 *            the file containing the script
	 */
	public ScriptProperties(Template template, Map<String, List<File>> propertyContainer) {
		this.template = template;
		this.propertyContainer = propertyContainer;
	}

	/**
	 * Clears the found properties.
	 */
	private void clearFoundProperties() {
		name2Properties.clear();
		allProperties = null;
	}

	/**
	 * Gets the property for the key and the property file (without extension).
	 * 
	 * @param name
	 *            is the name of the property file (without ".properties"
	 *            extension)
	 * @param key
	 *            is the key
	 * @return the value for the given key
	 * @throws IOException
	 *             if the property can't be reached
	 */
	public String getProperty(String name, String key) throws IOException {
		Properties properties = getOrCreateProperties(name);
		if (properties != null) {
			return properties.getProperty(key);
		} else {
			return null;
		}
	}

	/**
	 * Try to get or create a properties set with the given name.
	 * 
	 * @param name
	 *            name of the property
	 * @return the properties set
	 * @throws IOException
	 *             if the property file can't be loaded
	 */
	private Properties getOrCreateProperties(String name) throws IOException {
		Properties properties = name2Properties.get(name);
		if (properties == null) {
			boolean found = false;
			List<File> containers = getPropertyContainers();
			Iterator<File> it = containers.iterator();
			while (it.hasNext() && !found) {
				File container = it.next();
				if (container.exists() && container.isDirectory()) {
					File[] members = container.listFiles();
					for (int i = 0; i < members.length && !found; i++) {
						if (members[i].isFile() && (name + ".properties").equals(members[i].getName())) { //$NON-NLS-1$
							properties = new Properties();
							properties.load(members[i].toURL().openStream());
							name2Properties.put(name, properties);
							found = true;
						}
					}
				}
			}
		}
		return properties;
	}

	/**
	 * Gets the value for the key in all the properties files.
	 * 
	 * @param key
	 *            is the key
	 * @return the value for the given key
	 * @throws CoreException
	 * @throws IOException
	 *             if the property can't be reached or created
	 */
	public String getProperty(String key) throws IOException {
		Properties[] properties = getOrCreateProperties();
		for (int i = 0; i < properties.length; i++) {
			if (properties[i] != null) {
				String value = properties[i].getProperty(key);
				if (value != null) {
					return value;
				}
			}
		}
		return null;
	}

	/**
	 * Try to get or create a properties array.
	 * 
	 * @return the properties array
	 * @throws IOException
	 *             if the property can't be reached or created
	 */
	private Properties[] getOrCreateProperties() throws IOException {
		if (allProperties == null) {
			List<Properties> allPropertiesList = new ArrayList<Properties>();
			List<File> containers = getPropertyContainers();
			for (File container : containers) {
				if (container.exists() && container.isDirectory()) {
					File[] members = container.listFiles();
					for (int i = 0; i < members.length; i++) {
						if (members[i].isFile() && members[i].getName() != null && members[i].getName().endsWith(".properties")) { //$NON-NLS-1$
							Properties properties = new Properties();
							properties.load(members[i].toURL().openStream());
							name2Properties.put(new File(members[i].getName()).getPath(), properties);
							allPropertiesList.add(properties);
						}
					}
				}
			}
			allProperties = allPropertiesList.toArray(new Properties[allPropertiesList.size()]);
		}
		return allProperties;
	}

	/**
	 * Getter for the containers of properties.
	 * 
	 * @return the properties files.
	 */
	private List<File> getPropertyContainers() {
		// List<File> result = new ArrayList<File>();
		// TODO je ne sais pas trop comment le calculer. Quels doivent être les
		// points d'entrée lors de la génération !!!
		// Il faut revoir le mécanisme de chain pour prendre des .emt ??? : on
		// n'a plus de correspondance file / script
		// if (chainFile != null && chainFile.getParentFile() != null) {
		// result.add(chainFile.getParentFile());
		// }
		// if (file != null && file.getParentFile() != null) {
		// result.add(file.getParentFile());
		// }
		String templateName = template.getName();
		for (Iterator<String> it = propertyContainer.keySet().iterator(); it.hasNext();) {
			String fileName = it.next();
			String templateFileName = templateName.replace('.', File.separatorChar);
			if (fileName.endsWith(templateFileName + ".mt") || fileName.endsWith(templateFileName + ".emt") || fileName.equals(EMT.STANDALONE_EXECUTION_CONTEXT)) {
				return (List<File>) propertyContainer.get(fileName);
			}
		}
		// TODO Return an error
		return null;
	}
}
