/*
 * Copyright (c) 2005-2008 Obeo.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Obeo - initial API and implementation
 */

package org.acceleo.engine.eval.imports.services;

import fr.obeo.acceleo.gen.template.eval.ENode;
import fr.obeo.acceleo.gen.template.eval.ENodeCastException;
import fr.obeo.acceleo.gen.template.eval.ENodeException;
import fr.obeo.acceleo.gen.template.eval.ENodeIterator;
import fr.obeo.acceleo.gen.template.eval.ENodeList;

import java.util.Comparator;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;

import org.acceleo.engine.AcceleoEngineMessages;
import org.acceleo.engine.context.EvalContext;
import org.acceleo.engine.template.core.impl.spec.TTemplate;
import org.acceleo.engine.template.impl.spec.IEvaluable;
import org.acceleo.parser.template.TemplateSyntaxException;
import org.acceleo.tools.factory.EFactory;
import org.acceleo.tools.factory.FactoryException;
import org.acceleo.tools.format.Conventions;
import org.acceleo.tools.template.TemplateConstants;
import org.eclipse.emf.ecore.EObject;

/**
 * System services for ENode elements.
 * 
 * @author www.obeo.fr
 * 
 */
public final class ENodeServices implements IContextService {

	/**
	 * Evaluation context.
	 */
	private EvalContext evalContext;

	/**
	 * Trace counter.
	 */
	private int trace;

	/**
	 * Gets all the nodes whose type is the given type.
	 * <p>
	 * An element is kept only if it's an EObject whose type is the given type
	 * or if it isn't an EObject.
	 * <p>
	 * Sample :
	 * <p>
	 * <li>a is an instance of class A</li>
	 * <li>b is an instance of class B</li>
	 * <li>c is an instance of class C</li>
	 * <li>B extends A</li>
	 * <p>
	 * If type equals "A" and node is a list {a, "\n", b, c}
	 * <p>
	 * the result is a list {a, "\n", b}.
	 * 
	 * @param node
	 *            is the list
	 * @param type
	 *            is the type kept
	 * @return a list that contains the nodes whose type is the given type.
	 * @throws ENodeCastException
	 *             if cast error occurs
	 * @throws FactoryException
	 *             it can't append
	 */
	// TODO check FactoryException
	public ENode filter(ENode node, String type) throws ENodeCastException, FactoryException {
		ENode filteredNodes;
		if (node.isList()) {
			final ENodeList res = new ENodeList();
			final ENodeIterator it = node.getList().iterator();
			while (it.hasNext()) {
				final ENode child = filter(it.next(), type);
				if (!child.isNull()) {
					res.add(child);
				}
			}
			filteredNodes = new ENode(res, node);
		} else if (node.isEObject()) {
			if (EFactory.eInstanceOf(node.getEObject(), type)) {
				filteredNodes = node;
			} else {
				filteredNodes = new ENode(ENode.EMPTY, node);
			}
		} else {
			filteredNodes = node;
		}
		return filteredNodes;
	}

	/**
	 * Gets all the nodes whose type is the given type.
	 * <p>
	 * An element is kept only if it's an EObject whose type is the given type
	 * or if it isn't an EObject.
	 * <p>
	 * Sample :
	 * <p>
	 * <li>a is an instance of class A</li>
	 * <li>b is an instance of class B</li>
	 * <li>c is an instance of class C</li>
	 * <li>B extends A</li>
	 * <p>
	 * If type equals "A" and node is a list {a, "\n", b, c}
	 * <p>
	 * the result is a list {a, "\n", b}.
	 * 
	 * @param node
	 *            is the list
	 * @param type
	 *            is the type kept
	 * @return a list that contains the nodes whose type is the given type.
	 * @throws ENodeCastException
	 *             if cast error occurs
	 * @throws FactoryException
	 *             it can't append
	 * @deprecated
	 */
	// TODO check FactoryException
	@Deprecated
	public ENode cast(ENode node, String type) throws ENodeCastException, FactoryException {
		return filter(node, type);
	}

	/**
	 * Add a separator between each element of the list. It returns the given
	 * node if it isn't a list.
	 * <p>
	 * Sample :
	 * <p>
	 * If separator equals "\t" and node is a list {a, b, c} the result is a
	 * list {a, "\t", b, "\t", c}.
	 * 
	 * @param node
	 *            is the list
	 * @param separator
	 *            is the separator
	 * @return a list that contains a separator between each element.
	 * @throws ENodeCastException
	 *             if cast error occurs
	 * @throws FactoryException
	 *             it can't append
	 */
	// TODO check FactoryException
	public ENode sep(ENode node, String separator) throws ENodeCastException, FactoryException {
		if (node.isList()) {
			final ENodeList list = node.getList();
			final ENodeList res = new ENodeList();
			final ENodeIterator it = list.iterator();
			if (it.hasNext()) {
				ENode element = it.next();
				boolean needSeparator = element.size() > 0;
				res.add(element);
				while (it.hasNext()) {
					element = it.next();
					if (needSeparator && element.size() > 0) {
						res.add(new ENode(separator, node));
						needSeparator = true;
					} else if (!needSeparator) {
						needSeparator = element.size() > 0;
					}
					res.add(element);
				}
			}
			return new ENode(res, node);
		} else {
			return node;
		}
	}

	/**
	 * Add a separator between each element of the list and serialize the list.
	 * It returns node.toString() if it isn't a list.
	 * <p>
	 * Sample :
	 * <p>
	 * If separator equals "\t" and node is a list {a, b, c} the result is a
	 * string : a + "\t" + b + "\t" + c.
	 * 
	 * @param node
	 *            is the list
	 * @param separator
	 *            is the separator
	 * @return a string that contains a separator between each element
	 * @throws ENodeCastException
	 *             if cast error occurs
	 * @throws FactoryException
	 *             it can't append
	 * @deprecated
	 */
	// TODO check FactoryException
	@Deprecated
	public ENode sepStr(ENode node, String separator) throws ENodeCastException, FactoryException {
		if (node.isList()) {
			final ENodeList list = node.getList();
			final ENodeList res = new ENodeList();
			final ENodeIterator it = list.iterator();
			if (it.hasNext()) {
				ENode element = it.next();
				boolean needSeparator = element.size() > 0;
				res.add(element);
				while (it.hasNext()) {
					element = it.next();
					if (needSeparator && element.size() > 0) {
						res.add(new ENode(separator, node));
						needSeparator = true;
					} else if (!needSeparator) {
						needSeparator = element.size() > 0;
					}
					res.add(element);
				}
			}
			final ENode result = new ENode(res, node);
			result.asString();
			return result;
		} else {
			if (!node.isString()) {
				node.asString();
			}
			return node;
		}
	}

	/**
	 * Get recursively the feature value of an object. The recurrence is stopped
	 * when an element of the given type is found.
	 * <p>
	 * The given node must be an EObject.
	 * <p>
	 * Sample :
	 * <p>
	 * <li>a is an instance of class A</li>
	 * <li>b is an instance of class B</li>
	 * <li>c is an instance of class C</li>
	 * <li>a contains b</li>
	 * <li>b contains c</li>
	 * <li>'container' feature is defined on a, b, and c</li>
	 * <li>c.container returns b</li>
	 * <li>b.container returns a</li>
	 * <li>a.container returns null</li>
	 * <p>
	 * until(c,"container","A") returns a
	 * <p>
	 * until(b,"container","A") returns a
	 * <p>
	 * until(a,"container","A") returns a
	 * <p>
	 * until(b,"container","C") returns null
	 * <p>
	 * 
	 * @param node
	 *            is an EObject
	 * @param link
	 *            is the feature name
	 * @param type
	 *            is the type which stops the recurrence
	 * @return the found object or null
	 * @throws ENodeCastException
	 *             if cast error occurs
	 * @throws FactoryException
	 *             if the link can't be used during recurrence
	 * @deprecated
	 */
	@Deprecated
	public ENode until(ENode node, String link, String type) throws ENodeCastException, FactoryException {
		EObject object = node.getEObject();

		while (object != null) {
			if (EFactory.eInstanceOf(object, type)) {
				return new ENode(object, node);
			} else {
				object = EFactory.eGetAsEObject(object, link);
			}
		}
		return null;
	}

	/**
	 * Returns the current node.
	 * 
	 * @param node
	 *            is the current node
	 * @return the current node
	 */
	public ENode current(ENode node) {
		return node;
	}

	/**
	 * Returns the current node at the level.
	 * 
	 * @param node
	 *            is the current node
	 * @param arg
	 *            is the level to reach
	 * @return the current node
	 * @throws ENodeCastException
	 *             if cast error occurs
	 */
	public ENode current(ENode node, ENode arg) throws ENodeCastException {
		ENode currentNode;
		if (arg.isString()) {
			final String type = arg.getString();
			int i = 0;
			Object value = evalContext.contextAt(TTemplate.SCRIPT_NODE, i);
			while (value != null) {
				final ENode current = ENode.createTry(value, node);
				if (current != null && current.isEObject()) {
					if (EFactory.eInstanceOf(current.getEObject(), type)) {
						return current;
					}
				}
				i++;
				value = evalContext.contextAt(TTemplate.SCRIPT_NODE, i);
			}
			currentNode = new ENode(ENode.EMPTY, node);
		} else if (arg.isInt()) {
			final Object value = evalContext.contextAt(TTemplate.SCRIPT_NODE, arg.getInt());
			currentNode = ENode.createTry(value, node);
		} else {
			currentNode = new ENode(ENode.EMPTY, node);
		}
		return currentNode;
	}

	/**
	 * Transforms the node into a string.
	 * 
	 * @param node
	 *            is the current node
	 * @return the string node
	 */
	public ENode toString(ENode node) {
		if (!node.isString()) {
			node.asString();
		}
		return node;
	}

	/**
	 * Returns the adaptive value for the given type short name.
	 * 
	 * @param node
	 *            is the node
	 * @param type
	 *            is the type name : "EObject", "ENodeList", "String",
	 *            "boolean", "int", "double", "List", "ENode"
	 * @return the adaptive value
	 * @throws ENodeCastException
	 *             if cast error occurs
	 */
	public ENode adapt(ENode node, String type) throws ENodeCastException {
		Class<?> c;
		if ("EObject".equalsIgnoreCase(type)) { //$NON-NLS-1$
			c = EObject.class;
		} else if ("ENodeList".equalsIgnoreCase(type)) { //$NON-NLS-1$
			c = ENodeList.class;
		} else if ("String".equalsIgnoreCase(type)) { //$NON-NLS-1$
			c = String.class;
		} else if ("boolean".equalsIgnoreCase(type)) { //$NON-NLS-1$
			c = boolean.class;
		} else if ("int".equalsIgnoreCase(type)) { //$NON-NLS-1$
			c = int.class;
		} else if ("double".equalsIgnoreCase(type)) { //$NON-NLS-1$
			c = double.class;
		} else if ("List".equalsIgnoreCase(type)) { //$NON-NLS-1$
			c = List.class;
		} else if ("ENode".equalsIgnoreCase(type)) { //$NON-NLS-1$
			c = ENode.class;
		} else {
			c = null;
		}
		final ENode result = ENode.createTry(node.getAdapterValue(c), node);
		if (result != null) {
			return result;
		} else {
			return new ENode(ENode.EMPTY, node);
		}
	}

	/**
	 * It prints current node.
	 * 
	 * @param node
	 *            is the current node
	 * @return the current node
	 */
	public ENode debug(ENode node) {
		return trace(node);
	}

	/**
	 * It prints current node.
	 * 
	 * @param node
	 *            is the current node
	 * @return the current node
	 */
	public ENode trace(ENode node) {
		return trace(node, ""); //$NON-NLS-1$
	}

	/**
	 * It prints current node.
	 * 
	 * @param node
	 *            is the current node
	 * @param template
	 *            is the template to evaluate for each node
	 * @return the current node
	 */
	public ENode trace(ENode node, String template) {
		trace++;
		String prefix;
		if (template != null && template.length() > 0) {
			try {
				prefix = Conventions.formatString(template) + '=' + Conventions.formatString(evaluate(node, template).toString());
			} catch (final TemplateSyntaxException e) {
				prefix = ""; //$NON-NLS-1$
			} catch (final FactoryException e) {
				prefix = ""; //$NON-NLS-1$
			}
		} else {
			prefix = ""; //$NON-NLS-1$
		}
		final String marker = ('[' + trace + "]          ").substring(0, 10); //$NON-NLS-1$
		System.out.println(marker + prefix + ' ' + AcceleoEngineMessages.getString("ENodeServices.TraceMessage", new Object[] { node.getType(), Conventions.formatString(node.toString()), })); //$NON-NLS-1$
		return node;
	}

	/**
	 * Evaluations of the current node.
	 * 
	 * @param current
	 *            Current node
	 * @param call
	 *            String to evaluate
	 * @return evaluated node
	 * @throws TemplateSyntaxException
	 *             if compilation of the call failed
	 * @throws FactoryException
	 *             if the generation of the compiled call failed
	 */
	private ENode evaluate(ENode current, String call) throws TemplateSyntaxException, FactoryException {
		ENode evaluatedNode;
		final String localCall = call.replaceAll("'", TemplateConstants.SPEC); //$NON-NLS-1$
		// TODO YLU compute offset
		int offset;
		if (current.getContainerTemplateElement() != null) {
			offset = current.getContainerTemplateElement().getBegin();
		} else {
			offset = 0;
		}
		final IEvaluable expression = evalContext.compile(offset, localCall);
		try {
			evaluatedNode = expression.gen(current, evalContext);
		} catch (final ENodeException e) {
			evaluatedNode = new ENode(ENode.EMPTY, current);
		}
		return evaluatedNode;
	}

	// /**
	// *
	// * @return
	// */
	// private int getBegin() {
	// Int2 pos = (Int2) script.contextPeek(ScriptContext.ARGUMENT_POSITION);
	// if (pos != null) {
	// return pos.b();
	// } else {
	// return 0;
	// }
	// }

	// private String computeCall(int begin, String call) {
	// if (begin > 0) {
	// StringBuffer result = new StringBuffer(""); //$NON-NLS-1$
	// for (int i = 0; i < begin; i++) {
	// result.append(' ');
	// }
	// result.append(call);
	// return result.toString();
	// } else {
	// return call;
	// }
	// }

	/***************************************************************************
	 * 
	 * List Services
	 * 
	 **************************************************************************/

	/**
	 * Returns the size of the node.
	 * 
	 * @param node
	 *            is the current node
	 *            <li>isEObject() : return 1</li>
	 *            <li>isList() : return getList().size()</li>
	 *            <li>isString() : return getString().length()</li>
	 *            <li>isBoolean() : return 1</li>
	 *            <li>isInt() : return 1</li>
	 *            <li>isNull() : return 0</li>
	 * @return value size
	 */
	public int nSize(ENode node) {
		return node.size();
	}

	/**
	 * Returns the child node at the given index in the current node.
	 * 
	 * @param node
	 *            is the current node
	 * @param index
	 *            is the index of the child
	 * @return the child node
	 * @throws ENodeCastException
	 *             if cast error occurs
	 * @throws FactoryException
	 *             it can't append
	 */
	// TODO check FactoryException
	public ENode nGet(ENode node, int index) throws ENodeCastException, FactoryException {
		ENode childNode;
		if (node.isList()) {
			final ENodeList list = node.getList();
			final ENode result = list.get(index);
			if (result != null) {
				childNode = result;
			} else {
				childNode = new ENode(ENode.EMPTY, node);
			}
		} else if (index == 0) {
			childNode = node;
		} else {
			childNode = new ENode(ENode.EMPTY, node);
		}
		return childNode;
	}

	/**
	 * Returns the children in the given range.
	 * 
	 * @param node
	 *            is the current node
	 * @param begin
	 *            is the beginning index
	 * @param end
	 *            is the ending index
	 * @return the children in the given range
	 * @throws ENodeCastException
	 *             if cast error occurs
	 * @throws FactoryException
	 *             it can't append
	 */
	// TODO check FactoryException
	public ENode nGet(ENode node, int begin, int end) throws ENodeCastException, FactoryException {
		ENode children;
		if (node.isList()) {
			final ENodeList list = node.getList();
			final ENodeList result = new ENodeList();
			for (int i = begin; i < list.size() && (i < end || end == -1); i++) {
				final ENode child = list.get(i);
				result.add(child);
			}
			children = new ENode(result, node);
		} else if (begin == 0 && (end > 0 || end == -1)) {
			children = node;
		} else {
			children = new ENode(ENode.EMPTY, node);
		}
		return children;
	}

	/**
	 * Returns the first child of the current node.
	 * 
	 * @param node
	 *            is the current node
	 * @return the first child node
	 * @throws ENodeCastException
	 *             if cast error occurs
	 * @throws FactoryException
	 *             it can't append
	 */
	// TODO check FactoryException
	public ENode nFirst(ENode node) throws ENodeCastException, FactoryException {
		ENode firstChild;
		if (node.isList()) {
			final ENodeList list = node.getList();
			final ENode result = list.get(0);
			if (result != null) {
				firstChild = result;
			} else {
				firstChild = new ENode(ENode.EMPTY, node);
			}
		} else {
			firstChild = node;
		}
		return firstChild;
	}

	/**
	 * Returns the last child of the current node.
	 * 
	 * @param node
	 *            is the current node
	 * @return the last child node
	 * @throws ENodeCastException
	 *             if cast error occurs
	 * @throws FactoryException
	 *             it can't append
	 */
	// TODO check FactoryException
	public ENode nLast(ENode node) throws ENodeCastException, FactoryException {
		ENode lastChild;
		if (node.isList()) {
			final ENodeList list = node.getList();
			final ENode result = list.get(list.size() - 1);
			if (result != null) {
				lastChild = result;
			} else {
				lastChild = new ENode(ENode.EMPTY, node);
			}
		} else {
			lastChild = node;
		}
		return lastChild;
	}

	/**
	 * Removes all duplicated nodes.
	 * 
	 * @param node
	 *            is the current node
	 * @return the minimized node
	 * @throws ENodeCastException
	 *             if cast error occurs
	 * @deprecated
	 */
	@Deprecated
	public ENode minimize(ENode node) throws ENodeCastException {
		return nMinimize(node);
	}

	/**
	 * Removes all duplicated nodes.
	 * 
	 * @param node
	 *            is the current node
	 * @return the minimized node
	 * @throws ENodeCastException
	 *             if cast error occurs
	 */
	public ENode nMinimize(ENode node) throws ENodeCastException {
		if (node.isList()) {
			final ENodeList result = new ENodeList(true);
			result.addAll(node.getList());
			return new ENode(result, node);
		} else {
			return node;
		}
	}

	/**
	 * Reverses the order of the elements of the specified node.
	 * 
	 * @param node
	 *            is the node whose elements are to be reversed
	 * @return the new node
	 * @throws ENodeCastException
	 *             if cast error occurs
	 * @throws FactoryException
	 *             it can't append
	 * @deprecated
	 */
	// TODO check FactoryException
	@Deprecated
	public ENode reverse(ENode node) throws ENodeCastException, FactoryException {
		return nReverse(node);
	}

	/**
	 * Reverses the order of the elements of the specified node.
	 * 
	 * @param node
	 *            is the node whose elements are to be reversed
	 * @return the new node
	 * @throws ENodeCastException
	 *             if cast error occurs
	 * @throws FactoryException
	 *             it can't append
	 */
	// TODO check FactoryException
	public ENode nReverse(ENode node) throws ENodeCastException, FactoryException {
		if (node.isList()) {
			final ENodeList result = new ENodeList();
			final ENodeIterator it = node.getList().iterator();
			while (it.hasNext()) {
				final ENode child = it.next();
				if (!child.isNull()) {
					result.add(0, child);
				}
			}
			return new ENode(result, node);
		} else {
			return node;
		}
	}

	/**
	 * Returns true if this list contains the specified element.
	 * <p>
	 * More formally, returns true if and only if this list contains at least
	 * <p>
	 * one element e such that (o==null ? e==null : o.equals(e)).
	 * 
	 * @param node
	 *            is the current node
	 * @param element
	 *            is element whose presence in this list is to be tested
	 * @return true if this list contains the specified element
	 * @throws ENodeCastException
	 *             if cast error occurs
	 */
	public boolean nContains(ENode node, ENode element) throws ENodeCastException {
		if (node.isList()) {
			return node.getList().contains(element);
		} else {
			return node.equals(element);
		}
	}

	/**
	 * Sorts the node and its children.
	 * 
	 * @param node
	 *            is the current node
	 * @return the sorted node
	 * @throws ENodeCastException
	 *             if cast error occurs
	 * @deprecated
	 */
	@Deprecated
	public ENode sort(ENode node) throws ENodeCastException {
		return nMinimize(nSort(node));
	}

	/**
	 * Sorts the node and its children.
	 * 
	 * @param node
	 *            is the current node
	 * @return the sorted node
	 * @throws ENodeCastException
	 *             if cast error occurs
	 */
	public ENode nSort(ENode node) throws ENodeCastException {
		if (node.isList()) {
			node.getList().sort();
		}
		return node;
	}

	/**
	 * Sorts the node and its children.
	 * 
	 * @param current
	 *            is the current node of generation
	 * @param call
	 *            is the expression to evaluate for each node
	 * @return the sorted node
	 * @throws ENodeCastException
	 *             if cast error occurs
	 * @deprecated
	 */
	@Deprecated
	public ENode sort(ENode current, String call) throws ENodeCastException {
		return nMinimize(nSort(current, call));
	}

	/**
	 * Sorts the node and its children.
	 * 
	 * @param current
	 *            is the current node of generation
	 * @param call
	 *            is the expression to evaluate for each node
	 * @return the sorted node
	 * @throws ENodeCastException
	 *             if cast error occurs
	 */
	public ENode nSort(ENode current, String call) throws ENodeCastException {
		final String template = call;
		if (current.isList()) {
			final Set<ENode> set = new TreeSet<ENode>(new Comparator<ENode>() {
				public int compare(ENode arg0, ENode arg1) {
					ENode n0 = null;
					ENode n1 = null;
					try {
						n0 = evaluate(arg0, template);
					} catch (TemplateSyntaxException e) {
						n0 = new ENode(ENode.EMPTY, arg0);
					} catch (FactoryException e) {
						n0 = new ENode(ENode.EMPTY, arg0);
					}

					try {
						n1 = evaluate(arg1, template);
					} catch (TemplateSyntaxException e) {
						n1 = new ENode(ENode.EMPTY, arg1);
					} catch (FactoryException e) {
						n1 = new ENode(ENode.EMPTY, arg1);
					}

					return n0.compareTo(n1);
				}
			});
			final ENodeIterator it = current.getList().iterator();
			while (it.hasNext()) {
				set.add(it.next());
			}
			final ENodeList result = new ENodeList();
			for (final ENode node : set) {
				result.add(node);
			}
			return new ENode(result, current);
		} else {
			return current;
		}
	}

	public void setEvalContext(EvalContext evalContext) {
		this.evalContext = evalContext;
	}
}
