/*
 * Copyright (c) 2005-2008 Obeo.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Obeo - initial API and implementation
 */

package org.acceleo.engine.eval.imports.services;

import fr.obeo.acceleo.gen.template.eval.ENode;
import fr.obeo.acceleo.gen.template.eval.ENodeCastException;
import fr.obeo.acceleo.gen.template.eval.ENodeIterator;
import fr.obeo.acceleo.gen.template.eval.ENodeList;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.WeakHashMap;

import org.acceleo.tools.factory.EFactory;
import org.acceleo.tools.factory.ETools;
import org.acceleo.tools.resources.FileContentMap;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.util.EcoreUtil;

/**
 * System services for EObject elements.
 * 
 * @author www.obeo.fr
 * 
 */
public final class EObjectServices implements IThreadSafeService {

	/**
	 * Cache for eAllContenant.
	 */
	private final Map<String, List<EObject>> eAllContents = new WeakHashMap<String, List<EObject>>();

	/**
	 * Root package.
	 */
	private EPackage eRootPackage;

	/**
	 * Cache for containments.
	 */
	private final Map<String, List<String>> eClass2containments = new HashMap<String, List<String>>();

	/**
	 * Cache for subclasses.
	 */
	private Map<EClass, List<EClass>> eClass2subTypes;

	/**
	 * Cache for I/O accesses.
	 */
	private final FileContentMap load = new FileContentMap();

	/**
	 * Gets the container of an EObject.
	 * 
	 * @param current
	 *            is the object
	 * @return the container
	 */
	public EObject eContainer(EObject current) {
		return current.eContainer();
	}

	/**
	 * Gets recursively the container of an EObject until an element of the
	 * given type is found.
	 * 
	 * @param current
	 *            Current model element
	 * @param type
	 *            is the class name
	 * @return recursively the container
	 */
	public EObject eContainer(EObject current, String type) {
		EObject localCurrent = current;
		while (localCurrent != null) {
			if (EFactory.eInstanceOf(localCurrent, type)) {
				return localCurrent;
			} else {
				localCurrent = localCurrent.eContainer();
			}
		}
		return null;
	}

	/**
	 * Returns the root container; it may be this object itself and it will have
	 * a <code>null</code> {@link EObject#eContainer container}.
	 * <p>
	 * The root container must be
	 * {@link Resource#getContents directly contained} in a resource for its
	 * {@link EObject#eAllContents tree} to be
	 * {@link Resource#save(java.util.Map) serializable}.
	 * </p>
	 * 
	 * @param eObject
	 *            the object to get the root container for.
	 * @return the root container.
	 */
	public EObject getRootContainer(EObject eObject) {
		return EcoreUtil.getRootContainer(eObject);
	}

	/**
	 * Gets the metamodel class of an EObject.
	 * 
	 * @param current
	 *            is the object
	 * @return the metamodel class
	 */
	public EObject eClass(EObject current) {
		return current.eClass();
	}

	/**
	 * Gets the children of an EObject.
	 * 
	 * @param current
	 *            is the object
	 * @return the children
	 */
	public EList<EObject> eContents(EObject current) {
		return current.eContents();
	}

	/**
	 * Gets all the direct contents and indirect contents of this object.
	 * 
	 * @param current
	 *            is the object
	 * @return the contents
	 */
	public ENodeList eAllContents(ENode current) {
		final ENodeList result = new ENodeList();
		try {
			if (current.isList()) {
				final ENodeIterator it = current.getList().iterator();
				while (it.hasNext()) {
					result.addAll(eAllContents(it.next()));
				}
			} else if (current.isEObject()) {
				final TreeIterator<EObject> it = current.getEObject().eAllContents();
				while (it.hasNext()) {
					result.add(ENode.createTry(it.next(), current));
				}
			}
		} catch (final ENodeCastException e) {
			// Never catch
		}
		return result;
	}

	/**
	 * Gets all the direct contents and indirect contents of the object. It
	 * keeps only objects of the given type.
	 * 
	 * @param current
	 *            is the object
	 * @param type
	 *            is the type of the objects to select
	 * @return the contents
	 */
	public synchronized ENodeList eAllContents(ENode current, String type) {
		final ENodeList result = new ENodeList();
		try {
			if (current.isList()) {
				final ENodeIterator it = current.getList().iterator();
				while (it.hasNext()) {
					result.addAll(eAllContents(it.next(), type));
				}
			} else if (current.isEObject()) {
				final StringBuffer buffer = new StringBuffer();
				final Resource eResource = current.getEObject().eResource();
				if (eResource != null) {
					buffer.append(eResource.getURI().path());
					buffer.append(':');
					buffer.append(eResource.getURIFragment(current.getEObject()));
				} else {
					buffer.append(':');
					buffer.append(ETools.getURI(current.getEObject()));
				}
				buffer.append(':');
				buffer.append(type);
				final String uri = buffer.toString();
				List<EObject> children = eAllContents.get(uri);
				if (children == null) {
					final int i = type.lastIndexOf("."); //$NON-NLS-1$
					String typeName;
					if (i > -1) {
						typeName = type.substring(i + 1);
					} else {
						typeName = type;
					}
					children = eAllContents(current.getEObject(), type, typeName);
					eAllContents.put(uri, children);
				}
				for (final EObject child : children) {
					result.add(ENode.createTry(child, current));
				}
			}
		} catch (final ENodeCastException e) {
			// Never catch
		}
		return result;
	}

	/**
	 * Get the contained objects of the given EObject that are form the given
	 * type.
	 * 
	 * @param object
	 *            the EBoject from which we want the content
	 * @param type
	 *            the string representing the type
	 * @param typeName
	 *            the EClassifier name
	 * @return the filtered list of contained objects
	 */
	private List<EObject> eAllContents(EObject object, String type, String typeName) {
		final List<EObject> result = new ArrayList<EObject>();
		final List<String> containmentTypeNames = getContainmentNames(object);
		if (containmentTypeNames.contains(typeName)) {
			for (final EObject eContent : object.eContents()) {
				if (EFactory.eInstanceOf(eContent, type)) {
					result.add(eContent);
				}
				result.addAll(eAllContents(eContent, type, typeName));
			}
		}
		return result;
	}

	/**
	 * Getter for containments names.
	 * 
	 * @param object
	 *            the object
	 * @return the list of containments names.
	 */
	private List<String> getContainmentNames(EObject object) {
		final EClass eClass = object.eClass();
		final EPackage ePackage = (EPackage) getRootContainer(eClass);
		if (ePackage != eRootPackage) {
			eRootPackage = ePackage;
			eClass2containments.clear();
			eClass2subTypes = null;
		}
		final String eClassName = eClass.getName();
		List<String> containmentNames = eClass2containments.get(eClassName);
		if (containmentNames == null) {
			containmentNames = new ArrayList<String>();
			computeContainments(new ArrayList<EClassifier>(), containmentNames, ePackage, eClass.getEAllReferences());
			eClass2containments.put(eClassName, containmentNames);
		}
		return containmentNames;
	}

	/**
	 * 
	 * @param containmentTypes
	 * @param containmentNames
	 * @param ePackage
	 * @param eReferences
	 */
	private void computeContainments(List<EClassifier> containmentTypes, List<String> containmentNames, EPackage ePackage, List<EReference> eReferences) {
		for (final EReference eReference : eReferences) {
			if (eReference.isContainment()) {
				final Set<EClassifier> types = new HashSet<EClassifier>();
				types.add(eReference.getEType());
				if (eReference.getEType() instanceof EClass) {
					types.addAll(((EClass) eReference.getEType()).getEAllSuperTypes());
				}
				for (final EClass subType : eClass2subTypes(ePackage, eReference.getEType())) {
					types.add(subType);
					for (final EClass superType : subType.getEAllSuperTypes()) {
						final String name = superType.getName();
						if (!containmentNames.contains(name)) {
							containmentNames.add(name);
						}
					}
				}
				for (final EClassifier type : types) {
					if (type instanceof EClass && !containmentTypes.contains(type)) {
						final String name = type.getName();
						containmentTypes.add(type);
						if (!containmentNames.contains(name)) {
							containmentNames.add(name);
						}
						computeContainments(containmentTypes, containmentNames, ePackage, ((EClass) type).getEReferences());
					}
				}
			}
		}
	}

	/**
	 * List all sub types of EClassifier that belong to the EPackage.
	 * 
	 * @param ePackage
	 *            the EPackage where to find sub types
	 * @param eClass
	 *            the EClassifier from which we want the sub types
	 * @return the list of matched EClass
	 */
	private List<EClass> eClass2subTypes(EPackage ePackage, EClassifier eClass) {
		if (eClass2subTypes == null) {
			eClass2subTypes = new HashMap<EClass, List<EClass>>();
			for (final EClassifier classifier : ETools.computeAllClassifiersList(ePackage, false)) {
				if (classifier instanceof EClass) {
					final EClass c = (EClass) classifier;
					for (final EClass superType : c.getEAllSuperTypes()) {
						List<EClass> subTypes = eClass2subTypes.get(superType);
						if (subTypes == null) {
							subTypes = new ArrayList<EClass>();
							eClass2subTypes.put(superType, subTypes);
						}
						if (!subTypes.contains(c)) {
							subTypes.add(c);
						}
					}
				}
			}
		}
		List<EClass> types = eClass2subTypes.get(eClass);
		if (types == null) {
			types = new ArrayList<EClass>();
		}
		return types;
	}

	/**
	 * Gets the containing feature of an EObject.
	 * 
	 * @param current
	 *            is the object
	 * @return the containing feature
	 */
	public EStructuralFeature eContainingFeature(EObject current) {
		return current.eContainingFeature();
	}

	/**
	 * Gets the containment feature of an EObject.
	 * 
	 * @param current
	 *            is the object
	 * @return the containment feature
	 */
	public EReference eContainmentFeature(EObject current) {
		return current.eContainmentFeature();
	}

	/**
	 * Gets the cross referenced objects.
	 * 
	 * @param current
	 *            is the object
	 * @return the cross referenced objects
	 */
	public List<EObject> eCrossReferences(EObject current) {
		return current.eCrossReferences();
	}

	/**
	 * Gets the containing resource, or null.
	 * 
	 * @param current
	 *            is the object
	 * @return the resource
	 */
	public String eResource(EObject current) {
		if (current != null && current.eResource() != null) {
			return current.eResource().getURI().path();
		} else {
			return null;
		}
	}

	/**
	 * Gets the containing resource name, or null.
	 * 
	 * @param current
	 *            is the object
	 * @return the resource
	 */
	public String eResourceName(EObject current) {
		if (current != null && current.eResource() != null) {
			return current.eResource().getURI().lastSegment();
		} else {
			return null;
		}
	}

	/**
	 * Loads the root element of the given model.
	 * 
	 * @param node
	 *            is the current node
	 * @param path
	 *            is the path of the model to load
	 * @return the root element of the model
	 */
	public synchronized EObject load(ENode node, String path) {
		String localPath = path;
		if (localPath.startsWith("/resource")) { //$NON-NLS-1$
			localPath = localPath.substring("/resource".length()); //$NON-NLS-1$
		}
		final File file = new File(localPath);
		if (file.exists()) {
			// TODO there is no cache when the file is in a jar
			EObject result = (EObject) load.get(file);
			if (result == null) {
				result = ETools.loadXMI(localPath);
				load.put(file, result);
			}
			return result;
		} else {
			return ETools.loadXMI(localPath);
		}
	}
}