/*
 * Copyright (c) 2008 Obeo.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Obeo - initial API and implementation
 */

package org.acceleo.engine.eval.tools;

import fr.obeo.acceleo.gen.template.eval.ENode;
import fr.obeo.acceleo.gen.template.eval.ENodeCastException;
import fr.obeo.acceleo.gen.template.eval.ENodeIterator;
import fr.obeo.acceleo.gen.template.eval.ENodeList;

import org.acceleo.engine.AcceleoEngineMessages;
import org.acceleo.tools.strings.Int2;
import org.acceleo.tools.strings.TextSearch;
import org.acceleo.tools.template.TemplateConstants;

/**
 * Expression evaluation tools. It evaluates ENode expressions.
 * <p>
 * Samples : || , && , == , != , + , - , / , * , > , >= , < , <=
 * 
 * @author www.obeo.fr
 * @see ENode
 */
public final class ExpressionTools {
	/**
	 * Constant for invalid expression localization message.
	 */
	private static final String INVALID_EXPRESSION_MESSAGE = "ExpressionTools.InvalidExpression"; //$NON-NLS-1$

	/**
	 * Constant for adapter not found localization message.
	 */
	private static final String ADAPTER_NOT_FOUND_MESSAGE = "ExpressionTools.AdapterNotFound"; //$NON-NLS-1$

	/**
	 * Constant for brackets printing.
	 */
	private static final String[] BRACKETS_PRINTING_LITERAL = { " [", "]" }; //$NON-NLS-1$ //$NON-NLS-2$

	/**
	 * Constant for decimal separator.
	 */
	private static final String DECIMAL_SEPARATOR_LITERAL = "."; //$NON-NLS-1$

	/**
	 * Private constructor.
	 */
	private ExpressionTools() {

	}

	/**
	 * Evaluates an expression '||' between two nodes.
	 * <p>
	 * 
	 * @param e1
	 *            is the first node
	 * @param e2
	 *            is the second node
	 * @return the result node
	 * @throws ENodeCastException
	 *             if one of the nodes can't be cast to a Boolean.
	 */
	public static ENode or(ENode e1, ENode e2) throws ENodeCastException {
		ENode res;
		if (e1.isList()) {
			ENodeList resList = new ENodeList();
			if (e2.isList()) {
				resList.addAll(e1.getList());
				resList.addAll(e2.getList());
			} else {
				resList.addAll(e1.getList());
				resList.add(e1);
			}
			res = new ENode(resList, e1);
		} else if (e1.isBoolean() && e2.isBoolean()) {
			res = new ENode(e1.getBoolean() || e2.getBoolean(), e1);
		} else {
			res = new ENode(((Boolean) e1.getAdapterValue(boolean.class)).booleanValue() || ((Boolean) e2.getAdapterValue(boolean.class)).booleanValue(), e1);
		}
		return res;
	}

	/**
	 * Evaluates an expression '&&' between two nodes.
	 * <p>
	 * 
	 * @param e1
	 *            is the first node
	 * @param e2
	 *            is the second node
	 * @return the result node
	 * @throws ENodeCastException
	 *             if one of the nodes can't be casted to a Boolean.
	 */
	public static ENode and(ENode e1, ENode e2) throws ENodeCastException {
		ENode res;
		if (e1.isList()) {
			ENodeList resList = new ENodeList();
			if (e2.isList()) {
				ENodeIterator it = e1.getList().iterator();
				while (it.hasNext()) {
					ENode n1 = it.next();
					if (e2.getList().contains(n1)) {
						resList.add(n1);
					}
				}
			} else {
				ENodeIterator it = e1.getList().iterator();
				while (it.hasNext()) {
					ENode n1 = it.next();
					if (e2.equals(n1)) {
						resList.add(n1);
					}
				}
			}
			res = new ENode(resList, e1);
		} else if (e1.isBoolean() && e2.isBoolean()) {
			res = new ENode(e1.getBoolean() && e2.getBoolean(), e1);
		} else {
			res = new ENode(((Boolean) e1.getAdapterValue(boolean.class)).booleanValue() && ((Boolean) e2.getAdapterValue(boolean.class)).booleanValue(), e1);
		}
		return res;
	}

	/**
	 * Evaluates an expression '==' between two nodes.
	 * <p>
	 * Returns e1.equals(e2)
	 * <p>
	 * 
	 * @param e1
	 *            is the first node
	 * @param e2
	 *            is the second node
	 * @return the result node
	 * @throws ENodeCastException
	 *             if the types of the nodes are not compatible
	 */
	public static ENode equals(ENode e1, ENode e2) throws ENodeCastException {
		return new ENode(e1.equals(e2), e1);
	}

	/**
	 * Evaluates an expression '!=' between two nodes.
	 * <p>
	 * Returns !e1.equals(e2)
	 * <p>
	 * 
	 * @param e1
	 *            is the first node
	 * @param e2
	 *            is the second node
	 * @return the result node
	 * @throws ENodeCastException
	 *             if the types of the nodes are not compatible
	 */
	public static ENode notEquals(ENode e1, ENode e2) throws ENodeCastException {
		return new ENode(!e1.equals(e2), e1);
	}

	/**
	 * Evaluates an expression '+' between two nodes.
	 * <p>
	 * 
	 * @param e1
	 *            is the first node
	 * @param e2
	 *            is the second node
	 * @return the result node
	 * @throws ENodeCastException
	 *             if one of the nodes can't be casted to an int or Double.
	 */
	public static ENode add(ENode e1, ENode e2) throws ENodeCastException {
		ENode res;
		if (e1.isList()) {
			ENodeList resList = new ENodeList();
			if (e2.isList()) {
				resList.addAll(e1.getList());
				resList.addAll(e2.getList());
			} else {
				resList.addAll(e1.getList());
				resList.add(e2);
			}
			res = new ENode(resList, e1);
		} else if (e2.isList()) {
			ENodeList resList = new ENodeList();
			resList.add(e1);
			resList.addAll(e2.getList());
			res = new ENode(resList, e1);
		} else if (e1.isInt() && e2.isInt()) {
			res = new ENode(e1.getInt() + e2.getInt(), e1);
		} else if (e1.isDouble() && e2.isDouble()) {
			res = new ENode(e1.getDouble() + e2.getDouble(), e1);
		} else if (e1.isDouble() && e2.isInt()) {
			res = new ENode(e1.getDouble() + e2.getInt(), e1);
		} else if (e1.isInt() && e2.isDouble()) {
			res = new ENode(e1.getInt() + e2.getDouble(), e1);
		} else if (e1.isString() || e2.isString()) {
			ENode result = new ENode("", e1); //$NON-NLS-1$
			result.append(e1);
			result.append(e2);
			res = result;
		} else if (e1.isNull()) {
			res = e2;
		} else if (e2.isNull()) {
			res = e1;
		} else {
			ENodeList resList = new ENodeList();
			resList.add(e1);
			resList.add(e2);
			res = new ENode(resList, e1);
		}
		return res;
	}

	/**
	 * Evaluates an expression '-' between two nodes.
	 * <p>
	 * 
	 * @param e1
	 *            is the first node
	 * @param e2
	 *            is the second node
	 * @return the result node
	 * @throws ENodeCastException
	 *             if one of the nodes can't be casted to an Integer or a
	 *             Double. if the expression is not valid.
	 */
	public static ENode sub(ENode e1, ENode e2) throws ENodeCastException {
		ENode res = null;
		if (e1.isList()) {
			ENodeList resList = new ENodeList();
			resList.addAll(e1.getList());
			if (e2.isList()) {
				ENodeIterator it = e2.getList().iterator();
				while (it.hasNext()) {
					ENode n2 = it.next();
					resList.remove(n2);
				}
			} else {
				resList.remove(e2);
			}
			res = new ENode(resList, e1);
		} else if (e1.isInt() && e2.isInt()) {
			res = new ENode(e1.getInt() - e2.getInt(), e1);
		} else if (e1.isDouble() && e2.isDouble()) {
			res = new ENode(e1.getDouble() - e2.getDouble(), e1);
		} else if (e1.isDouble() && e2.isInt()) {
			res = new ENode(e1.getDouble() - e2.getInt(), e1);
		} else if (e1.isInt() && e2.isDouble()) {
			res = new ENode(e1.getInt() - e2.getDouble(), e1);
		} else if (e1.isNull()) {
			res = e1;
		} else if (e2.isNull()) {
			res = e1;
		} else if (e1.isString()) {
			Int2[] positions = TextSearch.getDefaultSearch().allIndexOf(e1.getString(), e2.toString());
			if (positions.length == 0) {
				res = e1;
			} else {
				String s1 = e1.getString();
				StringBuffer result = new StringBuffer(s1.substring(0, positions[0].b()));
				for (int i = 1; i < positions.length; i++) {
					result.append(s1.substring(positions[i - 1].e(), positions[i].b()));
				}
				result.append(s1.substring(positions[positions.length - 1].e(), s1.length()));
				res = new ENode(result.toString(), e1);
			}
		} else if (e1.isEObject()) {
			if (e1.equals(e2)) {
				res = new ENode(ENode.EMPTY, e1);
			} else {
				res = e1;
			}
		}
		if (res != null) {
			return res;
		} else {
			final String expression = e1.getType() + ' ' + TemplateConstants.OPERATOR_SUB + ' ' + e2.getType();
			throw new ENodeCastException(AcceleoEngineMessages.getString(INVALID_EXPRESSION_MESSAGE, new Object[] { expression, }));
		}
	}

	/**
	 * Evaluates an expression '/' between two nodes.
	 * <p>
	 * 
	 * @param e1
	 *            is the first node
	 * @param e2
	 *            is the second node
	 * @return the result node
	 * @throws ENodeCastException
	 *             if one of the nodes can't be casted to an Integer or a
	 *             Double. if the expression is not valid.
	 */
	public static ENode div(ENode e1, ENode e2) throws ENodeCastException {
		ENode res = null;
		if (e1.isInt() && e2.isInt()) {
			res = new ENode(e1.getInt() / div0(e2.getInt()), e1);
		} else if (e1.isDouble() && e2.isDouble()) {
			res = new ENode(e1.getDouble() / div0(e2.getDouble()), e1);
		} else if (e1.isDouble() && e2.isInt()) {
			res = new ENode(e1.getDouble() / div0(e2.getInt()), e1);
		} else if (e1.isInt() && e2.isDouble()) {
			res = new ENode(e1.getInt() / div0(e2.getDouble()), e1);
		} else {
			try {
				// adapter "int / int"
				if (e2.isInt() || (e2.isString() && e2.getString().indexOf(DECIMAL_SEPARATOR_LITERAL) == -1)) {
					res = new ENode(((Integer) e1.getAdapterValue(int.class)).intValue() / div0(((Integer) e2.getAdapterValue(int.class)).intValue()), e1);
				}
			} catch (ENodeCastException ex1) {
				// step
			}
			try {
				// adapter "double / double" (it includes "int / double" and
				// "double / int")
				if (e2.isDouble() || e2.isString()) {
					res = new ENode(((Double) e1.getAdapterValue(double.class)).doubleValue() / div0(((Double) e2.getAdapterValue(double.class)).doubleValue()), e1);
				}
			} catch (ENodeCastException ex2) {
				// step
			}
			if (res == null) {
				final String expression = e1.getType() + ' ' + TemplateConstants.OPERATOR_DIV + ' ' + e2.getType();
				throw new ENodeCastException(AcceleoEngineMessages.getString(INVALID_EXPRESSION_MESSAGE, new Object[] { expression, }));
			}
		}
		return res;
	}

	/**
	 * Divide by an int with exception management.
	 * 
	 * @param i
	 *            the divisor
	 * @return the result of the division
	 * @throws ENodeCastException
	 *             if i == 0.
	 */
	private static int div0(int i) throws ENodeCastException {
		if (i == 0) {
			throw new ENodeCastException(AcceleoEngineMessages.getString("ExpressionTools.ZeroDivide")); //$NON-NLS-1$
		}
		return i;
	}

	/**
	 * Divide by a Double with exception management.
	 * 
	 * @param d
	 *            the divisor
	 * @return the result of the division
	 * @throws ENodeCastException
	 *             if d == 0.
	 */
	private static double div0(double d) throws ENodeCastException {
		if (d == 0) {
			throw new ENodeCastException(AcceleoEngineMessages.getString("ExpressionTools.ZeroDivide")); //$NON-NLS-1$
		}
		return d;
	}

	/**
	 * Evaluates an expression '*' between two nodes.
	 * <p>
	 * 
	 * @param e1
	 *            is the first node
	 * @param e2
	 *            is the second node
	 * @return the result node
	 * @throws ENodeCastException
	 *             if one of the nodes can't be casted to an Integer or a
	 *             Double. if the expression is not valid.
	 */
	public static ENode mul(ENode e1, ENode e2) throws ENodeCastException {
		ENode res = null;
		if (e1.isInt() && e2.isInt()) {
			res = new ENode(e1.getInt() * e2.getInt(), e1);
		} else if (e1.isDouble() && e2.isDouble()) {
			res = new ENode(e1.getDouble() * e2.getDouble(), e1);
		} else if (e1.isDouble() && e2.isInt()) {
			res = new ENode(e1.getDouble() * e2.getInt(), e1);
		} else if (e1.isInt() && e2.isDouble()) {
			res = new ENode(e1.getInt() * e2.getDouble(), e1);
		} else {
			try {
				// adapter "int * int"
				boolean isInt1 = e1.isInt() || (e1.isString() && e1.getString().indexOf(DECIMAL_SEPARATOR_LITERAL) == -1);
				boolean isInt2 = e2.isInt() || (e2.isString() && e2.getString().indexOf(DECIMAL_SEPARATOR_LITERAL) == -1);
				if (isInt1 && isInt2) {
					res = new ENode(((Integer) e1.getAdapterValue(int.class)).intValue() * ((Integer) e2.getAdapterValue(int.class)).intValue(), e1);
				}
			} catch (ENodeCastException ex1) {
				// step
			}
			try {
				// adapter "double * double" (it includes "int * double" and
				// "double * int")
				if ((e1.isDouble() || e1.isString()) && (e2.isDouble() || e2.isString())) {
					res = new ENode(((Double) e1.getAdapterValue(double.class)).doubleValue() * ((Double) e2.getAdapterValue(double.class)).doubleValue(), e1);
				}
			} catch (ENodeCastException ex2) {
				// step
			}
			if (res == null) {
				final String expression = e1.getType() + ' ' + TemplateConstants.OPERATOR_MUL + ' ' + e2.getType();
				throw new ENodeCastException(AcceleoEngineMessages.getString(INVALID_EXPRESSION_MESSAGE, new Object[] { expression, }));
			}
		}
		return res;
	}

	/**
	 * Evaluates an expression '>' between two nodes.
	 * <p>
	 * 
	 * @param e1
	 *            is the first node
	 * @param e2
	 *            is the second node
	 * @return the result node
	 * @throws ENodeCastException
	 *             if one of the nodes can't be casted to an Integer or a
	 *             Double. if the expression is not valid.
	 */
	public static ENode sup(ENode e1, ENode e2) throws ENodeCastException {
		ENode res = null;
		if (e1.isInt() && e2.isInt()) {
			res = new ENode(e1.getInt() > e2.getInt(), e1);
		} else if (e1.isDouble() && e2.isDouble()) {
			res = new ENode(e1.getDouble() > e2.getDouble(), e1);
		} else if (e1.isDouble() && e2.isInt()) {
			res = new ENode(e1.getDouble() > e2.getInt(), e1);
		} else if (e1.isInt() && e2.isDouble()) {
			res = new ENode(e1.getInt() > e2.getDouble(), e1);
		} else if (e1.isString() && e2.isString()) {
			res = new ENode(e1.getString().compareTo(e2.getString()) > 0, e1);
		} else {
			try {
				// adapter "double > double" (it includes "int > double" and
				// "double > int")
				res = new ENode(((Double) e1.getAdapterValue(double.class)).doubleValue() > ((Double) e2.getAdapterValue(double.class)).doubleValue(), e1);
			} catch (ENodeCastException ex) {
				final String expression = e1.toString() + TemplateConstants.OPERATOR_SUP + e2.toString() + BRACKETS_PRINTING_LITERAL[0] + e1.getType() + TemplateConstants.OPERATOR_SUP + e2.getType()
						+ BRACKETS_PRINTING_LITERAL[1];
				throw new ENodeCastException(AcceleoEngineMessages.getString(ADAPTER_NOT_FOUND_MESSAGE, new Object[] { TemplateConstants.OPERATOR_SUP, expression, }));
			}
		}
		return res;
	}

	/**
	 * Evaluates an expression '>=' between two nodes.
	 * <p>
	 * 
	 * @param e1
	 *            is the first node
	 * @param e2
	 *            is the second node
	 * @return the result node
	 * @throws ENodeCastException
	 *             if one of the nodes can't be casted to an Integer or a
	 *             Double. if the expression is not valid.
	 */
	public static ENode supE(ENode e1, ENode e2) throws ENodeCastException {
		ENode res = null;
		if (e1.isInt() && e2.isInt()) {
			res = new ENode(e1.getInt() >= e2.getInt(), e1);
		} else if (e1.isDouble() && e2.isDouble()) {
			res = new ENode(e1.getDouble() >= e2.getDouble(), e1);
		} else if (e1.isDouble() && e2.isInt()) {
			res = new ENode(e1.getDouble() >= e2.getInt(), e1);
		} else if (e1.isInt() && e2.isDouble()) {
			res = new ENode(e1.getInt() >= e2.getDouble(), e1);
		} else if (e1.isString() && e2.isString()) {
			res = new ENode(e1.getString().compareTo(e2.getString()) >= 0, e1);
		} else {
			try {
				// adapter "double >= double" (it includes "int >= double"
				// and "double >= int")
				res = new ENode(((Double) e1.getAdapterValue(double.class)).doubleValue() >= ((Double) e2.getAdapterValue(double.class)).doubleValue(), e1);
			} catch (ENodeCastException ex) {
				final String expression = e1.toString() + TemplateConstants.OPERATOR_SUP_EQUALS + e2.toString() + BRACKETS_PRINTING_LITERAL[0] + e1.getType() + TemplateConstants.OPERATOR_SUP_EQUALS
						+ e2.getType() + BRACKETS_PRINTING_LITERAL[1];
				throw new ENodeCastException(AcceleoEngineMessages.getString(ADAPTER_NOT_FOUND_MESSAGE, new Object[] { TemplateConstants.OPERATOR_SUP_EQUALS, expression, }));
			}
		}
		return res;
	}

	/**
	 * Evaluates an expression '<' between two nodes.
	 * <p>
	 * 
	 * @param e1
	 *            is the first node
	 * @param e2
	 *            is the second node
	 * @return the result node
	 * @throws ENodeCastException
	 *             if one of the nodes can't be casted to an Integer or a
	 *             Double. if the expression is not valid.
	 */
	public static ENode inf(ENode e1, ENode e2) throws ENodeCastException {
		ENode res = null;
		if (e1.isInt() && e2.isInt()) {
			res = new ENode(e1.getInt() < e2.getInt(), e1);
		} else if (e1.isDouble() && e2.isDouble()) {
			res = new ENode(e1.getDouble() < e2.getDouble(), e1);
		} else if (e1.isDouble() && e2.isInt()) {
			res = new ENode(e1.getDouble() < e2.getInt(), e1);
		} else if (e1.isInt() && e2.isDouble()) {
			res = new ENode(e1.getInt() < e2.getDouble(), e1);
		} else if (e1.isString() && e2.isString()) {
			res = new ENode(e1.getString().compareTo(e2.getString()) < 0, e1);
		} else {
			try {
				// adapter "double < double" (it includes "int < double" and
				// "double < int")
				res = new ENode(((Double) e1.getAdapterValue(double.class)).doubleValue() < ((Double) e2.getAdapterValue(double.class)).doubleValue(), e1);
			} catch (ENodeCastException ex) {
				final String expression = e1.toString() + TemplateConstants.OPERATOR_INF + e2.toString() + BRACKETS_PRINTING_LITERAL[0] + e1.getType() + TemplateConstants.OPERATOR_INF + e2.getType()
						+ BRACKETS_PRINTING_LITERAL[1];
				throw new ENodeCastException(AcceleoEngineMessages.getString(ADAPTER_NOT_FOUND_MESSAGE, new Object[] { TemplateConstants.OPERATOR_INF, expression, }));
			}
		}
		return res;
	}

	/**
	 * Evaluates an expression '<=' between two nodes.
	 * <p>
	 * 
	 * @param e1
	 *            is the first node
	 * @param e2
	 *            is the second node
	 * @return the result node
	 * @throws ENodeCastException
	 *             if one of the nodes can't be casted to an Integer or a
	 *             Double. if the expression is not valid.
	 */
	public static ENode infE(ENode e1, ENode e2) throws ENodeCastException {
		ENode res = null;
		if (e1.isInt() && e2.isInt()) {
			res = new ENode(e1.getInt() <= e2.getInt(), e1);
		} else if (e1.isDouble() && e2.isDouble()) {
			res = new ENode(e1.getDouble() <= e2.getDouble(), e1);
		} else if (e1.isDouble() && e2.isInt()) {
			res = new ENode(e1.getDouble() <= e2.getInt(), e1);
		} else if (e1.isInt() && e2.isDouble()) {
			res = new ENode(e1.getInt() <= e2.getDouble(), e1);
		} else if (e1.isString() && e2.isString()) {
			res = new ENode(e1.getString().compareTo(e2.getString()) <= 0, e1);
		} else {
			try {
				// adapter "double <= double" (it includes "int <= double"
				// and "double <= int")
				res = new ENode(((Double) e1.getAdapterValue(double.class)).doubleValue() <= ((Double) e2.getAdapterValue(double.class)).doubleValue(), e1);
			} catch (ENodeCastException ex) {
				final String expression = e1.toString() + TemplateConstants.OPERATOR_INF_EQUALS + e2.toString() + BRACKETS_PRINTING_LITERAL[0] + e1.getType() + TemplateConstants.OPERATOR_INF_EQUALS
						+ e2.getType() + BRACKETS_PRINTING_LITERAL[1];
				throw new ENodeCastException(AcceleoEngineMessages.getString(ADAPTER_NOT_FOUND_MESSAGE, new Object[] { TemplateConstants.OPERATOR_INF_EQUALS, expression, }));
			}
		}
		return res;
	}
}
