/*
 * Copyright (c) 2008 Obeo.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Obeo - initial API and implementation
 */
package org.acceleo.engine.template.core.impl.spec;

import java.io.File;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.acceleo.engine.context.ScriptProperties;
import org.acceleo.engine.eval.imports.services.ContextServices;
import org.acceleo.engine.eval.imports.services.ENodeServices;
import org.acceleo.engine.eval.imports.services.EObjectServices;
import org.acceleo.engine.eval.imports.services.PropertiesServices;
import org.acceleo.engine.eval.imports.services.RequestServices;
import org.acceleo.engine.eval.imports.services.ResourceServices;
import org.acceleo.engine.eval.imports.services.StringServices;
import org.acceleo.engine.eval.imports.services.XpathServices;
import org.acceleo.template.Resource;
import org.acceleo.template.core.CoreFactory;
import org.acceleo.template.core.Metamodel;
import org.acceleo.template.core.Script;
import org.acceleo.template.core.Service;
import org.acceleo.template.core.impl.TemplateImpl;
import org.acceleo.tools.factory.ETools;
import org.acceleo.tools.template.TemplateConstants;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EcorePackage;

/**
 * Current template management.
 * 
 * @author David Bonneau
 * @author www.obeo.fr
 * 
 */
public class TTemplate extends TemplateImpl {

    /** Context key for while index. */
    public static final Integer WHILE_INDEX = new Integer(0);

    /** Context key for template arguments. */
    public static final Integer SCRIPT_ARGS = new Integer(1);

    /** Context key for service position. */
    public static final Integer ARGUMENT_POSITION = new Integer(2);

    /** Context key for current node. */
    public static final Integer CURRENT_NODE = new Integer(3);

    /** Context key for current node. */
    public static final Integer SCRIPT_NODE = new Integer(4);

    /** Utility class for properties management. */
    private ScriptProperties properties;

    /** Map script (name, type) -> script. */
    private Map<String, Script> scripts;

    /** List of system services. */
    private static List<TService> systemServices;

    /** List of accessible services. */
    private List<TService> javaServices;

    /** List Script names. */
    private List<String> scriptNames;

    /**
     * Constructor.
     * 
     * @param template
     *            is the current template
     */
    public void init(Map<String, List<File>> propertyContainer) {
        javaServices = new ArrayList<TService>();
        // Initialisation des imports de services
        for (final Resource currentImport : getImports()) {
            if (currentImport instanceof Service) {
                javaServices.add((TService) currentImport);
            }
        }

        scriptNames = new ArrayList<String>();
        for (final Script script : getScripts()) {
            scriptNames.add(script.getDescriptor().getName());
        }

        properties = new ScriptProperties(this, propertyContainer);
        javaServices.addAll(getSystemServices());
    }

    private static List<TService> getSystemServices() {
        if (systemServices == null) {
            systemServices = new ArrayList<TService>();
            Service serv;
            serv = CoreFactory.eINSTANCE.createService();
            serv.setName(StringServices.class.getName());
            ((TService) serv).init(new StringServices());
            systemServices.add((TService) serv);

            serv = CoreFactory.eINSTANCE.createService();
            serv.setName(EObjectServices.class.getName());
            ((TService) serv).init(new EObjectServices());
            systemServices.add((TService) serv);

            serv = CoreFactory.eINSTANCE.createService();
            serv.setName(XpathServices.class.getName());
            ((TService) serv).init(new XpathServices());
            systemServices.add((TService) serv);

            serv = CoreFactory.eINSTANCE.createService();
            serv.setName(ResourceServices.class.getName());
            ((TService) serv).init(new ResourceServices());
            systemServices.add((TService) serv);

            serv = CoreFactory.eINSTANCE.createService();
            serv.setName(ContextServices.class.getName());
            ((TService) serv).init(new ContextServices());
            systemServices.add((TService) serv);

            serv = CoreFactory.eINSTANCE.createService();
            serv.setName(ENodeServices.class.getName());
            ((TService) serv).init(new ENodeServices());
            systemServices.add((TService) serv);

            serv = CoreFactory.eINSTANCE.createService();
            serv.setName(RequestServices.class.getName());
            ((TService) serv).init(new RequestServices());
            systemServices.add((TService) serv);

            serv = CoreFactory.eINSTANCE.createService();
            serv.setName(PropertiesServices.class.getName());
            ((TService) serv).init(new PropertiesServices());
            systemServices.add((TService) serv);
        }

        return systemServices;
    }

    /**
     * Get the Script with a given name that can be applied to the given
     * EObject.
     * 
     * @param eObj
     *            the source object of the script
     * @param name
     *            the name of the script
     * @return the script
     */
    public Script getScript(EObject eObj, String name) {
        Script script = null;
        final EClass eClass = eObj.eClass();
        if (scriptNames.contains(name)) {
            script = getScript(eClass, name);
            if (script == null) {
                if (eClass.getESuperTypes().isEmpty() && eClass != EcorePackage.eINSTANCE.getEObject()) {
                    /*
                     * We want to add an "artificial" specialization on EObject
                     */
                    script = getScript(EcorePackage.eINSTANCE.getEObject(), name);
                } else {
                    script = getScriptFromSupertype(eClass.getESuperTypes(), name);
                    if (script != null) {
                        scripts.put(script.getDescriptor().getName() + ETools.getEClassifierShortPath(eClass), script);
                    }
                }
            }
        }

        return script;
    }

    /**
     * Gets script from supertype if applicable.
     * 
     * @param superTypes
     *            on which searched script could be available
     * @param name
     *            script name
     * @return script from supertype if applicable
     */
    private Script getScriptFromSupertype(EList<EClass> superTypes, String name) {
        EClass superType = null;
        Script script = null;
        for (final Iterator<EClass> it = superTypes.iterator(); it.hasNext() && script == null;) {
            superType = it.next();
            script = getScript(superType, name);
        }
        if (script == null) {
            for (final Iterator<EClass> it = superTypes.iterator(); it.hasNext() && script == null;) {
                script = getScriptFromSupertype(it.next().getESuperTypes(), name);
            }
        }
        return script;
    }

    /**
     * Get the Script with a given name that can be applied to the given
     * EClassifier in the current template only (imports not scanned).
     * 
     * @param eClass
     *            the EClassifier on witch the Script can be applied
     * @param name
     *            the name of the script we are looking for
     * @return the matching Script if founded, null otherwise
     */
    private Script getScript(EClassifier eClass, String name) {
        if (scripts == null) {
            scripts = new ConcurrentHashMap<String, Script>();
            String nsURI = null;
            for (final Resource resource : getImports()) {
                if (resource instanceof Metamodel) {
                    nsURI = ((Metamodel) resource).getName();
                }
            }
            EPackage metamodel = EPackage.Registry.INSTANCE.getEPackage(nsURI);
            if (metamodel == null) {
                metamodel = ETools.ecore2EPackage(nsURI, true);
            }
            // scan Scripts of the current template
            for (final Script script : getScripts()) {
                if (script.getDescriptor() != null) {
                    final EClassifier scriptType = ETools.getEClassifier(metamodel, script.getDescriptor().getType());
                    if (scriptType != null) {
                        scripts.put(script.getDescriptor().getName() + ETools.getEClassifierShortPath(scriptType), script);
                    }
                }
            }
        }
        return scripts.get(name + ETools.getEClassifierShortPath(eClass));
    }

    /**
     * Getter for properties.
     * 
     * @return properties attribute
     */
    public ScriptProperties getProperties() {
        return properties;
    }

    /**
     * Getter for available services.
     * 
     * @return available services
     */
    public List<TService> getServices() {
        return javaServices;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        String ls = "\n"; //$NON-NLS-1$
        String space = " "; //$NON-NLS-1$
        final StringBuffer buffer = new StringBuffer(""); //$NON-NLS-1$
        if (!getImports().isEmpty()) {
            buffer.append(TemplateConstants.IMPORT_BEGIN);
            buffer.append(ls);
            for (Resource resource : getImports()) {
                if (resource instanceof Metamodel && resource.getName() != null) {
                    buffer.append(TemplateConstants.MODELTYPE_WORD);
                    buffer.append(space);
                    buffer.append(resource.getName());
                    buffer.append(ls);
                } else if (resource.getName() != null) {
                    buffer.append(TemplateConstants.IMPORT_WORD);
                    buffer.append(space);
                    buffer.append(resource.getName());
                    buffer.append(ls);
                }
            }
            buffer.append(ls);
        }

        for (Script script : getScripts()) {
            buffer.append(script.toString());
            buffer.append(ls);
            buffer.append(ls);
        }
        return buffer.toString();
    }
}
