/*
 * Copyright (c) 2005-2008 Obeo.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Obeo - initial API and implementation
 */
package org.acceleo.gen;

import fr.obeo.acceleo.gen.template.eval.ILauncherManager;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Map;

import org.acceleo.engine.EMT;
import org.acceleo.engine.IModel;
import org.acceleo.gen.extension.Extension;
import org.acceleo.gen.merge.MergeTools;
import org.acceleo.tools.resources.Resources;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EObject;

/**
 * This class can generate files for a single EMT file.
 * 
 * @author <a href="mailto:yvan.lussaud@obeo.fr">Yvan Lussaud</a>
 * 
 */
public class AcceleoGenerate {

	/**
	 * Generation root folder path. The default value is the present working
	 * directory followed by generated.
	 */
	private String genRootDir = System.getProperty("user.dir") + File.separator + "generated"; //$NON-NLS-1$ //$NON-NLS-2$

	/**
	 * Generates files from a source model by using EMT scenario.
	 * 
	 * @param emt
	 *            EMT
	 * @param eObj
	 *            Source model
	 * @param recursive
	 *            is a recursive generation
	 * @param mode
	 *            Execution mode
	 * @return The list of generated file paths
	 * @throws Exception
	 *             if model generation failed
	 */
	public List<String> generate(EMT emt, EObject eObj, boolean recursive, ILauncherManager mode) throws Exception {
		List<String> res = new ArrayList<String>();

		emt.setThreadsNumber(Extension.getThreadsNumber());
		emt.setEObjectsPerThread(Extension.getEObjectsPerThread());
		Map<String, String> generatedFiles = emt.gen(eObj, recursive);
		for (String fileName : generatedFiles.keySet()) {
			res.addAll(writeFile(fileName, generatedFiles.get(fileName), mode));
		}
		return res;
	}

	/**
	 * Generates files from a source model by using EMT scenario.
	 * 
	 * @param emt
	 *            EMT
	 * @param model
	 *            Source
	 * @param recursive
	 *            is a recursive
	 * @param mode
	 *            Execution mode
	 * @return The list of generated file paths
	 * @throws Exception
	 *             if model generation failed
	 */
	public List<String> generate(EMT emt, IModel model, boolean recursive, ILauncherManager mode) throws Exception {
		List<String> res = new ArrayList<String>();

		emt.setThreadsNumber(Extension.getThreadsNumber());
		emt.setEObjectsPerThread(Extension.getEObjectsPerThread());
		Map<String, String> generatedFiles = emt.gen(model, recursive);
		String errors = emt.getLogs().toString();
		if (errors.length() > 0) {
			String errorLogName = genRootDir + "error.log"; //$NON-NLS-1$
			writeFile(errorLogName, errors, mode);
			System.out.println(AcceleoGenerateMessages.getString("AcceleoGenerate.ErrorsReported", new Object[] { errorLogName, })); //$NON-NLS-1$
		}
		for (String fileName : generatedFiles.keySet()) {
			res.addAll(writeFile(fileName, generatedFiles.get(fileName), mode));
		}
		return res;
	}

	/**
	 * Generates files from a source model by using EMT scenario.
	 * 
	 * @param emtFileName
	 *            EMT file path
	 * @param modelFileName
	 *            Source model file path
	 * @param mode
	 *            Execution mode
	 * @return The list of generated file paths
	 * @throws Exception
	 *             if model generation failed
	 */
	// TODO exception - exception management
	public List<String> generate(String emtFileName, String modelFileName, ILauncherManager mode) throws Exception {
		List<String> res = null;

		URI templateURI = Resources.getResourceURI(emtFileName);
		EMT emt = new EMT(Extension.getResourceSet(), templateURI, Extension.getClassLoader());
		URI modelURI = Resources.getResourceURI(modelFileName);
		Model model = new Model(Extension.getResourceSet(), emt.getMetamodel().getPackageClass(), modelURI);
		res = generate(emt, model, true, mode);

		return res;
	}

	/**
	 * rootDir setter.
	 * 
	 * @param rootDir
	 *            Generation root folder path
	 */
	public void setGenRootDir(String rootDir) {
		if (!rootDir.endsWith(File.separator)) {
			rootDir = rootDir + File.separator;
		}
		genRootDir = rootDir;
	}

	/**
	 * Serializes file content and returns the list of file paths.
	 * 
	 * @param fileName
	 *            Name of the file
	 * @param fileContent
	 *            File content
	 * @param mode
	 *            Execution mode
	 * @return The list of file paths
	 * @throws IOException
	 *             If an I/O error occurs
	 */
	public List<String> writeFile(String fileName, String fileContent, ILauncherManager mode) throws IOException {
		List<String> ret = new ArrayList<String>();
		StringBuffer fileContentSB = new StringBuffer(fileContent);
		File outDir = new File(genRootDir + File.separator + fileName.substring(0, java.lang.Math.max(0, fileName.lastIndexOf(File.separatorChar))));
		File outFile = new File(genRootDir + File.separator + fileName);
		outDir.mkdirs();

		// TODO YLU test merge "user code"
		if (outFile.exists()) {
			StringBuffer oldBuffer = Resources.getAbsolutFileContent(outFile.getAbsolutePath());
			if (oldBuffer.length() > 0) {
				String lostCode = MergeTools.merge(outFile, fileContentSB, oldBuffer, MergeTools.DEFAULT_USER_BEGIN, MergeTools.DEFAULT_USER_END);
				if (lostCode.length() > 0) {
					lostCode = '[' + AcceleoGenerateMessages.getString("AcceleoGenerate.LostCode") + "] " + new Date().toString() + '\n' + lostCode + '\n'; //$NON-NLS-1$ //$NON-NLS-2$
					// TODO what's the difference between target and
					// targetGenerated ? (Acceleo 2.x CGenerate)
					if (mode.getMode() == ILauncherManager.PREVIEW_MODE) {
						File lostCodeFile = new File(outFile.getAbsolutePath().concat(MergeTools.LOST_FILE_EXTENSION));
						FileWriter f = new FileWriter(lostCodeFile);
						BufferedWriter out = new BufferedWriter(f);
						out.append(lostCode);
						out.close();
						ret.add(lostCodeFile.getAbsolutePath());
					} else if (mode.getMode() != ILauncherManager.PHANTOM_MODE) {
						File lostCodeFile = new File(outFile.getAbsolutePath().concat(MergeTools.LOST_FILE_EXTENSION));
						FileWriter f = new FileWriter(lostCodeFile);
						BufferedWriter out = new BufferedWriter(f);
						out.append(lostCode);
						out.close();
						ret.add(lostCodeFile.getAbsolutePath());
					}
				}
			}
		}
		// TODO manage LaunchManager - Create a generation method for each mode
		// ?

		Resources.writeFile(outFile, fileContentSB.toString());
		ret.add(outFile.getAbsolutePath());
		return ret;
	}
}
