/*
 * Copyright (c) 2005-2008 Obeo.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Obeo - initial API and implementation
 */

package org.acceleo.gen.merge;

import java.io.File;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.Map.Entry;

import org.acceleo.gen.AcceleoGenerateMessages;
import org.acceleo.gen.extension.Extension;
import org.acceleo.gen.extension.vm.IOverdoneCodeStrategy;
import org.acceleo.tools.strings.Int2;
import org.acceleo.tools.template.TemplateConstants;

/**
 * Round-trip support for templates.
 * 
 * @author www.obeo.fr
 * 
 */
public final class MergeTools {

	/**
	 * Lost code file extension.
	 */
	public static final String LOST_FILE_EXTENSION = "lost"; //$NON-NLS-1$

	/**
	 * Default tag used to start the user code.
	 */
	public static final String DEFAULT_USER_BEGIN = TemplateConstants.DEFAULT_USER_BEGIN;

	/**
	 * Default tag used to stop the user code.
	 */
	public static final String DEFAULT_USER_END = TemplateConstants.DEFAULT_USER_END;

	/**
	 * The identifier of the extension point use to overdone the user code after
	 * the generation.
	 */
	public static final String OVERDONE_CODE_EXTENSION_ID = "org.acceleo.chain.overdonecode"; //$NON-NLS-1$

	/**
	 * Constant for the new line string.
	 */
	private static final String NEW_LINE = "\n"; //$NON-NLS-1$

	/**
	 * Private constructor.
	 */
	private MergeTools() {
		// nothing to do here
	}

	/**
	 * Merges the two buffers. It inserts the user code of the old buffer in the
	 * new buffer.
	 * 
	 * @param file
	 *            is the file to generate
	 * @param newBuffer
	 *            is the new buffer. It will contain the result of the merge
	 *            after call return.
	 * @param oldBuffer
	 *            is the old buffer
	 * @param beginTag
	 *            is used to start the user code
	 * @param endTag
	 *            is used to stop the user code
	 * @return lost code
	 */
	public static String merge(File file, StringBuffer newBuffer, StringBuffer oldBuffer, String beginTag, String endTag) {
		StringBuffer lost = new StringBuffer(""); //$NON-NLS-1$
		Map<String, Int2> oldName2Pos = tagName2Pos(oldBuffer, beginTag, endTag, lost);
		Map<Int2, String> newPos2Name = tagPos2Name(newBuffer, beginTag, endTag);
		List<Object[]> notFound = new ArrayList<Object[]>();
		for (Entry<Int2, String> entry : newPos2Name.entrySet()) {
			Int2 newPos = entry.getKey();
			String name = entry.getValue();
			Int2 oldPos = oldName2Pos.get(name);
			if (oldPos != null) {
				String oldText = oldBuffer.substring(oldPos.b(), oldPos.e());
				newBuffer.delete(newPos.b(), newPos.e());
				newBuffer.insert(newPos.b(), oldText);
				int shift = oldText.length() - (newPos.e() - newPos.b());
				oldName2Pos.remove(name);
				if (notFound.size() > 0) {
					for (Object[] notFoundEntry : notFound) {
						int b = ((Int2) notFoundEntry[0]).b() + shift;
						int e = ((Int2) notFoundEntry[0]).e() + shift;
						notFoundEntry[0] = new Int2(b, e);
					}
				}
			} else {
				notFound.add(new Object[] { entry.getKey(), entry.getValue() });
			}
		}
		if (notFound.size() == 1 && oldName2Pos.size() == 1) {
			List<String> toRemove = new ArrayList<String>();
			for (Entry<String, Int2> entry : oldName2Pos.entrySet()) {
				String oldName = entry.getKey();
				Iterator<Object[]> notFoundIt = notFound.iterator();
				while (notFoundIt.hasNext()) {
					Object[] notFoundEntry = notFoundIt.next();
					String newName = (String) notFoundEntry[1];
					if (nameSimilarityMetric(oldName, newName) > 0.5) {
						Int2 oldPos = entry.getValue();
						Int2 newPos = (Int2) notFoundEntry[0];
						String oldText = oldBuffer.substring(oldPos.b(), oldPos.e());
						newBuffer.delete(newPos.b(), newPos.e());
						newBuffer.insert(newPos.b(), oldText);
						toRemove.add(oldName);
						notFoundIt.remove();
						lost.append('\n');
						lost.append(AcceleoGenerateMessages.getString("MergeTools.MovedCode", new Object[] { oldName, newName, })); //$NON-NLS-1$
						lost.append('\n');
						lost.append(oldName);
						lost.append(oldBuffer.substring(oldPos.b(), oldPos.e()));
						lost.append('\n');
						break;
					}
				}
			}
			for (String key : toRemove) {
				oldName2Pos.remove(key);
			}
		}
		for (Entry<String, Int2> entry : oldName2Pos.entrySet()) {
			String name = entry.getKey();
			Int2 oldPos = entry.getValue();
			lost.append('\n');
			lost.append(name);
			lost.append(oldBuffer.substring(oldPos.b(), oldPos.e()));
			lost.append('\n');
			int eOld = oldBuffer.indexOf(NEW_LINE, oldPos.e() + 1);
			if (eOld == -1) {
				eOld = oldBuffer.length();
			}
			lost.append(oldBuffer.substring(oldPos.e(), eOld).trim());
			lost.append('\n');
		}
		keepOverdoneCode(file, oldBuffer, newBuffer, lost);
		return lost.toString();
	}

	private static Map<Int2, String> tagPos2Name(StringBuffer buffer, String beginTag, String endTag) {
		Map<Int2, String> tags = new TreeMap<Int2, String>(new Comparator<Int2>() {
			public int compare(Int2 arg0, Int2 arg1) {
				if (arg0.b() > arg1.b()) {
					return -1;
				} else {
					return 1;
				}
			}
		});
		int i = 0;
		while (i < buffer.length()) {
			Int2 bUser = nextTag(buffer, beginTag, i);
			Int2 eUser = nextTag(buffer, endTag, bUser.e());
			if (eUser.b() > -1) {
				// position and name
				Int2 pos = new Int2(bUser.e(), eUser.b());
				String name = buffer.substring(bUser.b(), bUser.e()).trim();
				tags.put(pos, name);
				i = eUser.e();
			} else {
				break;
			}
		}
		return tags;
	}

	private static Map<String, Int2> tagName2Pos(StringBuffer buffer, String beginTag, String endTag, StringBuffer lost) {
		Map<String, Int2> tags = new HashMap<String, Int2>();
		int i = 0;
		while (i < buffer.length()) {
			Int2 bUser = nextTag(buffer, beginTag, i);
			Int2 eUser = nextTag(buffer, endTag, bUser.e());
			if (eUser.b() > -1) {
				// position and name
				Int2 pos = new Int2(bUser.e(), eUser.b());
				String name = buffer.substring(bUser.b(), bUser.e()).trim();
				// tag exists?
				if (tags.get(name) == null) {
					tags.put(name, pos);
				} else {
					lost.append('\n');
					lost.append(AcceleoGenerateMessages.getString("MergeTools.DuplicatedTag", new Object[] { name, })); //$NON-NLS-1$
					lost.append(buffer.substring(pos.b(), pos.e()));
					lost.append('\n');
				}
				i = eUser.e();
			} else {
				break;
			}
		}
		return tags;
	}

	/**
	 * Heuristic for string matching.
	 * 
	 * @param str1
	 *            the first string
	 * @param str2
	 *            the second string
	 * @return the level of similarity
	 */
	private static double nameSimilarityMetric(String str1, String str2) {
		double result = 0;
		if (str1 == null || str2 == null) {
			result = 0;
		} else {
			if (str1.length() == 1 || str2.length() == 1) {
				if (str1.equals(str2)) {
					result = 1.0;
				} else {
					result = 0;
				}
			} else {
				List<String> pairs1 = pairs(str1);
				List<String> pairs2 = pairs(str2);
				double union = pairs1.size() + pairs2.size();
				if (union == 0) {
					result = 0;
				} else {
					pairs1.retainAll(pairs2);
					int inter = pairs1.size();
					result = inter * 2.0 / union;
					if (result > 1) {
						result = 1;
					}
					final double similarityValue = 0.999999;
					if (result == 1.0 && !str1.equals(str2)) {
						result = similarityValue;
					}
				}
			}
		}
		return result;
	}

	private static List<String> pairs(String source) {
		List<String> result = new LinkedList<String>();
		if (source != null) {
			for (int i = 0; i < source.length() - 1; i = i + 1) {
				result.add(source.toUpperCase().substring(i, i + 2));
				if (source.length() % 2 == 1 && source.length() > 1) {
					result.add(source.toUpperCase().substring(source.length() - 2, source.length() - 1));
				}
			}
		}
		return result;
	}

	/**
	 * Gets the position of the next tag in the buffer.
	 * 
	 * @param buffer
	 *            is the buffer
	 * @param tag
	 *            is the tag to search
	 * @param pos
	 *            is the beginning index
	 * @return the position of the next tag
	 */
	public static Int2 nextTag(StringBuffer buffer, String tag, int pos) {
		if (pos > -1) {
			int b = buffer.indexOf(tag, pos);
			if (b > -1) {
				// End of the tag
				int eUser = -1;
				String[] eDelimiters = { NEW_LINE };
				for (int i = 0; i < eDelimiters.length; i++) {
					int eUserTmp = buffer.indexOf(eDelimiters[i], b + tag.length());
					if (eUser == -1 || eUserTmp > -1 && eUserTmp < eUser) {
						eUser = eUserTmp;
					}
				}
				if (eUser == -1) {
					eUser = buffer.length();
				}
				// Begin of the tag
				int bUser = -1;
				String[] bDelimiters = { NEW_LINE };
				for (int i = bDelimiters.length - 1; i >= 0; i--) {
					int bUserTmp = buffer.lastIndexOf(bDelimiters[i], b);
					if (bUser == -1 || bUserTmp > -1 && bUserTmp > bUser) {
						bUser = bUserTmp;
					}
				}
				if (bUser == -1) {
					bUser = 0;
				}
				return new Int2(bUser, eUser);
			}
		}
		return Int2.NOT_FOUND;
	}

	/**
	 * Indicates if the offset is between the given tags.
	 * 
	 * @param buffer
	 *            is the buffer
	 * @param beginTag
	 *            is the beginning tag
	 * @param endTag
	 *            is the ending tag
	 * @param index
	 *            is the offset to test
	 * @return true if the offset is between the given tags
	 */
	public static boolean isBetweenTags(StringBuffer buffer, String beginTag, String endTag, int index) {
		int pos = buffer.substring(0, index).lastIndexOf(NEW_LINE) + 1;
		Int2 bUser = nextTag(buffer, beginTag, pos);
		Int2 eUser = nextTag(buffer, endTag, pos);
		return eUser.b() > -1 && (bUser.b() == -1 || eUser.b() < bUser.b());
	}

	/**
	 * Call extension to keep overdone code.
	 * 
	 * @param file
	 *            the file to generate
	 * @param oldBuffer
	 *            the old file content
	 * @param newBuffer
	 *            the new file content
	 * @param lost
	 *            the lost content
	 */
	private static void keepOverdoneCode(File file, StringBuffer oldBuffer, StringBuffer newBuffer, StringBuffer lost) {
		if (file != null && file.exists()) {
			String fileExt = file.getName().substring(file.getName().lastIndexOf('.'), file.getName().length());
			IOverdoneCodeStrategy strategy = Extension.getOverdoneStrategy(fileExt);
			if (strategy == null) {
				strategy = Extension.getOverdoneStrategy("*"); //$NON-NLS-1$
			}
			if (strategy != null) {
				strategy.keepOverdoneCode(file, oldBuffer, newBuffer, lost);
			}
		}
	}
}
