/*
 * Copyright (c) 2005-2008 Obeo.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Obeo - initial API and implementation
 */
package org.acceleo.parser.template.statements;

import org.acceleo.template.core.Template;
import org.acceleo.template.statements.Statement;

import java.util.ArrayList;
import java.util.List;

import org.acceleo.parser.AcceleoParserMessages;
import org.acceleo.parser.template.AbstractParser;
import org.acceleo.parser.template.TemplateSyntaxException;
import org.acceleo.tools.strings.Int2;
import org.acceleo.tools.strings.TextSearch;
import org.acceleo.tools.template.TemplateConstants;

/**
 * Acceleo's Statement parser.
 * 
 * @author <a href="mailto:yvan.lussaud@obeo.fr">Yvan Lussaud</a>
 */
public class StatementParser extends AbstractParser {
	/** Constant for local unknown positions. */
	private static final int UNKNOWN_POSITION = -2;

	/** Constant for the new line string. */
	private static final String NEW_LINE = "\n"; //$NON-NLS-1$

	/**
	 * No public access to the constructor.
	 */
	protected StatementParser() {
		// nothing to do here
	}

	/**
	 * {@inheritDoc}
	 */
	public static Statement fromString(int offset, String buffer, Int2 range, Template template) throws TemplateSyntaxException {
		// TODO check implement if needed
		return null;
	}

	/**
	 * Formats the text of a template in a script.
	 * 
	 * @param buffer
	 *            is the full text of a script
	 * @param range
	 *            delimits the part of the text to parse for this template
	 * @param nbReturn
	 *            is the number of return character to ignore at the end of the
	 *            template
	 * @return is the new limits that delimits the part of the text to parse for
	 *         this template
	 */
	public static Int2 formatTemplate(String buffer, Int2 range, int nbReturn) {
		Int2 res = range;

		String text = ""; //$NON-NLS-1$
		int posBegin = -1;
		int posEnd = posBegin - 1;

		if (range.b() == -1 || range.b() > range.e()) {
			res = Int2.NOT_FOUND;
		} else {
			text = buffer.substring(range.b(), range.e());
			// Search first valid char
			posBegin = searchFirstValidChar(text);
			// Search last valid char
			if (text.length() > 0) {
				posEnd = searchLastValidChar(text, posBegin, nbReturn);
			}
		}

		if (posEnd >= posBegin) {
			boolean formatWithComment = false;
			String sub = text.substring(posBegin, posEnd);
			if (sub.startsWith(TemplateConstants.COMMENT_BEGIN) && sub.indexOf(TemplateConstants.COMMENT_END) > -1) {
				posBegin = text.indexOf(TemplateConstants.COMMENT_END, posBegin) + TemplateConstants.COMMENT_END.length();
				formatWithComment = true;
				if (posEnd > posBegin) {
					sub = sub.substring(posBegin);
				} else {
					sub = ""; //$NON-NLS-1$
				}
			}
			if (sub.endsWith(TemplateConstants.COMMENT_END) && sub.lastIndexOf(TemplateConstants.COMMENT_BEGIN, posEnd) > -1) {
				posEnd = text.lastIndexOf(TemplateConstants.COMMENT_BEGIN, posEnd);
				formatWithComment = true;
			}
			if (formatWithComment) {
				res = formatTemplate(buffer, new Int2(range.b() + posBegin, range.b() + posEnd), nbReturn);
			} else {
				res = new Int2(range.b() + posBegin, range.b() + posEnd);
			}
		}

		return res;
	}

	/**
	 * Search the position of the first valid char.
	 * 
	 * @param text
	 *            the text to parse
	 * @return the position of the first valid char
	 */
	private static int searchFirstValidChar(String text) {
		int posBegin = text.indexOf(NEW_LINE);
		if (posBegin == -1 || text.substring(0, posBegin).trim().length() > 0) {
			posBegin = 0;
		} else {
			posBegin++;
		}
		return posBegin;
	}

	/**
	 * Search the position of the last valid char.
	 * 
	 * @param text
	 *            the text to parse
	 * @param posBegin
	 *            the position to start
	 * @param nbReturn
	 *            is the number of return character to ignore at the end of the
	 *            template
	 * @return the position of the last valid char
	 */
	private static int searchLastValidChar(String text, int posBegin, int nbReturn) {
		int n = 0;
		int endLine = text.length();
		int posEnd = text.length() - 1;
		boolean stop = false;
		while (!stop) {
			char c = text.charAt(posEnd);
			if (nbReturn == 0 && c == '\n') {
				posEnd++;
				stop = true;
			} else if (c == '\n' && n < nbReturn) {
				n++;
				endLine = posEnd;
			} else if (c == '\r') {
				endLine = posEnd;
			} else if (c == ' ' || c == '\t') {
				// nothing to do
			} else {
				// Keep space at end
				posEnd = endLine;
				stop = true;
			}
			if (!stop) {
				if (posEnd > posBegin) {
					posEnd--;
				} else {
					if (nbReturn != 0) {
						posEnd = endLine;
					}
					stop = true;
				}
			}
		}
		return posEnd;
	}

	/**
	 * It checks the syntax and creates a template for the given part of the
	 * text. The part of the text to be parsed is delimited by the given limits.
	 * 
	 * @param offset
	 *            offset to add to the found element position
	 * @param buffer
	 *            is the textual representation of the script that contains
	 *            templates
	 * @param range
	 *            delimits the part of the text to be parsed for this template
	 * @param template
	 *            is the generator's configuration used to apply the new
	 *            template to model's objects
	 * @return the new template
	 * @throws TemplateSyntaxException
	 *             if the statements can't be parsed
	 */
	public static List<Statement> read(int offset, String buffer, Int2 range, Template template) throws TemplateSyntaxException {
		List<Statement> block = new ArrayList<Statement>();
		if (buffer != null) {
			int[] pos = new int[4];
			for (int i = 0; i < pos.length; i++) {
				pos[i] = UNKNOWN_POSITION;
			}
			int i = range.b();
			while (i < range.e()) {
				if (pos[0] != -1 && i > pos[0]) {
					pos[0] = TextSearch.getDefaultSearch().indexIn(buffer, TemplateConstants.COMMENT_BEGIN, i, range.e()).b();
				}
				if (pos[1] != -1 && i > pos[1]) {
					pos[1] = TextSearch.getDefaultSearch().indexIn(buffer, TemplateConstants.IF_BEGIN, i, range.e()).b();
				}
				if (pos[2] != -1 && i > pos[2]) {
					pos[2] = TextSearch.getDefaultSearch().indexIn(buffer, TemplateConstants.FOR_BEGIN, i, range.e()).b();
				}
				if (pos[3] != -1 && i > pos[3]) {
					pos[3] = TextSearch.getDefaultSearch().indexIn(buffer, TemplateConstants.FEATURE_BEGIN, i, range.e()).b();
				}
				int iTab = indexOfMin(pos);
				if (iTab == 0) {
					i = readStatement(offset, buffer, TemplateConstants.COMMENT_BEGIN, TemplateConstants.COMMENT_END, new Int2(i, range.e()), block, template);
				} else if (iTab == 1) {
					i = readStatement(offset, buffer, TemplateConstants.IF_BEGIN, TemplateConstants.IF_END, new Int2(i, range.e()), block, template);
				} else if (iTab == 2) {
					i = readStatement(offset, buffer, TemplateConstants.FOR_BEGIN, TemplateConstants.FOR_END, new Int2(i, range.e()), block, template);
				} else if (iTab == 3) {
					i = readStatement(offset, buffer, TemplateConstants.FEATURE_BEGIN, TemplateConstants.FEATURE_END, new Int2(i, range.e()), block, template);
				} else { // -1
					Int2 posText = new Int2(i, range.e());
					block.add(TextParser.fromString(offset, buffer, posText, null));
					i = range.e();
				}
			}
		}
		return block;
	}

	/**
	 * Gets the index of the minimum integer in the given table.
	 * 
	 * @param pos
	 *            is the table
	 * @return the index of the minimum integer
	 */
	protected static int indexOfMin(int[] pos) {
		int index = -1;
		int min = Integer.MAX_VALUE;
		for (int i = 0; i < pos.length; i++) {
			if (pos[i] > -1 && pos[i] < min) {
				index = i;
				min = pos[i];
			}
		}
		return index;
	}

	/**
	 * Reads the statements in the buffer, starting at a specified position, and
	 * returns the ending index.
	 * 
	 * @param offset
	 *            offset to add to the found element position
	 * @param buffer
	 *            is the textual representation of the script that contains
	 *            templates
	 * @param tagBegin
	 *            is the begin tag of the statement
	 * @param tagEnd
	 *            is the end tag of the statement
	 * @param range
	 *            delimits the part of the text to be parsed
	 * @param block
	 *            is the list of statements where to add the parsed statements
	 *            is the current template
	 * @param template
	 *            is the owning template
	 * @return the ending index, or -1
	 * @throws TemplateSyntaxException
	 *             if the statements can't be parsed
	 */
	protected static int readStatement(int offset, String buffer, String tagBegin, String tagEnd, Int2 range, List<Statement> block, Template template) throws TemplateSyntaxException {
		Int2 end = Int2.NOT_FOUND;
		Int2 begin = TextSearch.getDefaultSearch().indexIn(buffer, tagBegin, range.b(), range.e());
		// ASSERT (begin.b > -1)
		if (tagBegin == TemplateConstants.COMMENT_BEGIN) {
			end = TextSearch.getDefaultSearch().blockIndexEndIn(buffer, tagBegin, tagEnd, begin.b(), range.e(), false);
		} else {
			end = TextSearch.getDefaultSearch().blockIndexEndIn(buffer, tagBegin, tagEnd, begin.b(), range.e(), true, null, TemplateConstants.INHIBS_STATEMENT);
		}
		if (end.b() > -1) {
			boolean untab = tagBegin != TemplateConstants.FEATURE_BEGIN && isFirstSignificantOfLine(buffer, begin.b()) && isLastSignificantOfLine(buffer, end.e());
			if (begin.b() > range.b()) {
				Int2 posText = new Int2(range.b(), getIEndPos(buffer, untab, begin.b(), range.b()));
				block.add(TextParser.fromString(offset, buffer, posText, null));
			}
			if (tagBegin == TemplateConstants.IF_BEGIN) {
				block.add(IfParser.fromString(offset, buffer, new Int2(begin.e(), end.b()), template));
			}
			if (tagBegin == TemplateConstants.FOR_BEGIN) {
				block.add(ForParser.fromString(offset, buffer, new Int2(begin.e(), end.b()), template));
			}
			if (tagBegin == TemplateConstants.FEATURE_BEGIN) {
				block.add(FeatureParser.fromString(offset, buffer, new Int2(begin.e(), end.b()), template));
			}
			if (tagBegin == TemplateConstants.COMMENT_BEGIN) {
				block.add(CommentParser.fromString(offset, buffer, new Int2(begin.e(), end.b()), template));
			}
			if (untab) {
				// Delete (\s|\t)*\n after <%}%>
				int iNextLine = TextSearch.getDefaultSearch().indexIn(buffer, NEW_LINE, end.e(), range.e()).e();
				if (iNextLine == -1) {
					iNextLine = range.e();
				}
				if (buffer.substring(end.e(), iNextLine).trim().length() == 0) {
					return iNextLine;
				}
			}
			return end.e();
		} else {
			block.add(TextParser.fromString(offset, buffer, range, null));
			throw new TemplateSyntaxException(AcceleoParserMessages.getString("Template.UnclosedTag", new Object[] { tagEnd, tagBegin, }), template, new Int2(begin.b(), range.e())); //$NON-NLS-1$
		}
	}

	/**
	 * Give the position of the iEndText.
	 * 
	 * @param buffer
	 *            the buffer to parse
	 * @param untab
	 *            should remove tabs or not
	 * @param beginBegin
	 *            begins of the begin
	 * @param rangeBegin
	 *            begins of the range
	 * 
	 * @return the position of the iEndText
	 */
	private static int getIEndPos(String buffer, boolean untab, int beginBegin, int rangeBegin) {
		int iEndText = beginBegin;
		if (untab) {
			// Delete (\s|\t)* before <%if..., <%for...
			int iPrevLine = TextSearch.getDefaultSearch().lastIndexIn(buffer, NEW_LINE, rangeBegin, beginBegin).e();
			if (iPrevLine == -1) {
				iPrevLine = rangeBegin;
			}
			if (buffer.substring(iPrevLine, beginBegin).trim().length() == 0) {
				iEndText = iPrevLine;
			}
		}
		return iEndText;
	}

	/**
	 * Is the parsed statement the first significant of the line?
	 * 
	 * @param buffer
	 *            the buffer containing the text representation
	 * @param index
	 *            the current position of the statement
	 * @return if it's the first significant of the line or not
	 */
	private static boolean isFirstSignificantOfLine(String buffer, int index) {
		boolean res = true;
		boolean stop = false;
		int localIndex = index;
		if (localIndex > 0 && localIndex < buffer.length()) {
			localIndex--;
			while (localIndex >= 0 && !stop) {
				char c = buffer.charAt(localIndex);
				if (c == '\n') {
					res = true;
					stop = true;
				} else if (c == ' ' || c == '\t' || c == '\r') {
					localIndex--;
				} else {
					res = false;
					stop = true;
				}
			}
		}
		if (stop) {
			return res;
		} else {
			return true;
		}
	}

	/**
	 * Is the parsed statement the last significant of the line?
	 * 
	 * @param buffer
	 *            the buffer containing the text representation
	 * @param index
	 *            the current position of the statement
	 * @return if it's the last significant of the line or not
	 */
	private static boolean isLastSignificantOfLine(String buffer, int index) {
		boolean res = true;
		boolean stop = false;
		int localIndex = index;
		if (localIndex >= 0) {
			while (localIndex < buffer.length() && !stop) {
				char c = buffer.charAt(localIndex);
				if (c == '\n') {
					res = true;
					stop = true;
				} else if (c == ' ' || c == '\t' || c == '\r') {
					localIndex++;
				} else {
					res = false;
					stop = true;
				}
			}
		}
		if (stop) {
			return res;
		} else {
			return true;
		}
	}
}
