/*
 * Copyright (c) 2005-2008 Obeo.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Obeo - initial API and implementation
 */
package org.acceleo.tools.factory;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.acceleo.tools.AcceleoToolsMessages;
import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.Enumerator;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;

/**
 * Utility class for eObjects.
 * 
 * @author www.obeo.fr
 * 
 */
public final class EFactory {

	/**
	 * Private constructor.
	 */
	private EFactory() {
		// nothing to do here
	}

	/**
	 * Indicates if the object is instance of the class whose name is given.
	 * <p>
	 * Samples :
	 * <p>
	 * An instance of java.resources.Folder return true if name equals "Folder"
	 * or "resources.Folder".
	 * <p>
	 * An instance of java.resources.Folder return true if name equals "File"
	 * and Folder inherits File.
	 * 
	 * @param object
	 *            is the object
	 * @param name
	 *            is the class name
	 * @return true if the object is instance of the class whose name is given
	 */
	public static boolean eInstanceOf(EObject object, String name) {
		if (object == null) {
			return name == null;
		}
		return eInstanceOf(object.eClass(), name);
	}

	/**
	 * Returns true if name is equals to the name of eClass.
	 * 
	 * @param eClass
	 *            meta class of an object
	 * @param name
	 *            is the class name
	 * @return true if name is equals to the name of eClass.
	 */
	private static boolean eInstanceOf(EClass eClass, String name) {
		boolean isEInstanceOf;
		if (name.indexOf('.') == -1 && name.equals(eClass.getName())) {
			isEInstanceOf = true;
		} else {
			String instanceClassName = '.' + eClass.getInstanceClassName();
			String endsWith = '.' + name;
			if (instanceClassName.endsWith(endsWith)) {
				isEInstanceOf = true;
			} else {
				for (EClass eSuperClass : eClass.getESuperTypes()) {
					if (eInstanceOf(eSuperClass, name)) {
						return true;
					}
				}
				isEInstanceOf = false;
			}
		}
		return isEInstanceOf;
	}

	/**
	 * Gets the value of the given feature of the object, as an EObject.
	 * 
	 * @param object
	 *            is the object
	 * @param name
	 *            is the feature name
	 * @return the value or null if it isn't an EObject
	 * @throws FactoryException
	 *             if the given feature could not be accessed.
	 */
	public static EObject eGetAsEObject(EObject object, String name) throws FactoryException {
		Object eGet = eGet(object, name);
		if (eGet != null && eGet instanceof EObject) {
			return (EObject) eGet;
		} else {
			return null;
		}
	}

	/**
	 * Gets the value of the given feature of the object, as a String.
	 * 
	 * @param object
	 *            is the object
	 * @param name
	 *            is the feature name
	 * @return the value or null if it isn't a String
	 * @throws FactoryException
	 *             if the given feature could not be accessed.
	 */
	public static String eGetAsString(EObject object, String name) throws FactoryException {
		Object eGet = eGet(object, name);
		if (eGet != null) {
			return eGet.toString();
		} else {
			return null;
		}
	}

	/**
	 * Gets the value of the given feature of the object.
	 * 
	 * @param object
	 *            is the object
	 * @param name
	 *            is the feature name, or a method defined on EObject like
	 *            'eClass', 'eResource', 'eContainer', 'eContainingFeature',
	 *            'eContainmentFeature', 'eContents', 'eAllContents',
	 *            'eCrossReferences'
	 * @return the value of the given feature of the object
	 * @throws FactoryException
	 *             if the given feature could not be accessed.
	 */
	@SuppressWarnings("unchecked")
	public static Object eGet(EObject object, String name) throws FactoryException {
		Object result;
		Object featureValue;
		try {
			EStructuralFeature feature = eStructuralFeature(object, name);
			result = object.eGet(feature);
		} catch (FactoryException eGet) {
			try {
				result = eCall(object, name, null);
			} catch (FactoryException e) {
				throw eGet;
			}
		}
		if (result != null && result instanceof Enumerator) {
			featureValue = ((Enumerator) result).getLiteral();
		} else if (result != null && result instanceof BasicEList) {
			List<Object> list = new ArrayList<Object>();
			Iterator<Object> enums = ((BasicEList<Object>) result).iterator();
			while (enums.hasNext()) {
				Object next = enums.next();
				if (next instanceof Enumerator) {
					list.add(((Enumerator) next).getLiteral());
				} else {
					list.add(next);
				}
			}
			featureValue = list;
		} else {
			featureValue = result;
		}
		return featureValue;
	}

	/**
	 * Gets the structural feature of the given feature name of the object.
	 * 
	 * @param object
	 *            is the object
	 * @param name
	 *            is the feature name
	 * @return the structural feature
	 * @throws FactoryException
	 *             if an EStructuralFeature could not be created.
	 */
	public static EStructuralFeature eStructuralFeature(EObject object, String name) throws FactoryException {
		EStructuralFeature structuralFeature = object.eClass().getEStructuralFeature(name);
		if (structuralFeature != null) {
			return structuralFeature;
		} else {
			throw new FactoryException(AcceleoToolsMessages.getString("EFactory.UnresolvedLink", new Object[] { name, object.eClass().getName(), })); //$NON-NLS-1$
		}
	}

	/**
	 * Invokes the method called name from object.
	 * 
	 * @param object
	 *            is the object
	 * @param name
	 *            is the feature name, or a method defined on EObject like
	 *            'eClass', 'eResource', 'eContainer', 'eContainingFeature',
	 *            'eContainmentFeature', 'eContents', 'eAllContents',
	 *            'eCrossReferences'
	 * @param arg
	 *            arguments for the call
	 * @return the result of the method.
	 * @throws FactoryException
	 *             if the given feature could not be accessed.
	 */
	private static Object eCall(Object object, String name, Object arg) throws FactoryException {
		Class<?>[] parameterTypes;
		Object[] parameters;
		if (arg != null) {
			parameterTypes = new Class[] { arg.getClass() };
			parameters = new Object[] { arg };
		} else {
			parameterTypes = new Class[] {};
			parameters = new Object[] {};
		}
		Method method;
		try {
			method = object.getClass().getMethod(name, parameterTypes);
			return method.invoke(object, parameters);
		} catch (SecurityException e) {
			throw new FactoryException(e.getMessage());
		} catch (NoSuchMethodException e) {
			throw new FactoryException(e.getMessage());
		} catch (IllegalArgumentException e) {
			throw new FactoryException(e.getMessage());
		} catch (IllegalAccessException e) {
			throw new FactoryException(e.getMessage());
		} catch (InvocationTargetException e) {
			throw new FactoryException(e.getMessage());
		}
	}

	/**
	 * Indicates if the object has a value for the feature name.
	 * 
	 * @param object
	 *            is the object
	 * @param name
	 *            is the feature name
	 * @return if the feature is a list, return feature.size() > 0 else return
	 *         feature != null
	 */
	public static boolean eExist(EObject object, String name) {
		try {
			Object eGet = eGet(object, name);
			if (eGet != null) {
				if (eGet instanceof List) {
					return ((List) eGet).size() > 0;
				} else {
					return true;
				}
			} else {
				return false;
			}
		} catch (FactoryException e) {
			return false;
		}
	}

	/**
	 * Indicates if the object contains the given value for the feature name.
	 * 
	 * @param object
	 *            is the object
	 * @param name
	 *            is the feature name
	 * @param arg
	 *            is the value to find, null is allowed
	 * @return true if the object contains the given value for the feature name
	 */
	public static boolean eExist(EObject object, String name, Object arg) {
		try {
			Object eGet = eGet(object, name);
			if (eGet != null && eGet instanceof List) {
				return ((List) eGet).contains(arg);
			} else {
				return eGet == arg;
			}
		} catch (FactoryException e) {
			return false;
		}
	}
}
