/*
 * Copyright (c) 2005-2007 Obeo.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Obeo - initial API and implementation
 */

package org.acceleo.tools.log;

import java.util.Comparator;
import java.util.Iterator;
import java.util.TreeSet;

/**
 * This is a result log with a value, errors, and warnings.
 * 
 * @author www.obeo.fr
 * 
 */
// TODO revoir la gestion des logs
public abstract class AbstractTreeLog {

	/**
	 * The comparator used to order the sorted set of failures. It guarantees
	 * that the sorted set will be in ascending failure order, sorted according
	 * to the natural order of the failures (method position).
	 */
	protected static Comparator<IFailure> failureComparator = new Comparator<IFailure>() {
		public int compare(IFailure arg0, IFailure arg1) {
			if (arg0.position() < arg1.position()) {
				return -1;
			} else {
				return 1;
			}
		}
	};

	/**
	 * The value transmitted from a node to another.
	 */
	protected Object value;

	/**
	 * The sorted set of errors.
	 */
	protected TreeSet<IFailure> errors = new TreeSet<IFailure>(failureComparator);

	/**
	 * The sorted set of warnings.
	 */
	protected TreeSet<IFailure> severeWarnings = new TreeSet<IFailure>(failureComparator);

	/**
	 * Constructor.
	 */
	protected AbstractTreeLog() {
		// nothing to do here
	}

	/**
	 * Constructor.
	 * 
	 * @param value
	 *            is the value transmitted from a node to another
	 */
	protected AbstractTreeLog(Object value) {
		this.value = value;
	}

	/**
	 * Constructor.
	 * 
	 * @param failure
	 *            is the first failure
	 */
	protected AbstractTreeLog(IFailure failure) {
		addError(failure);
	}

	/**
	 * Give the value transmitted from a node to another.
	 * 
	 * @return the value transmitted from a node to another
	 */
	public Object getValue() {
		return value;
	}

	/**
	 * Set value transmitted from a node to another.
	 * 
	 * @param value
	 *            is the value transmitted from a node to another
	 */
	public void setValue(Object value) {
		this.value = value;
	}

	/**
	 * Adds the errors and warnings of the other log.
	 * 
	 * @param otherLog
	 *            is the other log
	 * @param errorsToSevereWarnings
	 *            indicates if the errors of the other log become warnings
	 */
	public void getAll(AbstractTreeLog otherLog, boolean errorsToSevereWarnings) {
		if (errorsToSevereWarnings) {
			severeWarnings.addAll(otherLog.errors);
		} else {
			errors.addAll(otherLog.errors);
		}
		severeWarnings.addAll(otherLog.severeWarnings);
	}

	/**
	 * Indicates if log is valid.
	 * 
	 * @return true if log is valid
	 */
	public boolean isOk() {
		return !hasError();
	}

	/**
	 * Indicates if log has errors. The warnings are ignored.
	 * 
	 * @return true if log has errors
	 */
	public boolean hasError() {
		return errors.size() > 0;
	}

	/**
	 * Indicates if log has warnings. The errors are ignored.
	 * 
	 * @return true if log has warnings
	 */
	public boolean hasSevereWarning() {
		return severeWarnings.size() > 0;
	}

	/**
	 * Adds a new error.
	 * 
	 * @param failure
	 *            is the new error
	 */
	protected void addError(IFailure failure) {
		errors.add(failure);
	}

	/**
	 * Iterates on all errors.
	 * 
	 * @return an iterator on all errors
	 */
	public Iterator<IFailure> allErrors() {
		return errors.iterator();
	}

	/**
	 * Gets last error.
	 * 
	 * @return the last error
	 */
	public IFailure lastError() {
		if (errors.size() > 0) {
			return errors.last();
		} else {
			return null;
		}
	}

	/**
	 * Adds a new warning.
	 * 
	 * @param failure
	 *            is the new warning
	 */
	protected void addSevereWarning(IFailure failure) {
		severeWarnings.add(failure);
	}

	/**
	 * Iterates on all warnings.
	 * 
	 * @return an iterator on all warnings
	 */
	public Iterator<IFailure> allSevereWarnings() {
		return severeWarnings.iterator();
	}

	/**
	 * Iterates on all errors and warnings.
	 * 
	 * @return an iterator on all errors and warnings
	 */
	public Iterator<IFailure> allOrderedErrorsAndSevereWarnings() {
		TreeSet<IFailure> all = new TreeSet<IFailure>(failureComparator);
		all.addAll(errors);
		all.addAll(severeWarnings);
		return all.iterator();
	}

	/**
	 * {@inheritDoc}
	 */
	public String toString() {
		String message = ""; //$NON-NLS-1$
		Iterator<IFailure> err = allOrderedErrorsAndSevereWarnings();
		while (err.hasNext()) {
			Object failure = err.next();
			message += failure.toString() + '\n';
		}
		return message;
	}
}
