/*
 * Copyright (c) 2005-2008 Obeo.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Obeo - initial API and implementation
 */
package org.acceleo.tools.resources;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;

import org.eclipse.emf.common.util.URI;

/**
 * Utility class for file system and resources access.
 * 
 * @author <a href="mailto:yvan.lussaud@obeo.fr">Yvan Lussaud</a>
 * 
 */
public final class Resources {

	/**
	 * Private Constructor. Utility class should not have default or public
	 * constructor.
	 */
	private Resources() {
		// nothing to do here
	}

	/**
	 * A tool method to get the content of the file specified by its absolute
	 * path.
	 * 
	 * @param fileName
	 *            the absolute path of the file
	 * @return the content of the file
	 */
	public static StringBuffer getAbsolutFileContent(String fileName) {
		final StringBuffer ret = new StringBuffer();
		FileInputStream is;
		try {
			is = new FileInputStream(fileName);
			int length = 0;
			final int bufferSize = 512;
			final byte[] buffer = new byte[bufferSize];
			try {
				while ((length = is.read(buffer)) > 0) {
					ret.append(new String(buffer, 0, length));
				}
			} catch (final IOException e) {
				// TODO YLU Auto-generated catch block
				e.printStackTrace();
			} finally {
				try {
					is.close();
				} catch (final IOException e) {
					// TODO YLU Auto-generated catch block
					e.printStackTrace();
				}
			}
		} catch (final FileNotFoundException e) {
			// TODO YLU Auto-generated catch block
			e.printStackTrace();
		}
		return ret;
	}

	/**
	 * A tool method to get the content of the specified file.
	 * 
	 * @param resourceName
	 *            the resource name
	 * @return the content of the file
	 */
	public static String getFileContent(String resourceName) {
		StringBuffer ret = new StringBuffer();
		final InputStream is = ClassLoader.getSystemResourceAsStream(resourceName);
		if (is != null) {
			int length = 0;
			final int bufferSize = 100 * 1024;
			final byte[] buffer = new byte[bufferSize];
			try {
				while ((length = is.read(buffer)) > 0) {
					ret.append(new String(buffer, 0, length));
				}
			} catch (final IOException e) {
				// TODO YLU Auto-generated catch block
				e.printStackTrace();
			} finally {
				try {
					is.close();
				} catch (final IOException e) {
					// TODO YLU Auto-generated catch block
					e.printStackTrace();
				}
			}
		} else {
			if (resourceName.startsWith("/")) { //$NON-NLS-1$
				ret = new StringBuffer(getFileContent(resourceName.substring(1)));
			}
		}
		return ret.toString();
	}

	/**
	 * Create the directory hierarchy and write the file with the given content.
	 * 
	 * @param outFileName
	 *            the name of the file to write
	 * @param fileContent
	 *            the content of the file
	 */
	public static void writeFile(String outFileName, String fileContent) {
		writeFile(new File(outFileName), fileContent);
	}

	/**
	 * Create the directory hierarchy and write the file with the given content.
	 * 
	 * @param outFile
	 *            the file to write
	 * @param fileContent
	 *            the content of the file
	 */
	public static void writeFile(File outFile, String fileContent) {
		final File outDir = new File(outFile.getPath().substring(0, outFile.getPath().lastIndexOf(File.separatorChar)));
		outDir.mkdirs();
		FileWriter f;
		try {
			f = new FileWriter(outFile);
			final BufferedWriter out = new BufferedWriter(f);
			out.write(fileContent);
			out.close();
		} catch (final IOException e) {
			// TODO YLU Auto-generated catch block
			e.printStackTrace();
		}
	}

	/**
	 * Return URI from a resource name.
	 * 
	 * @param resourceName
	 *            name of the resource
	 * @return URI from a resource name.
	 */
	public static URI getResourceURI(String resourceName) {
		URI res = null;
		final URL resourceURL = ClassLoader.getSystemResource(resourceName);
		if (resourceURL != null) {
			res = URI.createURI(resourceURL.toString());
		} else {
			if (resourceName.startsWith("/")) { //$NON-NLS-1$
				res = getResourceURI(resourceName.substring(1));
			}
			if (res == null) {
				res = URI.createFileURI(resourceName);
			}
		}
		return res;
	}

	/**
	 * Return true if a resource is associated to the given name.
	 * 
	 * @param resourceName
	 *            name of the resource
	 * @return true if a resource is associated to the given name.
	 */
	public static boolean resourceExist(String resourceName) {
		boolean res = ClassLoader.getSystemResource(resourceName) != null;
		if (!res && resourceName.startsWith("/")) { //$NON-NLS-1$
			res = resourceExist(resourceName.substring(1));
		}
		return res;
	}

	/**
	 * Return java.io.File from a resource name.
	 * 
	 * @param resourceName
	 *            name of the resource
	 * @return java.io.File from a resource name.
	 */
	public static File getResourceFile(String resourceName) {
		File res = null;
		// TODO YLU add some tests (resource not found, file not found, ...)
		final URL resourceURL = ClassLoader.getSystemResource(resourceName);
		if (resourceURL != null) {
			res = new File(resourceURL.getFile());
			if (!res.exists()) {
				res = null;
			}
		}
		if (res == null) {
			res = new File(resourceName);
			if (!res.exists()) {
				res = null;
			}
		}

		if (res == null) {
			final byte[] buffer = new byte[100 * 1024];
			int length;

			final InputStream resourceStream = ClassLoader.getSystemResourceAsStream(resourceName);
			if (resourceStream != null) {
				try {
					final File temp = File.createTempFile("resource", resourceName.substring(resourceName.lastIndexOf('.'))); //$NON-NLS-1$
					temp.deleteOnExit();

					final FileOutputStream out = new FileOutputStream(temp);

					while ((length = resourceStream.read(buffer)) >= 0) {
						out.write(buffer, 0, length);
					}
					res = temp;
				} catch (final IOException e) {
					// res = null;
				}
			}
		}
		return res;
	}

	/**
	 * Creates a platform-relative path URI.
	 * 
	 * @param path
	 *            is the path of the URI to create
	 * @return the new URI
	 * @see org.eclipse.emf.common.util.URI#createPlatformResourceURI
	 */
	public static URI createPlatformResourceURI(String path) {
		if (path.startsWith("platform:/plugin")) { //$NON-NLS-1$
			return URI.createURI(path, true);
		} else {
			return URI.createPlatformResourceURI(path, true);
		}
	}
}
