/*
 * Copyright (c) 2008 Obeo.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Obeo - initial API and implementation
 */

package org.acceleo.engine.eval.imports.services;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.ecore.EObject;

/**
 * Xpath axes.
 * 
 * @author www.obeo.fr
 * 
 */
public final class XpathServices implements IThreadSafeService {

	/**
	 * Same meaning as the "ancestor" axe in Xpath : selects all ancestors
	 * (parent, grandparent, etc.) of the current node.
	 * 
	 * @param object
	 *            is the current node of generation
	 * @return all ancestors (parent, grandparent, etc.) of the current node
	 */
	public List<EObject> ancestor(EObject object) {
		EObject localObject = object;
		final List<EObject> result = new ArrayList<EObject>();
		localObject = localObject.eContainer();
		while (localObject != null) {
			result.add(localObject);
			localObject = localObject.eContainer();
		}
		return result;
	}

	/**
	 * Same meaning as the "parent" axe in Xpath : selects the parent of the
	 * current node.
	 * 
	 * @param object
	 *            is the current node of generation
	 * @return the parent of the current node
	 */
	public EObject parent(EObject object) {
		return object.eContainer();
	}

	/**
	 * Same meaning as the "self" axe in Xpath : selects the current node.
	 * 
	 * @param object
	 *            is the current node of generation
	 * @return the current node
	 */
	public EObject self(EObject object) {
		return object;
	}

	/**
	 * Same meaning as the "child" axe in Xpath : selects all children of the
	 * current node.
	 * 
	 * @param object
	 *            is the current node of generation
	 * @return all children of the current node
	 */
	public List<EObject> child(EObject object) {
		return object.eContents();
	}

	/**
	 * Same meaning as the "descendant" axe in Xpath: Gets all the descendants
	 * (children, grandchildren, etc.) of the current node.
	 * 
	 * @param object
	 *            is the current node
	 * @return the descendants of the current node
	 */
	public List<EObject> descendant(EObject object) {
		final List<EObject> result = new ArrayList<EObject>();
		final TreeIterator<EObject> it = object.eAllContents();
		while (it.hasNext()) {
			result.add(it.next());
		}
		return result;
	}

	/**
	 * Same meaning as the "preceding-sibling" axe in Xpath: selects all
	 * siblings before the current node.
	 * 
	 * @param object
	 *            is the current node
	 * @return all siblings before the current node
	 */
	public List<EObject> precedingSibling(EObject object) {
		final List<EObject> result = new ArrayList<EObject>();
		final EObject parent = object.eContainer();
		if (parent != null) {
			for (final EObject o : parent.eContents()) {
				if (o == object) {
					break;
				} else {
					result.add(o);
				}
			}
		}
		return result;
	}

	/**
	 * Same meaning as the "preceding" axe in Xpath: selects everything in the
	 * model that is before the current node.
	 * 
	 * @param object
	 *            is the current node
	 * @return everything in the model that is before the current node
	 */
	public List<EObject> preceding(EObject object) {
		final List<EObject> result = new ArrayList<EObject>();
		final List<EObject> ancestors = ancestor(object);
		ancestors.add(object);
		// now return all preceding siblings and their respective children...
		for (final EObject ancestor : ancestors) {
			if (ancestor != object) {
				result.add(ancestor);
			}
			result.addAll(precedingSibling(ancestor));
			for (final EObject precedingSibling : precedingSibling(ancestor)) {
				result.addAll(descendant(precedingSibling));
			}
		}
		return result;
	}

	/**
	 * Same meaning as the "following-sibling" axe in Xpath: selects all
	 * siblings after the current node.
	 * 
	 * @param object
	 *            is the current node
	 * @return all siblings after the current node
	 */
	public List<EObject> followingSibling(EObject object) {
		final List<EObject> result = new ArrayList<EObject>();
		final EObject parent = object.eContainer();
		final Iterator<EObject> it = parent.eContents().iterator();
		while (it.hasNext()) {
			final Object child = it.next();
			if (child == object) {
				break;
			}
		}
		while (it.hasNext()) {
			result.add(it.next());
		}
		return result;
	}

	/**
	 * Same meaning as the "following" axe in Xpath: selects everything in the
	 * model after the current node.
	 * 
	 * @param object
	 *            is the current node
	 * @return everything after the current node
	 */
	public List<EObject> following(EObject object) {
		final List<EObject> result = new ArrayList<EObject>();
		final Iterator<EObject> it = followingSibling(object).iterator();
		while (it.hasNext()) {
			final EObject following = it.next();
			result.add(following);
			result.addAll(descendant(following));
		}
		return result;
	}

}
