/*
 * Copyright (c) 2005-2008 Obeo.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Obeo - initial API and implementation
 */

package org.acceleo.tools.factory;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.WeakHashMap;

import org.acceleo.tools.AcceleoToolsMessages;
import org.acceleo.tools.resources.Resources;
import org.eclipse.core.runtime.Platform;
import org.eclipse.emf.common.EMFPlugin;
import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.Diagnostic;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.EcoreFactory;
import org.eclipse.emf.ecore.EcorePackage;
import org.eclipse.emf.ecore.InternalEObject;
import org.eclipse.emf.ecore.impl.EcorePackageImpl;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.emf.ecore.util.Diagnostician;
import org.eclipse.emf.ecore.xmi.impl.EcoreResourceFactoryImpl;
import org.eclipse.emf.ecore.xmi.impl.XMIResourceFactoryImpl;

/**
 * This contains general support for ecore browsing.
 * 
 * @author www.obeo.fr
 * 
 */
public final class ETools {

	/** Resource set to load XMI. */
	private static ResourceSet resourceSet;

	/**
	 * Private constructor.
	 */
	private ETools() {
		// nothing to do here
	}

	/**
	 * Returns a URI for the eObject, i.e., either the eProxyURI, the URI of the
	 * eResource with the fragment produced by the eResource, or the URI
	 * consisting of just the fragment that would be produced by a default
	 * Resource with the eObject as its only contents.
	 * 
	 * @param object
	 *            the object for which to get the URI.
	 * @return the URI for the object.
	 */
	public static String getURI(EObject object) {
		String toReturn;
		if (object == null) {
			toReturn = ""; //$NON-NLS-1$
		} else if (object.eResource() != null) {
			toReturn = object.eResource().getURIFragment(object);
		} else {
			// inspired from EMF sources
			StringBuffer result = new StringBuffer("//"); //$NON-NLS-1$
			List<String> uriFragmentPath = new ArrayList<String>();
			EObject localObject = object;
			for (EObject container = localObject.eContainer(); container != null; container = localObject.eContainer()) {
				uriFragmentPath.add(((InternalEObject) container).eURIFragmentSegment(localObject.eContainmentFeature(), localObject));
				localObject = container;
			}
			int size = uriFragmentPath.size();
			if (size > 0) {
				for (int i = size - 1; i >= 0; --i) {
					result.append(uriFragmentPath.get(i));
					result.append('/');
				}
			}
			toReturn = result.toString();
		}
		return toReturn;
	}

	/**
	 * Creates root package for the metamodel nsURI or the metamodel file path.
	 * <p>
	 * uri2EPackage(path,true)
	 * 
	 * @param path
	 *            is the metamodel nsURI or the metamodel file path
	 * @return the root package of the metamodel
	 */
	// public static EPackage uri2EPackage(String path) {
	// return uri2EPackage(path, true);
	// }
	/**
	 * Creates root package for the metamodel nsURI or the metamodel file path.
	 * 
	 * @param path
	 *            is the metamodel nsURI or the metamodel file path
	 * @param forceReload
	 *            indicates that the reload of the metamodel is forced, if
	 *            forceReload is false then file modification stamp is used for
	 *            optimization
	 * @return the root package of the metamodel
	 */
	// public static EPackage uri2EPackage(String path, boolean forceReload) {
	// if (path != null && path.length() > 0) {
	// path = path.trim();
	// EPackage regValue = EPackage.Registry.INSTANCE.getEPackage(path);
	// if (regValue != null) {
	// return regValue;
	// } else {
	// return ecore2EPackage(path, forceReload);
	// }
	// } else {
	// return null;
	// }
	// }
	/**
	 * Creates root package for the ecore file path.
	 * 
	 * @param path
	 *            is the ecore file path
	 * @param forceReload
	 *            indicates that the reload of the metamodel is forced, if
	 *            forceReload is false then file modification stamp is used
	 * @return the root package of the metamodel
	 */
	public static EPackage ecore2EPackage(String path, boolean forceReload) {
		// TODO maps to store metamodels (for a fast loading) have been not
		// reimplemented for the moment (see Acceleo 2.X ETools class)
		// TODO use URL instead of File
		File ecoreFile = new File(path);
		boolean isExternal = true;
		if (!ecoreFile.exists() && EMFPlugin.IS_ECLIPSE_RUNNING) {
			ecoreFile = new File(Platform.getLocation() + path);
			isExternal = false;
		}
		if (ecoreFile != null && ecoreFile.exists()) {
			if (forceReload) {
				return subEcore2EPackage(path, isExternal);
			}
		} else if (forceReload) {
			return subEcore2EPackage(path, true);
		}
		return null;
	}

	private static Map<String, EPackage> ecore2EPackage = new WeakHashMap<String, EPackage>();

	private static Map<String, Double> ecore2OldModificationStamp = new HashMap<String, Double>();

	private static EPackage subEcore2EPackage(String path, boolean isExternal) {
		URI ecoreURI;
		if (isExternal) {
			ecoreURI = Resources.getResourceURI(path);
		} else {
			ecoreURI = Resources.createPlatformResourceURI(path);
		}
		org.eclipse.emf.ecore.resource.Resource.Factory.Registry reg = org.eclipse.emf.ecore.resource.Resource.Factory.Registry.INSTANCE;
		reg.getExtensionToFactoryMap().put("ecore", new EcoreResourceFactoryImpl()); //$NON-NLS-1$
		org.eclipse.emf.ecore.resource.Resource ecoreResource = getResourceSet().getResource(ecoreURI, true);
		// EcoreUtil.resolveAll(ecoreResourceSet);
		Object result = ecoreResource.getContents().size() > 0 ? ecoreResource.getContents().get(0) : null;
		if (result instanceof EPackage) {
			String nsURI = getNsURI((EPackage) result);
			if (nsURI != null) {
				EPackage regValue = EPackage.Registry.INSTANCE.getEPackage(nsURI);
				if (regValue != null) {
					return regValue;
				}
			}
			// TODO Msg to return : ETools.ModelValidationNeeded = You have to
			// validate the ecore {0}
			return (EPackage) validate((EPackage) result, false, "ETools.ModelValidationNeeded"); //$NON-NLS-1$
		} else {
			return null;
		}
	}

	private static String getNsURI(EPackage p) {
		String pNsURI = p.getNsURI();
		if (pNsURI != null && pNsURI.length() > 0) {
			return pNsURI;
		}
		if (p.getESubpackages().size() == 1) {
			return getNsURI(p.getESubpackages().get(0));
		} else {
			return null;
		}
	}

	/**
	 * Indicates if the metamodel nsURI or the metamodel file path exists.
	 * 
	 * @param path
	 *            is the metamodel nsURI or the metamodel file path
	 * @param nsUriOnly
	 *            indicates if only nsURI are allowed
	 * @return true if the metamodel nsURI exists or if the metamodel file path
	 *         exists
	 */
	public static boolean validateURI(String path, boolean nsUriOnly) {
		boolean res = EPackage.Registry.INSTANCE.get(path) != null;

		if (!res && !nsUriOnly) {
			String localPath = path.substring(0, path.lastIndexOf(".")) + "ecore"; //$NON-NLS-1$ //$NON-NLS-2$
			if (Resources.resourceExist(path) && Resources.resourceExist(localPath)) {
				res = true;
			} else {
				res = Resources.resourceExist(path);
			}
		}
		return res;
	}

	/**
	 * Search all the classifiers recursively in a package.
	 * 
	 * @param ePackage
	 *            is the container
	 * @return table of classifiers
	 */
	public static EClassifier[] computeAllClassifiers(EPackage ePackage) {
		List<EClassifier> classifiers = computeAllClassifiersList(ePackage);
		return classifiers.toArray(new EClassifier[] {});
	}

	/**
	 * Search all the classifiers recursively in a package.
	 * 
	 * @param ePackage
	 *            is the container
	 * @return list of classifiers
	 */
	public static List<EClassifier> computeAllClassifiersList(EPackage ePackage) {
		return computeAllClassifiersList(ePackage, false);
	}

	/**
	 * Search all the classifiers recursively in a package.
	 * 
	 * @param ePackage
	 *            is the container
	 * @param classOnly
	 *            indicates that only the classes are kept
	 * @return list of classifiers
	 */
	public static List<EClassifier> computeAllClassifiersList(EPackage ePackage, boolean classOnly) {
		List<EClassifier> classifiers = new BasicEList<EClassifier>();
		if (ePackage != null) {
			computeAllClassifiersList(ePackage, classifiers, classOnly);
		}
		return classifiers;
	}

	/**
	 * Adds all the classifiers from ePackage in all list.
	 * 
	 * @param ePackage
	 *            is the container
	 * @param all
	 *            list of classifiers
	 * @param classOnly
	 *            indicates that only the classes are kept
	 */
	private static void computeAllClassifiersList(EPackage ePackage, List<EClassifier> all, boolean classOnly) {
		for (EClassifier classifier : ePackage.getEClassifiers()) {
			if (!classOnly) {
				all.add(classifier);
			} else {
				if (classifier instanceof EClass && !((EClass) classifier).isAbstract() && !((EClass) classifier).isInterface()) {
					all.add(classifier);
				}
			}
		}
		for (EPackage subPackage : ePackage.getESubpackages()) {
			computeAllClassifiersList(subPackage, all, classOnly);
		}
	}

	/**
	 * Search a classifier recursively in a package.
	 * <p>
	 * Remarks :
	 * <li>It never returns classifier java.resources.Folder if name = "File"</li>
	 * <li>It never returns classifier java.resources.Folder if name = "older"</li>
	 * <li>It returns classifier java.resources.Folder for "Folder" or
	 * "resources.Folder"</li>
	 * 
	 * @param ePackage
	 *            is the container
	 * @param name
	 *            is the classifier identifier
	 * @return classifier or null if not found
	 */
	public static EClassifier getEClassifier(EPackage ePackage, String name) {
		if (ePackage == null || name == null) {
			return null;
		}
		String trimedName = name.trim();
		EClassifier get = null;
		/*
		 * Special case for EObject as we want a generic type for all the
		 * elements.
		 */
		if (trimedName.equals("EObject") || trimedName.equals("ecore.EObject")) { //$NON-NLS-1$ //$NON-NLS-2$
			get = EcorePackage.eINSTANCE.getEObject();
		}
		Iterator<EClassifier> classifiers = ePackage.getEClassifiers().iterator();
		while (get == null && classifiers.hasNext()) {
			EClassifier classifier = classifiers.next();
			String instanceClassName = '.' + getEClassifierPath(classifier);
			String endsWith = '.' + trimedName;
			if (instanceClassName.endsWith(endsWith)) {
				get = classifier;
			}
		}
		Iterator<EPackage> packages = ePackage.getESubpackages().iterator();
		while (get == null && packages.hasNext()) {
			EClassifier classifier = getEClassifier(packages.next(), trimedName);
			if (classifier != null) {
				get = classifier;
			}
		}
		return get;
	}

	/**
	 * Returns the feature with this classifier and this name.
	 * 
	 * @param currentEClassifier
	 *            is the classifier
	 * @param name
	 *            is the feature name
	 * @return the feature
	 */
	public static EStructuralFeature getEStructuralFeature(EClassifier currentEClassifier, String name) {
		if (currentEClassifier != null && currentEClassifier instanceof EClass) {
			return getEStructuralFeature((EClass) currentEClassifier, name);
		} else {
			return null;
		}
	}

	/**
	 * Returns the feature with this class and this name.
	 * 
	 * @param currentEClass
	 *            is the class
	 * @param name
	 *            is the feature name
	 * @return the feature
	 */
	public static EStructuralFeature getEStructuralFeature(EClass currentEClass, String name) {
		String trimedName = name.trim();
		if (currentEClass != null) {
			return currentEClass.getEStructuralFeature(trimedName);
		} else {
			return null;
		}
	}

	/**
	 * Returns a factory path for a classifier. There is a factory by package.
	 * The factory is used to create instances of classifiers.
	 * <p>
	 * Sample : "java.resources.ResourcesFactory" is the path of the factory
	 * java.resources.ResourcesFactory
	 * 
	 * @param eClassifier
	 *            is the classifier
	 * @return the factory full path
	 */
	// public static String getEClassifierFactoryPath(EClassifier eClassifier) {
	// return getEClassifierFactoryID(eClassifier) + '.' +
	// getEClassifierFactoryName(eClassifier);
	// }
	/**
	 * Returns a factory ID for a classifier. There is a factory by package. The
	 * factory is used to create instances of classifiers.
	 * <p>
	 * Sample : "java.resources" is the ID of the factory
	 * java.resources.ResourcesFactory
	 * 
	 * @param eClassifier
	 *            is the classifier
	 * @return the factory ID
	 */
	// public static String getEClassifierFactoryID(EClassifier eClassifier) {
	// if (eClassifier != null) {
	// String path = eClassifier.getInstanceClassName();
	// if (path != null) {
	// int i = path.lastIndexOf('.');
	// if (i > -1) {
	// return path.substring(0, i);
	// }
	// } else {
	// if (eClassifier.getEPackage().eResource() != null) {
	// boolean genmodelExists = false;
	// String ecoreModelPath =
	// eClassifier.getEPackage().eResource().getURI().toString();
	// String s = "platform:/resource"; //$NON-NLS-1$
	// if (ecoreModelPath.startsWith(s)) {
	// ecoreModelPath = ecoreModelPath.substring(s.length());
	// genmodelExists = Resources.findFile(new
	// Path(ecoreModelPath).removeFileExtension().addFileExtension("genmodel"))
	// != null; //$NON-NLS-1$
	// }
	// if (genmodelExists) {
	// URI genModelURI =
	// eClassifier.getEPackage().eResource().getURI().trimFileExtension().appendFileExtension("genmodel");
	// //$NON-NLS-1$
	// GenModel genModel = (GenModel) genModels.get(genModelURI);
	// if (genModel == null && !genModels.containsKey(genModelURI)) {
	// Resource.Factory.Registry reg = Resource.Factory.Registry.INSTANCE;
	// reg.getExtensionToFactoryMap().put("genmodel", new
	// XMIResourceFactoryImpl()); //$NON-NLS-1$
	// ResourceSet genModelResourceSet = new ResourceSetImpl();
	// Resource genModelResource = genModelResourceSet.getResource(genModelURI,
	// true);
	// if (genModelResource.getContents().size() > 0) {
	// genModel = (GenModel) genModelResource.getContents().get(0);
	// }
	// genModels.put(genModelURI, genModel);
	// }
	// if (genModel != null) {
	// GenPackage genPackage = findGenPackage(genModel.getGenPackages(),
	// eClassifier.getEPackage());
	// if (genPackage != null) {
	// String basePackage = genPackage.getBasePackage();
	// if (basePackage != null && basePackage.length() > 0) {
	// return basePackage + '.' + eClassifier.getEPackage().getName();
	// }
	// }
	// }
	// }
	// }
	// if (eClassifier.getEPackage() != null) {
	// String name = eClassifier.getEPackage().getName();
	// EPackage container = (EPackage) eClassifier.getEPackage().eContainer();
	// while (container != null) {
	// name = container.getName() + '.' + name;
	// container = (EPackage) container.eContainer();
	// }
	// return name;
	// }
	// }
	// }
	// return ""; //$NON-NLS-1$
	// }
	// private static Map genModels = new WeakHashMap();
	// private static GenPackage findGenPackage(List genPackages, EPackage
	// ecorePackage) {
	// for (Iterator i = genPackages.iterator(); i.hasNext();) {
	// GenPackage genPackage = (GenPackage) i.next();
	// if (EcoreUtil.equals(genPackage.getEcorePackage(), ecorePackage)) {
	// return genPackage;
	// } else {
	// GenPackage nestedGenPackage =
	// findGenPackage(genPackage.getNestedGenPackages(), ecorePackage);
	// if (nestedGenPackage != null) {
	// return nestedGenPackage;
	// }
	// }
	// }
	// return null;
	// }
	/**
	 * Returns a factory name for a classifier. There is a factory by package.
	 * The factory is used to create instances of classifiers.
	 * <p>
	 * Sample : "ResourcesFactory" is the name of the factory
	 * java.resources.ResourcesFactory
	 * 
	 * @param eClassifier
	 *            is the classifier
	 * @return the factory name
	 */
	public static String getEClassifierFactoryName(EClassifier eClassifier) {
		return getEClassifierFactoryShortName(eClassifier) + "Factory"; //$NON-NLS-1$
	}

	/**
	 * Returns a factory short name for a classifier. There is a factory by
	 * package. The factory is used to create instances of classifiers.
	 * <p>
	 * Sample : "Resources" is the short name of the factory
	 * java.resources.ResourcesFactory
	 * 
	 * @param eClassifier
	 *            is the classifier
	 * @return the factory short name
	 */
	public static String getEClassifierFactoryShortName(EClassifier eClassifier) {
		if (eClassifier != null) {
			EPackage p = eClassifier.getEPackage();
			String name = p.getName();
			if (name != null && name.length() > 0) {
				return name.substring(0, 1).toUpperCase() + name.substring(1);
			}
		}
		return ""; //$NON-NLS-1$
	}

	/**
	 * Returns the full path of the classifier.
	 * <p>
	 * Sample : "java.resources.JavaFile" is the full path for the classifier
	 * java.resources.JavaFile
	 * 
	 * @param eClassifier
	 *            is the classifier
	 * @return full path of the classifier
	 */
	public static String getEClassifierPath(EClassifier eClassifier) {
		String toReturn;
		if (eClassifier != null) {
			String instanceClassName = eClassifier.getInstanceClassName();
			String name = eClassifier.getName();
			if (eClassifier.getEPackage() != null) {
				EPackage container = eClassifier.getEPackage();
				if (container != null && instanceClassName != null && instanceClassName.endsWith(container.getName() + '.' + name)) {
					return instanceClassName;
				}
				while (container != null) {
					name = container.getName() + '.' + name;
					container = container.getESuperPackage();
				}
			}
			toReturn = name;
		} else {
			toReturn = null;
		}
		return toReturn;
	}

	/**
	 * Returns the short path of the classifier.
	 * <p>
	 * Sample : "resources.JavaFile" is the short path for the classifier
	 * java.resources.JavaFile
	 * 
	 * @param eClassifier
	 *            is the classifier
	 * @return short path of the classifier
	 */
	public static String getEClassifierShortPath(EClassifier eClassifier) {
		String name = eClassifier.getName();
		if (eClassifier.getEPackage() != null) {
			name = eClassifier.getEPackage().getName() + '.' + name;
		}
		return name;
	}

	/**
	 * Creates a package and his children in an ecore model. The children are
	 * separated by '.'
	 * <p>
	 * Sample : "a.b.c" is a package full path,
	 * <p>
	 * "a", "b", and "c" packages are created, the root package "a" is returned.
	 * 
	 * @param path
	 *            is the package full path
	 * @return the root package
	 */
	public static EPackage createPackageHierarchy(String path) {
		EPackage ePackage = null;
		if (path != null && path.length() > 0) {
			EcorePackage p = EcorePackageImpl.init();
			EcoreFactory factory = p.getEcoreFactory();
			EPackage parent = null;
			StringTokenizer st = new StringTokenizer(path, "."); //$NON-NLS-1$
			while (st.hasMoreTokens()) {
				String name = st.nextToken();
				EPackage child = factory.createEPackage();
				child.setName(name);
				if (parent != null) {
					parent.getESubpackages().add(child);
				} else {
					ePackage = child;
				}
				parent = child;
			}
		}
		return ePackage;
	}

	/**
	 * Get a package in a parent package.
	 * <p>
	 * Sample : "a.b" is a parent package and "c.d" is a relative path,
	 * <p>
	 * "a.b.c.d" package is returned.
	 * 
	 * @param parent
	 *            is the parent package
	 * @param path
	 *            is the relative path of the required package
	 * @return the required package, null if not found
	 */
	public static EPackage getEPackage(EPackage parent, String path) {
		EPackage ePackage = parent;
		if (path != null && path.length() > 0) {
			StringTokenizer st = new StringTokenizer(path, "."); //$NON-NLS-1$
			while (st.hasMoreTokens()) {
				String name = st.nextToken();
				boolean found = false;
				Iterator<EPackage> subPackages = ePackage.getESubpackages().iterator();
				while (!found && subPackages.hasNext()) {
					EPackage subPackage = subPackages.next();
					if (subPackage.getName().equals(name)) {
						found = true;
						ePackage = subPackage;
					}
				}
				if (!found) {
					return null;
				}
			}
		}
		return ePackage;
	}

	/**
	 * Indicates if an instance of the classifier is an instance of the type.
	 * 
	 * @param classifier
	 *            is the classifier
	 * @param type
	 *            is the type
	 * @return true if an instance of the classifier is an instance of the type
	 */
	public static boolean ofType(EClassifier classifier, String type) {
		if (classifier instanceof EClass) {
			return ofType((EClass) classifier, type);
		} else {
			return ofClass(classifier, type);
		}
	}

	/**
	 * Indicates if an instance of the class is an instance of the type.
	 * 
	 * @param eClass
	 *            is the class
	 * @param type
	 *            is the type
	 * @return true if an instance of the class is an instance of the type
	 */
	public static boolean ofType(EClass eClass, String type) {
		boolean isOfType = false;
		if (ofClass(eClass, type)) {
			isOfType = true;
		}
		for (EClass superType : eClass.getESuperTypes()) {
			if (ofType(superType, type)) {
				isOfType = true;
			}
		}
		return isOfType;
	}

	/**
	 * Indicates if the type corresponds to the name of the classifier.
	 * 
	 * @param classifier
	 *            is the classifier
	 * @param type
	 *            is the type
	 * @return true if the type corresponds to the name of the classifier
	 */
	public static boolean ofClass(EClassifier classifier, String type) {
		String path = ETools.getEClassifierPath(classifier);
		return ('.' + path).endsWith('.' + type);
	}

	/**
	 * Loads an EMF model.
	 * 
	 * @param path
	 *            is the path of the model to load
	 * @return the root element of the model
	 */
	public static EObject loadXMI(String path) {
		URI modelURI = null;
		// TODO path policy - check who to manage path :
		// compilation time ?
		// map new resources
		// (EcorePlugin.getPlatformResourceMap().put("resource", uri))
		// if (path.startsWith("platform:/plugin")) { //$NON-NLS-1$
		// modelURI = URI.createURI(path, true);
		// } else {
		modelURI = Resources.getResourceURI(path);
		// modelURI = URI.createPlatformResourceURI(path, true);
		// }
		String fileExtension = modelURI.fileExtension();
		if (fileExtension == null || fileExtension.length() == 0) {
			fileExtension = Resource.Factory.Registry.DEFAULT_EXTENSION;
		}
		Resource.Factory.Registry reg = Resource.Factory.Registry.INSTANCE;
		Object resourceFactory = reg.getExtensionToFactoryMap().get(fileExtension);
		if (resourceFactory != null) {
			getResourceSet().getResourceFactoryRegistry().getExtensionToFactoryMap().put(fileExtension, resourceFactory);
		} else {
			getResourceSet().getResourceFactoryRegistry().getExtensionToFactoryMap().put(fileExtension, new XMIResourceFactoryImpl());
		}

		Resource modelResource = getResourceSet().getResource(modelURI, true);

		// EcoreUtil.resolveAll(resourceSet);
		EObject result;
		if (modelResource.getContents().size() > 0) {
			result = modelResource.getContents().get(0);
		} else {
			result = null;
		}
		return ETools.validate(result, false, AcceleoToolsMessages.getString("ETools.ModelValidationRecommended", new Object[] { path, })); //$NON-NLS-1$
	}

	/**
	 * It validates the given EMF object.
	 * <p>
	 * An error is put in the acceleo log if the validation failed.
	 * <p>
	 * <li>!blocker || Diagnostic.OK => root</li>
	 * <li>blocker && !Diagnostic.OK => null</li>
	 * <li>root == null => null</li>
	 * 
	 * @param root
	 *            is the object to validate
	 * @param blocker
	 *            indicates if the result must be Diagnostic.OK
	 * @param message
	 *            is the error message to put in the acceleo log
	 * @return the given object or null
	 */
	public static EObject validate(EObject root, boolean blocker, String message) {
		if (root != null && !root.getClass().getName().startsWith("org.eclipse.uml2")) { //$NON-NLS-1$
			if (Diagnostician.INSTANCE.validate(root).getSeverity() != Diagnostic.OK) {
				// TODO Manage how to log this action
				// AcceleoEcorePlugin.getDefault().log(message, blocker);
				if (blocker) {
					return null;
				}
			}
		}
		return root;
	}

	public static ResourceSet getResourceSet() {
		if (resourceSet == null) {
			// Create a resource set.
			resourceSet = new ResourceSetImpl();

			// Register the default resource factory -- only needed for
			// stand-alone!
			resourceSet.getResourceFactoryRegistry().getExtensionToFactoryMap().put(org.eclipse.emf.ecore.resource.Resource.Factory.Registry.DEFAULT_EXTENSION, new XMIResourceFactoryImpl());
		}
		return resourceSet;
	}

	public static void setResourceSet(ResourceSet resourceSet) {
		ETools.resourceSet = resourceSet;
	}

}
